%%% -*- Module: fsa_tk; -*-
:- module(fsa_tk,
	  [ tk_fsa/1, tk_fsa/2,                        % fsa_tk
	    tk_fsa_anyway/1,
	    tk_fsa_file/1, tk_fsa_file/2,              % fsa_tk_file
	    tk_rx/0, tk_rx/1, tk_rx/2,                 % fsa_tk_rx
	    tk_regex/0, tk_regex/1, tk_regex/2,        % fsa_tk_regex
	    tk_read_regex/0, tk_read_regex/1,          % fsa_tk_regex_read
	    %% called from tk: wrap inside fsa_tk:
	    tk_operation/2,                            % -
	    recreate_fsa/1,                            
	    move_node/4, move_node0/4,
	    save_fsa/1, save_fsa/2,
	    display_sigma/0,
	    display_fa/0,
	    count_fa/0,
	    revert_fsa/1,
	    zoom/2,
	    tk_version/1,
	    tk_close/1,
	    protect/1,
	    expand_user_macros/2,                  % fsa_expand_user_macros
	    expand_all_macros/2                   % fsa_expand_all_macros
	    
	  ]).

:- dynamic current_trans/10.
:- dynamic current_state/7.
:- dynamic tkconsol/4.
:- dynamic current_state/6.

:- use_module(library(lists),  [ member/2            ]).
:- use_module(library(tcltk),  [ tcl_eval/3,
				 tk_new/2,
				 tk_do_one_event/0,
				 tcl_delete/1,
				 tk_terminal/5       ]).
:- use_module(library(system), [ delete_file/1 ]).

:- public unknown_predicate_handler/2.
unknown_predicate_handler(tkconsol(_,_,_,_),fail).
unknown_predicate_handler(current_state(_,_,_,_,_,_),fail).
unknown_predicate_handler(current_trans(_,_,_,_,_,_,_,_,_),fail).
unknown_predicate_handler(current_state(_,_,_,_,_,_,_),fail).
unknown_predicate_handler(current_trans(_,_,_,_,_,_,_,_,_,_),fail).
unknown_predicate_handler(fa(_,_,_,_,_,_),fail).
unknown_predicate_handler(fa(_,_,_,_,_,_,_),fail).
unknown_predicate_handler(final_td(_,_),fail).
unknown_predicate_handler(trans(_,_,_,_),fail).
unknown_predicate_handler(trans(_,_,_),fail).
unknown_predicate_handler(start(_),fail).
unknown_predicate_handler(final(_),fail).
unknown_predicate_handler(jump(_,_),fail).

%% if library(tcltk) is not present for some reason:
unknown_predicate_handler(tk_do_one_event,fail).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

user:help_info(class,fsa_tk,"The Graphical User Interface","
This section discusses the graphical user interface for FSA. It's mostly 
extremely obvious. So this is a kind of `If you click on the help
button, a help text will be displayed' explanation. Note that currently
the graphical user interface is available only under SICStus Prolog.
Take a look at http://www.let.rug.nl/~vannoord/Fsa/Manual/dump.png if you 
would like to know what the graphical user interface looks like.

*MENU*


The menu consists of a number of menu buttons. Really. The actions
associated with the menu buttons are:

* *File*

[Load]: Loads a file that is assumed to contain an automaton.

[LoadAux]: Loads a file that is assumed to contain auxiliary regular
expression macro's and operators.

[ReconsultAux]: Reconsults a file that is assumed to
contain auxiliary regular expression macro's and operators. This
allows tracing of your code.

[SaveAs]: Saves the automaton and the associated geometry
information in a file. Such a file can be read-in using the Load
button, or as the startup file. Various formats are available under
various [SaveAsX] buttons.

[Revert] Redraws the current automaton without the current
layout; a new layout will be computed.

[Redraw] Redraws the current automaton with the current 
layout.

[Close] Halts the graphical user interface, but FSA continues.

[Quit] Halts.

* *Settings*
A number of global variables can be set via this menu. Help is
available on-line.

* *Operations*
A number of unary operations on the current regular expression can be
fired through this menu. The operations are a subset of those allowed
in regular expressions.

* *Produce* Produces a number of example strings (pairs of
strings) accepted by the current finite automaton.

* *Visualization* Interface to a number of external
visualization tools. These only work if you have the tools installed
(really!), and the appropriate commands are in your PATH (yes, no
magic here either). 

* *Help* Well. What do you *think* this menubutton would do?

Now let's consider some of the other widgets maintained by the
graphical user interface:

* *Regex*

If a regular expression is typed in the field, then after hitting
<CR> the corresponding automaton will be visualised on the
canvas.

The [Expand User Macro's] and [Expand All Macro's] can be used to expand
all the macro's of the current regular expression. 

* *String*

If a sequence of symbols (separated by whatever the symbol_separator
flag requires) is typed in this field, then (after hitting <CR>, or
after pushing the `Submit' button) the system *runs* the current automaton 
on the input you provided. The actual way in which the automaton is run
depends on the value displayed in the radio-button available to the right 
of the `Submit' button.

* *Canvas*

The large canvas contains a picture of the current regular expression
(or automaton read-in from a file). Note that you can drag states to
alter the layout interactively. If you point your mouse to a label of
an edge, then the corresponding edge will become red temporarily (this
is useful for large labels). Also note that for states P and Q all
edges from P to Q are combined in a single edge. Start states are
green, final states are red and have a sunken relief. If a state is
both a start state and a final state, then it is green with a sunken
relief.

* *ToolBar*

The tool bar at the bottom consists of the following sub parts:

[EdgeAngle]: Text field should contain a real. Does a redraw using
the current (typically) new angle variable upon &lt;CR&gt;. Does not
require re-computation of layout.

[Xdistance]: Text field should contain an integer. Re-computes and
re-draws using the current (typically new) distance of states
parameter.

[Quality]: Re-computes and re-draws using the current (typically
new) parameters. 

[DisplaySigma]: Displays internal representation of alphabet and
symbols list of current automaton.

[DisplayFa]: Displays internal representation of current
automaton.

[CountFa]: Provides numerical information of current automaton.

[ClearCache]: Clears the cache of the regular expression
compiler.

[ZoomIn]: This does the opposite of ZoomOut.

[ZoomOut]: This does the opposite of ZoomIn.

Finally, the [Interp] button can be used to get the name of the current
Tcl/Tc interpreter. This is of interest only for development work.

* *TkConsol* (experimental)

As an experimental feature, you can include a widget displaying
standard input and standard output. If you want to try out this new
feature, you have to set the global variable *tkconsol* to *on*. E.g.:

	fsa tkconsol=on -tk

Note that this is currently not very robust. 


The following global variables are relevant for this module:

* tkconsol
* v_angle
* v_xdist
* use_graphviz
* no_display_beyond
").

user:help_info(fsa_tk,tk_fsa_file,"tk_fsa_file(+File)","
Starts a Tcl/Tk widget for the automaton read from File").
tk_fsa_file(File) :-
    tk_fsa_file(File,not_started).

tk_fsa_file(File,Started) :-
    remove_fsa(Started),
    tk_fsa_read_file(File,Fa),
    fsa_globals:fsa_global_set(fa,Fa),
    (	fsa_globals:fsa_global_get(regex,_)
    ->  true
    ;	fsa_globals:fsa_global_set(regex,file(File))
    ),
    recreate_fa(Fa,Started).

tk_fsa_read_file(File,Fa) :-
    fsa_globals:fsa_global_get(read,ReadFlag),
    tk_fsa_read_file(ReadFlag,File,Fa).

%% only if Flag=normal do we reconsult the file in order to get
%% perhaps existing geometry information
tk_fsa_read_file(Flag,File,Fa):-
    fsa_io:fsa_read_file(Flag,File,Fa),
    (	Flag == normal
    ->	prolog_flag(redefine_warnings,Old,off),
	reconsult(File),
	prolog_flag(redefine_warnings,_,Old)
    ;   true
    ).

user:help_info(fsa_tk,tk_fsa,"tk_fsa(+Fa)","
Starts a Tcl/Tk widget for the automaton Fa").
tk_fsa(Fa) :-
    tk_fsa(Fa,not_started).

tk_fsa(Fa,Started) :-
    remove_fsa(Started),
    fsa_globals:fsa_global_set(fa,Fa),
    initialize_tk(Started,Interp),
    print_states_and_transitions(Fa,Interp),
    tk_main_loop_if_no_top_level_events(Started,Interp).

tk_fsa_anyway(Started) :-
    fsa_globals:fsa_global_get(fa,Fa),
    fsa_globals:fsa_global_get(no_display_beyond,V0),
    fsa_globals:fsa_global_set(no_display_beyond,1000000),
    call_cleanup(tk_fsa(Fa,Started),
		 fsa_globals:fsa_global_set(no_display_beyond,V0)
		).


print_states_and_transitions(Fa,Interp) :-
    busy_fsa(started(Interp),'Computing geometry...'),
    call_cleanup(print_states_and_transitions_(Fa,Interp),
		 not_busy_fsa(started(Interp))
		).

print_states_and_transitions_(Fa,Interp):-
    fsa_data:fsa_states_number(Fa,States),
    fsa_globals:fsa_global_get(no_display_beyond,Value),
    (	States =< Value
    ->	fsa_globals:fsa_global_get(use_graphviz,Dot),
	print_states_and_transitions_(Fa,Interp,Dot)
    ;   %fsa_util:debug_message(0,
	% "no_display_beyond flag: ~w > ~w; not displaying automaton~n",
	%		       [States,Value])
	fsa_tk_message(Interp,"~w states; no_display_beyond=~w",
		       [States,Value])		       
    ).

% use tcldot
print_states_and_transitions_(Fa,Interp,on) :-
    fsa_visualization:fsa_temporary_file(Tmp),
    open(Tmp,write,Stream),
    current_output(Old),
    set_output(Stream),
    user:call_cleanup(fsa_visualization:fsa_dot(Fa),
                      ( set_output(Old), close(Stream) )),
    tcl(Interp,'\
           package require Tcldot
           set f [open ~a r]
	   set g [dotread $f]
	   close $f
	   $g setattributes rankdir LR
	   $g setattributes margin 0
	   $g setattributes center true
           $g layout
	   set c .t.c
           eval [$g render]',[Tmp]),
    delete_file(Tmp).

% use built-in renderer
print_states_and_transitions_(Fa,Interp,off) :-
    %% print the states
    tk_all_states(Fa,Interp,StateCoords),
    %% print the transitions:
    tk_transitions(StateCoords,StateCoords,Fa,Interp),
    ensure_visible(Interp).

fsa_tk_message(Interp,Format,Args) :-
    charsio:format_to_chars(Format,Args,Chars),
    tcl(Interp,"catch {destroy .t.c.too}",[]),
    tcl(Interp,"label .t.c.too -text {~s}",[Chars]),
    tcl(Interp,".t.c create window 0 0 -window .t.c.too -anchor center",[]),
    tcl(Interp,".t.c xview moveto 0.48",[]),
    tcl(Interp,".t.c xview moveto 0.48",[]).

tk_read_regex :-
    tk_read_regex(not_started).

tk_read_regex(Started) :-
    fsa_regex:default_ops(Ops),
    prompt(Old,'RegEx: '),
    fsa_util:read_with_ops(Ops,Regex),
    prompt(_,Old),
    tk_rx(Regex,Started).

tk_regex :-
    tk_regex('{}', not_started).

user:help_info(fsa_tk,tk_regex,"tk_regex(+Atom)","
Atom is an atom, converted to regular expression and compiled into
automaton. A Tcl/Tk widget is started for that automaton.").

tk_regex(AtomRegex) :-
    tk_regex(AtomRegex,not_started).

tk_regex(AtomRegex,Started) :-
    fsa_regex:default_ops(Ops),
    fsa_util:call_with_ops(Ops,fsa_util:atom_term(AtomRegex,Regex)),
    tk_rx(Regex,Started).

tk_rx :-
    tk_rx({},not_started).

user:help_info(fsa_tk,tk_rx,"tk_rx(+Expr)","
Atom is a regular expression and compiled into automaton. A Tcl/Tk 
widget is started for that automaton.").

tk_rx(Regex):-
    tk_rx(Regex,not_started).

tk_rx(Regex,Started) :-
    fsa_globals:fsa_global_set(regex,Regex),
    tk_regex0(Regex,Started).

tk_regex0(Regex,Started):-
    remove_fsa(Started),
    fsa_tk:busy_fsa(Started,'Compiling regular expression...'),
%    didn't work in 3.10.0
    call_cleanup( fsa_regex:fsa_regex_compile(Regex,Fa0),
		   not_busy_fsa_and_flags(Started)
		),
%    on_exception(E, fsa_regex:fsa_regex_compile(Regex,Fa0),
%		    ( not_busy_fsa_and_flags(Started),
%		      raise_exception(E)
%		    )
%		),
    not_busy_fsa_and_flags(Started),
    tk_fsa(Fa0,Started).

%% this is complicated because we want to compute only the last step
%% (i.e. the newly added operator), but we work with a complicated
%% regular expression whose structure we want to maintain in the regex
%% field.
tk_operation(Op,started(TclInterp)) :-
    fsa_globals:fsa_global_get(regex,OldRegex),
    fsa_globals:fsa_global_get(fa,Fa),
    Regex =.. [Op,fa(Fa)],
    tk_regex0(Regex,started(TclInterp)),
    OpOldRegex =.. [Op,OldRegex],
    fsa_globals:fsa_global_set(regex,OpOldRegex),
    send_flags(TclInterp). % extra, have this version of regex flag...

remove_fsa(not_started).

remove_fsa(started(Interp)) :-
    retractall(current_trans(_,_,_,_,_,_,_,_,_,Interp)),
    retractall(current_state(_,_,_,_,_,_,Interp)),
    tcl(Interp,".t.c delete all",[]).

busy_fsa(not_started,_).
busy_fsa(started(Interp),Text) :-
    tcl(Interp,"catch {destroy .t.c.busy}",[]),
    tcl(Interp,"label .t.c.busy -text {~a}",[Text]),
    tcl(Interp,".t.c create window 0 0 -window .t.c.busy -anchor center",[]),
    tcl(Interp,".t.c xview moveto 0.48",[]),
    tcl(Interp,".t.c yview moveto 0.48",[]),
    tcl(Interp,"update",[]).

not_busy_fsa_and_flags(not_started).
not_busy_fsa_and_flags(started(I)) :-
    tcl(I,"destroy .t.c.busy",[]),
    send_flags(I).


not_busy_fsa(not_started).
not_busy_fsa(started(Interp)) :-
    tcl(Interp,"destroy .t.c.busy",[]).

revert_fsa(Started) :-
    fsa_globals:fsa_global_get(fa,Fa),
    tk_fsa(Fa,Started).

ensure_visible(Interp) :-
    current_state(_,X0,Y0,_,_,_,Interp),
    !,
    X is  ((5000+X0)/10000) - 0.02,
    Y is  ((5000+Y0)/10000) - 0.02,
    tcl(Interp,".t.c xview moveto ~w",[X]),
    tcl(Interp,".t.c yview moveto ~w",[Y]).

%% already started
initialize_tk(started(Interp),Interp) :-
    send_flags(Interp).

initialize_tk(not_started,Interp) :-
    open_null_stream(Null),
    print(Null,' '),
    close(Null),
    tcltk:tk_new([top_level_events],'$TclInterp'(Interp)),
    user:bb_put(fsa_tk:last_interp,Interp),
    tcl(Interp,"set interp ~w",[Interp]),
    load_tcl_code(Interp),
    send_flags(Interp),
    menu_flags([c_with_main,
		cpp_with_main,
		count,
		debug,
		display_unused_states,
		hash_size,
		java_with_main,
		length_max,
		no_display_beyond,
		nr_sol_max,
		postscript_res,
		pred_module,
		pstricks_style,
		read,
		regex_cache,
		symbol_separator,
		symbol_separator_in,
		symbol_separator_out,
		tkconsol,
		to_c_conversion,
		to_cpp_conversion,
		to_java_conversion,
		use_graphviz,
		w_determinizer_minimum,
		v_algorithm,
		v_angle,
		v_tree_depth,
		v_xdist,
		write
	       ],Interp),
    start_help('$TclInterp'(Interp)),
    start_tkconsol('$TclInterp'(Interp)).

add_toplevel(I) :-
    tcl(I,"\
button .edits.prolog -text {Prolog Command} -command {
    prolog fsa_runtime_cmd_interpreter }
pack .edits.prolog -side right",[]).   

%% Windows systems do not support top_level events (i.e. allowing
%% tcltk to interpret events while SICStus is waiting for input).
tk_main_loop_if_no_top_level_events(started(_),_).
tk_main_loop_if_no_top_level_events(not_started,I) :-
    prolog_flag(host_type,Host),
    (   sub_atom(Host,_,_,_,win)
    ->  add_toplevel(I),
        tcltk:tk_main_loop
    ;   true
    ).

start_help(Interp) :-
    fsa_globals:fsa_global_get(tk_fsa_add_help_menu,Bool),
    start_help(Bool,Interp).
start_help(off,_).
start_help(on,Interp):-
    fsa_help:fsa_help_add_to_menu('.menu.help.m',Interp).

start_tkconsol(Interp) :-
    fsa_globals:fsa_global_get(tkconsol,Val),
    start_tkconsol(Val,Interp).

start_tkconsol(off,_Interp).
start_tkconsol(on,Interp) :-
    (	tkconsol(_,_,_,_)         %% only start a single tkconsol
    ->  true
    ;   continue_tkconsol(Interp)
    ).

continue_tkconsol(I) :-
    tcl_eval(I,'\
       frame .cons
       pack .cons -fill both  -after .t
       text .cons.console -width 80 -height 10 -yscrollcommand ".cons.sb set"
       bind Text <Control-u> {
           tkTextSetCursor %W {insert linestart}
       if {[%W compare insert == {insert lineend}]} {
           %W delete insert
       } else {
           %W delete insert {insert lineend}
       }
       }
       scrollbar .cons.sb -command ".cons.console yview"
       pack .cons.sb -side right -fill both
       pack .cons.console -fill both -expand 1  -side left',[]),
    (   on_exception(Error,tcltk:tk_terminal(I,".cons.console",InS,OutS,ErrS),
		     (   print_message(error,Error),
			 fail
		     )
		    )
    ->  prolog_flag(user_input, InOld, InS),
	prolog_flag(user_output, OutOld, OutS),
	prolog_flag(user_error, ErrOld, ErrS),
	assertz(tkconsol(I,InOld,OutOld,ErrOld)),
	set_input(InS),
	set_output(OutS),
	tcl_eval(I,'.menu.file.m add separator',[]),
	tcl_eval(I,
		 '.menu.file.m add command -label "Enlarge Console (1 line)" \
	   -command { .cons.console configure -height \
             [expr [lindex [.cons.console configure -height] 4]+1]
	   }',[]),
	tcl_eval(I,
		 '.menu.file.m add command -label "Enlarge Console (5 lines)" \
	   -command { .cons.console configure -height \
             [expr [lindex [.cons.console configure -height] 4]+5]
	   }',[]),
	tcl_eval(I,
		 '.menu.file.m add command -label "Shrink Console (1 line)" \
	   -command { .cons.console configure -height \
             [expr [lindex [.cons.console configure -height] 4]-1]
	   }',[]),
	tcl_eval(I,
		 '.menu.file.m add command -label "Shrink Console (5 lines)" \
	   -command { .cons.console configure -height \
             [expr [lindex [.cons.console configure -height] 4]-5]
	   }',[])
    ;   tcl_eval(I,'destroy .cons',[])
    ).

send_flags(Interp) :-
    fsa_globals:fsa_global_get(v_angle,Val),
    tcl(Interp,"set fsa(angle) ~w",[Val]),
    fsa_globals:fsa_global_get(v_xdist,Val2),
    tcl(Interp,"set fsa(xdist) ~w",[Val2]),
    (	fsa_globals:fsa_global_get(regex,Regex)
    ->	fsa_regex:default_ops(Ops),
	fsa_util:call_with_ops(Ops,fsa_tk:tcl(Interp,'next_regex {~q}',[Regex]))
    ;	tcl(Interp,'next_regex {""}',[])
    ).


%%%%%%%%%%%%
%% STATES %%
%%%%%%%%%%%%

%% print each state in turn
tk_all_states(Fa,Interp,StateCoords) :-
    fsa_visualization:compute_states(StateCoords,Fa),
    tk_the_states(StateCoords,Interp,Fa).

%% place the states on the canvas
tk_the_states([],_,_).
tk_the_states([s(H,X,Y)|Tail],Interp,Fa) :-
    tk_state(H,X,Y,Interp,Fa),
    tk_the_states(Tail,Interp,Fa).

tk_state(H,X,Y,Interp,Fa) :-
    color(H,Fa,Color),
    relief(H,Fa,Relief),
    create_state(H,X,Y,Color,Relief,Interp).

create_state(H,X,Y,Color,Relief,Interp) :-
    gen_sym(Name),
    tcl(Interp,"label .t.c.~w  -bg ~w -relief ~w -text {~w} -font {-*-times-bold-r-*-*-*-100-*-*-*-*-*-*}",
	    [Name,Color,Relief,H]),
    tcl(Interp,".t.c create window ~w ~w -window .t.c.~w -anchor center",
	    [X,Y,Name],Id),
    tcl(Interp,'bind .t.c.~w <1> "set cccx %x; set cccy %y"',
	   [Name]),
    tcl(Interp,'bind .t.c.~w <2> "prolog fsa_tk:showtrans_out(~w)"',[Name,H]),
    tcl(Interp,'bind .t.c.~w <3> "prolog fsa_tk:showtrans_in(~w)"',[Name,H]),
    tcl(Interp,'bind .t.c.~w <ButtonRelease-1> {
	   move_node {~w} [expr %x -$cccx] [expr %y - $cccy]}',
			       [Name,H]),
    assertz(current_state(H,X,Y,Id,Color,Relief,Interp)).

%% options for states:
%% start states are green,
%% final states are red (unless already green)
%% start and final are plum2
%% others are blue
%% 
%% final states have a ridge. So states that are both
%% start + final are green with a ridge.
color(State,Fa,Color) :-
    (	fsa_data:fsa_start_state(Fa,State)
    ->	(  fsa_data:fsa_final_state(Fa,State)
	->  Color = plum2
	;   Color = green
	)
    ;	(  fsa_data:fsa_final_state(Fa,State)
	->  Color = red
	;   Color = 'SteelBlue1'
	)
    ).

relief(State,Fa,Relief) :-
    (	fsa_data:fsa_final_state(Fa,State)
    ->	Relief = 'ridge -borderwidth 5'
    ;	Relief = flat
    ).

%%%%%%%%%%%%%%%%%
%% TRANSITIONS %%
%%%%%%%%%%%%%%%%%

%% print each transition in turn
tk_transitions([],_,_fa,_Interp).
tk_transitions([H|T],All,Fa,Interp) :-
    tk_transitions0(H,All,Fa,Interp),
    tk_transitions(T,All,Fa,Interp).

%% combine all transitions combining the same nodes into
%% a single transition with multiple labels
tk_transitions0(Start,All,Fa,Interp) :-
    (	setof(Sym,Fa^a_trans(Start,Sym,End,All,Fa),Syms),
	tk_transition(Start,Syms,End,Interp),
	fail
    ;	true
    ).

a_trans(s(Start,_X0,_Y0),Sym,s(End,X,Y),All,Fa) :-
    member(s(End,X,Y),All),
    fsa_visualization:transition_or_jump(Fa,Start,Sym,End).

tk_transition(s(Start,X0,Y0),Syms0,s(End,X,Y),Interp) :-
    fsa_visualization:abbreviate_label(Syms0,Sym0),
    escape_special(Sym0,Sym),
    gen_sym(Name),
    create_transition(Start,End,Sym,X0,Y0,X,Y,Name,Interp).

create_transition(Start,End,Sym,X0,Y0,X,Y,Name,Interp) :-
    (	X0/Y0 = X/Y
    ->	create_cycle(Start,End,Sym,X,Y,Name,Interp) % special case
    ;	compute_coords(X0,Y0,X,Y,Xa,Ya,Xb,Yb,Xc,Yc,Xtext,Ytext),
	tcl(Interp,'.t.c create line ~w ~w ~w ~w ~w ~w -arrow last -smooth 1 -tags {{~w ~w}} -arrowshape {25 25 3}',
		[Xa,Ya,Xb,Yb,Xc,Yc,Start,End],LineId), 
	tcl(Interp,'edge-label .t.c .t.c.~w ~w ~w "~w" {~w ~w}',
		[Name,Xtext,Ytext,Sym,Start,End],LabelId),
	assertz(current_trans(Start,End,Sym,X0,Y0,X,Y,
			      LineId,LabelId,Interp))
    ).

%% cycles are special transitions. Need more coordinates..
create_cycle(Start,End,Sym,X,Y,Name,Interp) :-
    compute_cycle_coords(X,Y,Xa,Ya,Xb,Yb,Xc,Yc,Xd,Yd,Xe,Ye,Xtext,Ytext),
    tcl(Interp,".t.c create line ~w ~w ~w ~w ~w ~w ~w ~w ~w ~w -arrow last -smooth 1 -tags {{~w ~w}} -arrowshape {25 25 5}",
	    [Xa,Ya,Xb,Yb,Xc,Yc,Xd,Yd,Xe,Ye,Start,End],LineId),
    tcl(Interp,'edge-label .t.c .t.c.~w ~w ~w "~w" {~w ~w} w',
	    [Name,Xtext,Ytext,Sym,Start,End],LabelId),
    assertz(current_trans(Start,End,Sym,X,Y,X,Y,LineId,LabelId,Interp)).

compute_cycle_coords(X,Y,X,Y,Xb,Yb,Xc,Yc,Xd,Yd,X,Y,Xtext,Ytext) :-
    Xb is X+30, Yb is Y-30,
    Xc is X+55, Yc is Y,
    Xd is X+30, Yd is Y+30,
    Xtext is Xc-10, Ytext is Yc.

%% computation of transition geometry
%% the idea is to have slanted lines between two states. The further
%% apart the states are, the more these lines leave the shortest path.
%% futhermore, arrows from left to right are above the shortest path,
%% and items from right to left are below.
compute_coords(X0,Y0,X,Y,X0,Y0,Xb,Yb,X,Y,Xtext,Ytext) :-
    ( X0 > X -> Xdir = left ;  Xdir = right ),
    ( Y0 > Y -> Ydir = up ;  Ydir = down ),
    MidX is (X0+X)/2,
    MidY is (Y0+Y)/2,
    Xdist is abs(X0-X),
    Ydist is abs(Y0-Y),
    Dist is sqrt(exp(Xdist,2)+exp(Ydist,2)),
    angle_factor(Factor),
    Schuifb is Factor*Dist,
    factors(Xdist,Ydist,Xfactor,Yfactor),
    add_or_subs(Ydir,MidX,Xfactor*Schuifb,Xb),
    add_or_subs(Ydir,MidX,Xfactor*Schuifb*0.5,Xtext),
    add_or_subs(Xdir,MidY,Yfactor*Schuifb,Yb),
    add_or_subs(Xdir,MidY,Yfactor*Schuifb*0.5,Ytext).

%% default angle
%% this determines how `angled' 
%% the transitions are. If 0 -> straight lines.
angle_factor(X) :-
    fsa_globals:fsa_global_get(v_angle,X0),
    (	atom(X0)
    ->	atom_codes(X0,Codes),
	number_codes(X,Codes)
    ;	X= X0
    ).

factors(Xd,Yd,Xf,Yf) :-
    (	Xd > Yd
    ->	(  Yd =:= 0
	->  Xf is 0, Yf is 1
	;   Xf is Yd/Xd,
	    Yf is 1-Xf
	)
    ;	(  Xd =:= 0 
	->  Yf is 0, Xf is 1
	;   Yf is Xd/Yd,
	    Xf is 1-Yf
	)
    ).

add_or_subs(left,Y,Dif,Yb) :-	Yb is Y+Dif.
add_or_subs(right,Y,Dif,Yb) :-	Yb is Y-Dif.
add_or_subs(down,Y,Dif,Yb) :-	Yb is Y+Dif.
add_or_subs(up,Y,Dif,Yb) :-	Yb is Y-Dif.

%%%%%%%%%%
%% SAVE %%
%%%%%%%%%%

%% saves current geometry to file
%% including fsa information in ordinary format..
save_fsa(File,Format):-
    format(user_error,"{Saving fsa in ~w (write=~w)}~n",[File,Format]),
    fsa_globals:fsa_global_get(fa,Fa),
    fsa_io:fsa_write_file(Format,File,Fa).

save_fsa(File) :-
    format(user_error,"{Saving fsa in ~w}~n",[File]),
    fsa_util:to_file(fsa_tk:save_fsa,File),
    format(user_error,"{Saved fsa in ~w}~n",[File]).

save_fsa :-
    fsa_globals:fsa_global_get(fa,Fa),
    fsa_io:fsa_write_file(Fa),
    (	fsa_globals:fsa_global_get(regex,Regex)
    ->	format(":- fsa_globals:fsa_global_set(regex,~k).~n",[Regex])
    ;	true
    ),
        format(":- dynamic current_state/6.~n",[]),
    (	current_state(A,B,C,D,E,F,_),
	format("current_state(~q,~q,~q,~q,~q,~q).~n",[A,B,C,D,E,F]),
	fail	
    ;	true	
    ),
    fsa_globals:fsa_global_get(v_angle,Angle),
    format(":- fsa_globals:fsa_global_set(v_angle,~w).~n",[Angle]).

%% display a fsa on the basis of a saved geometry:
recreate_fsa(Started) :-
    fsa_globals:fsa_global_get(fa,Fa),
    recreate_fa(Fa,Started).

recreate_fa(Fa,Started):-
    initialize_tk(Started,Interp),
    tcl(Interp,".t.c delete all",[]),
    (	current_state(_,_,_,_,_,_,Interp)
    ->	redraw_soft(Fa,Interp)
    ;   current_state(_,_,_,_,_,_)
    ->  add_int_to_current_state(Interp),
	redraw_soft(Fa,Interp)
    ;	tk_fsa(Fa,started(Interp))
    ).

add_int_to_current_state(Interp) :-
    (   retract(current_state(A,B,C,D,E,F)),
	assertz(current_state(A,B,C,D,E,F,Interp)),
	fail
    ;   true
    ).

redraw_soft(Fa,Interp):-
    (	retract(current_state(H,X,Y,_Id,Color,Relief,Interp)),
	create_state(H,X,Y,Color,Relief,Interp),
	fail
    ;	a_current_trans(Fa,Start,End,Sym,X0,Y0,X,Y,Interp),
	gen_sym(Name),
	create_transition(Start,End,Sym,X0,Y0,X,Y,Name,Interp),
	fail
    ;	true
    ),
    ensure_visible(Interp).

a_current_trans(Fa,Start,End,Sym,X0,Y0,X,Y,Interp) :-
    fsa_visualization:setof(S0,Fa^transition_or_jump(Fa,Start,S0,End),Syms),
    fsa_visualization:abbreviate_label(Syms,Sym0),
    escape_special(Sym0,Sym),
    current_state(Start,X0,Y0,_,_,_,Interp),
    current_state(End,X,Y,_,_,_,Interp).

move_node0(Node0,X,Y,Interp) :-
    move_node(Node0,X,Y,Interp).

move_node(Node,Xdiff,Ydiff,Interp) :-
    retract(current_state(Node,Xn,Yn,Id,Col,Rel,Interp)), !, 
    Xn1 is Xn+Xdiff, Yn1 is Yn+Ydiff,
    move_state(Node,Xn1,Yn1,Interp,Id,Col,Rel),
    %% nb. beware not to handle the transitions already moved..
    (	retract(current_trans(Node,End,Sym,X0,Y0,X,Y,LId,TId,Interp)),
	X1 is X0 + Xdiff, Y1 is Y0+Ydiff,
	move_trans(Node,End,Sym,X1,Y1,X,Y,Interp,LId,TId),
	fail
    ;	retract(current_trans(Start,Node,Sym,X0,Y0,X,Y,LId,TId,Interp)),
	X1 is X + Xdiff, Y1 is Y+Ydiff,
	move_trans(Start,Node,Sym,X0,Y0,X1,Y1,Interp,LId,TId),
	fail
    ;	true
    ).

move_trans(Start,End,Sym,X0,Y0,X,Y,Interp,Lid,Tid) :-
    (	X0/Y0 = X/Y		% cycle
    ->	move_cycle(Start,End,Sym,X,Y,Interp,Lid,Tid)
    ;	compute_coords(X0,Y0,X,Y,Xa,Ya,Xb,Yb,Xc,Yc,Xtext,Ytext),
	tcl(Interp,".t.c coords ~w ~w ~w ~w ~w ~w ~w",
		[Lid,Xa,Ya,Xb,Yb,Xc,Yc]),
	tcl(Interp,".t.c coords ~w ~w ~w",
		[Tid,Xtext,Ytext]),
	asserta(current_trans(Start,End,Sym,X0,Y0,X,Y,Lid,Tid,Interp))
    ).

move_cycle(Start,End,Sym,X,Y,Interp,Lid,Tid) :-
    compute_cycle_coords(X,Y,Xa,Ya,Xb,Yb,Xc,Yc,Xd,Yd,Xe,Ye,Xtext,Ytext),
    tcl(Interp,".t.c coords ~w ~w ~w ~w ~w ~w ~w ~w ~w ~w ~w",
	    [Lid,Xa,Ya,Xb,Yb,Xc,Yc,Xd,Yd,Xe,Ye]),
    tcl(Interp,".t.c coords ~w ~w ~w",
	    [Tid,Xtext,Ytext]),
    asserta(current_trans(Start,End,Sym,X,Y,X,Y,Lid,Tid,Interp)).

move_state(H,X,Y,Interp,Id,Col,Rel) :-
    tcl(Interp,".t.c coords ~w ~w ~w",[Id,X,Y]),
    asserta(current_state(H,X,Y,Id,Col,Rel,Interp)).

escape_special(Sym0,Sym) :-
    fsa_visualization:escape_specials(Sym0,Sym,
		[[92] -[92,92],
		 [123]-[92,123],
		 [125]-[92,125],
		 [91] -[92,91],
		 [93] -[92,93],
		 [34] -[92,34],
		 [36] -[92,36]]).

display_sigma :-
    fsa_globals:fsa_global_get(fa,Fa),
    fsa_data:fsa_symbols_decl(Fa,Sig),
    format(user_error,"~n%% sigma:~n",[]),
    format(user_error,"~w~n",[Sig]),
    format(user_error,"%% end of sigma~n",[]).

display_fa :-
    fsa_globals:fsa_global_get(fa,Fa),
    nl(user_error),
    fsa_io:fsa_write_file(user_error,Fa).

count_fa :-
    fsa_globals:fsa_global_get(fa,Fa),
    nl(user_error),
    fsa_data:fsa_count(Fa).

tk_version(Interp) :-
    fsa_globals:fsa_global_get(fsa_version,Version),
    tcl(Interp,'tk_dialog .dialog About {FSA utilities ~w All flames to vannoord@let.rug.nl} {} 0 {Seen enough?}',[Version]).

zoom(R,Interp) :-
    findall(current_state(State,X0,Y0,L,A,B,Interp),
	    retract(current_state(State,X0,Y0,L,A,B,Interp)),States0),
    zoom_states(States0,R),
    recreate_fsa(started(Interp)).

zoom_states([],_).
zoom_states([H|T],R):-
    zoom_state(H,R),
    zoom_states(T,R).

zoom_state(current_state(State,X0,Y0,L,A,B,Interp),R) :-
    X is X0*R,
    Y is Y0*R,
    assertz(current_state(State,X,Y,L,A,B,Interp)).

:- (   user:bb_get(fsa_tk:tk_id,_)
   ->  true
   ;   user:bb_put(fsa_tk:tk_id,0)
   ).

gen_sym(Id):-
    user:bb_get(fsa_tk:tk_id,Id),
    NewId is Id+1,
    user:bb_put(fsa_tk:tk_id,NewId).

%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% utilities %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%

tcl(Interp,Expr,Subs) :-
	%format(Expr,Subs),
    tcltk:tcl_eval('$TclInterp'(Interp),format(Expr,Subs),_).

tcl(Interp,Expr,Subs,Return) :-
    %format(Expr,Subs),
    tcltk:tcl_eval('$TclInterp'(Interp),format(Expr,Subs),Return0),
    atom_codes(Return,Return0).

% menu_flag(Interp,Flag,Values)
% adds a radiobutton for Flag under the Options menu, with Values
% an instantiated list of possible values

menu_flags([],_).
menu_flags([Flag|Flags],Interp):-
    (	fsa_globals:fsa_global_decl(Flag,_Text,Default,Vals,_)
    ->  menu_flag(Interp,Flag,Default,Vals)
    ;   true
    ),
    menu_flags(Flags,Interp).

menu_flag(Interp,Flag,Default,Values0) :-
    (   fsa_globals:fsa_global_get(Flag,Value)
    ->  true
    ;   Value=Default
    ),
    (	member(Value,Values0)
    ->	Values0=Values
    ;	Values=[Value|Values0]
    ),
    tcl(Interp,"set flag(~w) ~w",[Flag,Value]),
    menu_flag0(Interp,'.menu.options.m',Flag,Values).

menu_flag0(Interp,Menu,Flag,Values) :-
    tcl(Interp,'~w add cascade -label "~w" -menu ~w.~w',
	    [Menu,Flag,Menu,Flag]),
    tcl(Interp,'menu ~w.~w',[Menu,Flag],SubMenu),
    menu_flag_values(Values,Interp,Flag,SubMenu).

menu_flag_values([],Interp,Flag,Menu) :-
    tcl(Interp,'~w add command -label Help -command "display_help user global ~w"',
	    [Menu,Flag]).
menu_flag_values([V0|V],Interp,Flag,Menu) :-
    tcl(Interp,'~w add radiobutton -label ~w -value ~w -variable flag(~w) \
	      -command "prolog fsa_globals:fsa_global_set(~w,~w) "',
	      [Menu,V0,V0,Flag,Flag,V0]),
    menu_flag_values(V,Interp,Flag,Menu).


tk_close(Interp) :-
    retractall(current_trans(_,_,_,_,_,_,_,_,_,Interp)),
    retractall(current_state(_,_,_,_,_,_,Interp)),
    (	retract(tkconsol('$TclInterp'(Interp),In,Out,Err))
    ->  prolog_flag(user_input,_,In),
	prolog_flag(user_output,_,Out),
	prolog_flag(user_error,_,Err),
	set_input(In),
	set_output(Out)
    ;   true
    ),	
    tcltk:tcl_delete('$TclInterp'(Interp)).

protect(Call):-
    catch(fsa_tk:Call,Exc,user:print_message(error,Exc)).

expand_user_macros(Regex0,Interp):-
    expand_macros(user,Regex0,Interp).

expand_all_macros(Regex0,Interp):-
    expand_macros(all,Regex0,Interp).

expand_macros(Variant,Regex0,Interp):-
    fsa_regex:default_ops(Ops),
    fsa_util:call_with_ops(Ops,fsa_util:atom_term(Regex0,Regex1)),
    fsa_util:call_with_ops(Ops,fsa_tk:tcl(Interp,'next_regex {~q}',[Regex1])),
    expand(Regex1,Regex,Variant),
    fsa_globals:fsa_global_set(regex,Regex),
    fsa_util:call_with_ops(Ops,fsa_tk:tcl(Interp,'next_regex {~q}',[Regex])),
    fsa_util:call_with_ops(Ops,format(user_error,"~n~n~q~n~n",[Regex])).

expand(Expr0,Expr,Variant):-
    (	var(Expr0)
    ->	Expr0=Expr
    ;	fsa_regex_aux:current_predicate(protected,protected(_,_,_)),
	fsa_regex_aux:protected(Expr0,Expr,Variant)
    ->	true
    ;   protected(Expr0,Expr,Variant)
    ->  true
    ;	fsa_regex_aux:current_predicate(macro,macro(_,_)),
	fsa_regex_aux:macro(Expr0,Expr1)
    ->	expand(Expr1,Expr,Variant)
    ;	Variant==all,
	fsa_regex:macro(Expr0,Expr1)
    ->	expand(Expr1,Expr,Variant)
    ;   Expr0 =.. [Fun|Args0],
	length(Args0,N),
	length(Args,N),
	Expr =.. [Fun|Args],
	expand_args(Args0,Args,Variant)
    ).

expand_args([],[],_).
expand_args([H0|T0],[H|T],V):-
    expand(H0,H,V),
    expand_args(T0,T,V).

%% this should list operators which do something _special_ with
%% their arguments (where special means: not applying rx/2 to it;
%% i.e. it is a context in which macro expansion doesn't apply.
protected([H0|T0],[H|T],V):-
    !,
    expand_args([H0|T0],[H|T],V).   %% we don't want concat_list
protected('{}'(Conj0),'{}'(Conj),V):-
    !,
    expand(Conj0,Conj,V).

protected(A,A,_):-
    protected(A).

protected(file(_)).
protected(flag(_)).
protected(interval(_,_)).
protected(class(_)).
protected(negated_class(_)).
protected(fa(_)).
protected(escape(_)).
protected(dict(_)).
protected(word(_)).
protected(fsa_random(_,_,_,_)).
protected(det_random(_,_,_)).

% called from tcl
:- public fsa_minimum_path/0, produce_order_of_length/0,
    produce_stochastic_sample/0, submit_from_tk/2, showtrans/2,
    showtrans_in/1, showtrans_out/1.

%% minimum path
fsa_minimum_path :-
    fsa_globals:fsa_global_get(fa,Fa),
    fsa_minimum_path:fsa_minimum_path(Fa).

%% produce_sample
produce_order_of_length :-
    fsa_globals:fsa_global_get(fa,Fa),
    fsa_data:fsa_type(Fa,Type),
    fsa_interpreter:fsa_produces(Type,interpreted,Fa).

produce_stochastic_sample :-
    fsa_globals:fsa_global_get(fa,Fa),
    fsa_interpreter:fsa_sample(Fa).

%% submit from tk
submit_from_tk('Apply Exactly',W0) :-
    fsa_globals:fsa_global_get(fa,Fa),
    fsa_data:fsa_type(Fa,Type),
    fsa_interpreter:string_to_list_of_symbols(W0,W),
    fsa_interpreter:optp1(Type,interpreted,Fa,W).

submit_from_tk('Apply Approximately',W) :-
    fsa_globals:fsa_global_get(fa,Fa),
    fsa_interpreter:fsa_approx(Fa,W).

%% to be called regularly from within complex computations
%% tk_do_all_events :-
%%    (	  tk_do_one_event
%%    ->  tk_do_all_events
%%    ;   true
%%    ).

load_tcl_code(Interp) :-
    tcl(Interp,"
set entryfont -adobe-helvetica-bold-r-normal--14-140-*-*-*-*-iso8859-*

# history of regex and string
set current_cmd_no 0
set last_cmd_no 0
set string_current_cmd_no 0
set string_last_cmd_no 0

wm title . {Fsa Utilities; (c) Gertjan van Noord}
wm iconname . FSA
wm geometry . 1000x400
wm minsize . 100 100
frame .menu

menubutton .menu.file -text File -underline 0 -menu .menu.file.m
menubutton .menu.options -text Settings -underline 0 -menu .menu.options.m
menubutton .menu.op -text Operations -underline 0 -menu .menu.op.m
menubutton .menu.viz -text Vizualization -underline 0 -menu .menu.viz.m
menubutton .menu.pro -text Produce -underline 0 -menu .menu.pro.m
menubutton .menu.help -text Help -underline 0 -menu .menu.help.m

menu .menu.file.m
menu .menu.options.m
menu .menu.op.m
menu .menu.viz.m
menu .menu.pro.m
menu .menu.help.m

.menu.file.m add command -label Load -underline 0 \\
    -command {
	set filename [tk_getOpenFile]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_globals:fsa_global_set(regex,undefined))
	    prolog protect(fsa_tk:tk_fsa_file('$filename',started($interp)))
	}
    }
.menu.file.m add command -label LoadAux -underline 4\\
    -command {
	set filename [tk_getOpenFile]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_regex:fsa_load_aux_file('$filename'))
	}
    }

.menu.file.m add command -label ReconsultAux\\
    -command {
	set filename [tk_getOpenFile]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_regex:fsa_reconsult_aux_file('$filename'))
	}
    }

.menu.file.m add separator

.menu.file.m add command -label SaveAs -underline 0 \\
    -command {
	set filename [tk_getSaveFile]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename'))
	}
    }

.menu.file.m add command -label \"SaveAsPsTricks Picture\" \\
    -command {
	set filename [tk_getSaveFile -defaultextension .pic]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',pstricks))
	}
    }

.menu.file.m add command -label \"SaveAsPsTricks LaTeX Document\" \\
    -command {
	set filename [tk_getSaveFile -defaultextension .tex]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',pstricks_tex))
	}
    }

",[]),    tcl(Interp,"
.menu.file.m add command -label SaveAsPostScript\\
    -command {
	set filename [tk_getSaveFile -defaultextension .ps]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',postscript))
	}
    }

.menu.file.m add command -label SaveAsFast\\
    -command {
	set filename [tk_getSaveFile -defaultextension .fast]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',fast))
	}
    }

.menu.file.m add command -label SaveAsNormal\\
    -command {
	set filename [tk_getSaveFile]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',normal))
	}
    }


.menu.file.m add command -label SaveAsOld\\
    -command {
	set filename [tk_getSaveFile]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',old))
	}
    }

.menu.file.m add command -label SaveAsCompact\\
    -command {
	set filename [tk_getSaveFile]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',compact))
	}
    }

.menu.file.m add command -label SaveAsVcg\\
    -command {
	set filename [tk_getSaveFile -defaultextension .vcg]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',vcg))
	}
    }

.menu.file.m add command -label SaveAsDaVinci\\
    -command {
	set filename [tk_getSaveFile -defaultextension .dav]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',davinci))
	}
    }

.menu.file.m add command -label SaveAsDot\\
    -command {
	set filename [tk_getSaveFile -defaultextension .dot]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',dot))
	}
    }

.menu.file.m add command -label SaveAsProlog\\
    -command {
	set filename [tk_getSaveFile -defaultextension .pl]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',prolog))
	}
    }

.menu.file.m add command -label {SaveAsC}\\
    -command {
	set filename [tk_getSaveFile -defaultextension .c]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',c))
	}
    }

.menu.file.m add command -label {SaveAsJava}\\
    -command {
	set filename [tk_getSaveFile -defaultextension .java]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',java))
	}
    }

.menu.file.m add command -label {SaveAsC++}\\
    -command {
	set filename [tk_getSaveFile -defaultextension .cc]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',cpp))
	}
    }

",[]),    tcl(Interp,"
.menu.file.m add command -label SaveAsFSM\\
    -command {
	set filename [tk_getSaveFile -defaultextension .fsm]
	if {\"$filename\" != \"\"} {
	    prolog protect(fsa_tk:save_fsa('$filename',fsm))
	}
    }

.menu.file.m add separator

.menu.file.m add command -label Revert -underline 0\\
	-command { prolog protect(fsa_tk:revert_fsa(started($interp))) }

.menu.file.m add command -label Redraw -underline 2\\
	-command { prolog protect(fsa_tk:recreate_fsa(started($interp))) }

.menu.file.m add separator

.menu.file.m add command -label Close -underline 0\\
    -command { prolog fsa_tk:tk_close($interp) }

.menu.file.m add command -label Quit -underline 0\\
	-command exit

# .menu.options is filled in from the Prolog side

foreach op { 
    accessible
    cleanup
    coaccessible
    complement
    complete
    containment
    del
    determinize 
    dict
    domain
    efree
    reachable_efree
    co_reachable_efree
    expand_non_overlapping_predicates
    kleene_star
    kleene_plus
    identity
    ins
    inverse
    letter_transducer
    minimize
    mb
    mh
    no_weights
    option
    perfect_hash
    range
    reverse
    subs
    t_determinize
    t_minimize
    w_determinize
    w_minimize
    weights
    weighted_letter_transducer
    wt_determinize
    wt_minimize
           } {
    .menu.op.m add command -label $op -command \"
    prolog \\\"protect(fsa_tk:tk_operation($op,started($interp)))\\\" \"
#    prolog \\\"tk_regex('$op\\(\\$regex\\)',started('.t.c'))\\\" \"
}
",[]),    tcl(Interp,"

foreach viz {
    davinci
    dotty
    dot_ghostview
    dot_xv
    pstricks_ghostview
    vcg
} { .menu.viz.m add command -label $viz \\
	-command \"prolog protect(fsa_visualization:fsa_visualization($viz))\"
}

.menu.viz.m add command -label Internal -command \"prolog display_fa\"
.menu.viz.m add command -label {Tk (ignoring no_display_beyond)} -command \\
              \"prolog tk_fsa_anyway(started($interp))\"

.menu.pro.m add command -label {Produce in order of length}\\
    -command \"prolog protect(fsa_tk:produce_order_of_length)\"
.menu.pro.m add command -label {Stochastic Sample}\\
    -command \"prolog protect(fsa_tk:produce_stochastic_sample)\"
.menu.pro.m add command -label {Minimum path}\\
    -command \"prolog protect(fsa_tk:fsa_minimum_path)\"

.menu.help.m add command -label About -command {
    prolog fsa_tk:tk_version($interp)
}
.menu.help.m add command -label {Reference Manual in netscape}\\
	-command {
         send_url_to_netscape http://www.let.rug.nl/~~vannoord/Fsa/Manual
                 }

.menu.help.m add command -label {FSA Home Page}\\
	-command {
         send_url_to_netscape http://www.let.rug.nl/~~vannoord/Fsa/
                 }

pack .menu.file .menu.options .menu.op .menu.pro .menu.viz -side left
pack .menu.help -side right


frame .edits
label .edits.al -text {Edge Angle:}
entry .edits.a -textvariable fsa(angle) -relief sunken -width 5\\
        -font $entryfont
bind .edits.a <Return> \"
   prolog \\\"protect(fsa_globals:fsa_global_set(v_angle,\\$fsa(angle)))\\\"
   prolog \\\"protect(fsa_tk:recreate_fsa(started($interp)))\\\" \"
label .edits.xl -text {X-distance:}
entry .edits.x -textvariable fsa(xdist) -relief sunken -width 5\\
        -font $entryfont
bind .edits.x <Return> \"
	prolog protect(fsa_globals:fsa_global_set(v_xdist,\\$fsa(xdist)))
	prolog \\\"protect(fsa_tk:revert_fsa(started($interp)))\\\"\"

button .edits.sigma -text {Display Sigma} \\
         -command \"prolog protect(fsa_tk:display_sigma)\"

button .edits.fa -text {Display Fa} -command \"prolog protect(fsa_tk:display_fa)\"

button .edits.count -text {Count Fa} -command \"prolog protect(fsa_tk:count_fa)\"

button .edits.cache -text {Clear Cache} -command \"prolog fsa_regex:verbose_clear_cache\"

button .edits.zoomin -text {Zoom-out} -command \"prolog fsa_tk:zoom(0.8,$interp)\"

button .edits.zoomout -text {Zoom-in} -command \"prolog fsa_tk:zoom(1.125,$interp)\"
button .edits.interp -text {Interp} -command {prolog \"format('~~w~~n',\\[$interp\\])\"}

pack .edits.al .edits.a .edits.xl .edits.x \\
    .edits.sigma .edits.fa .edits.count .edits.cache \\
    .edits.zoomin .edits.zoomout .edits.interp -side left -fill x
    
frame .regexf
frame .stringf
frame .t
",[]),    tcl(Interp,"

label .regexf.regexl -text \"Regex:\"
entry .regexf.regex -textvariable regex -relief sunken -font $entryfont
bind .regexf.regex <Return> {
    regsub -all ' $regex \"\\\\\\\\'\" regex
    prolog \"protect(fsa_tk:tk_regex('$regex',started($interp)))\"
}
bind .regexf.regex <Up> {
    if [expr $current_cmd_no > 1] {
	set current_cmd_no [expr $current_cmd_no-1]
	set regex $regex_history($current_cmd_no);
    }
}
bind .regexf.regex <Down> {
    if [expr $current_cmd_no < $last_cmd_no] {
	set current_cmd_no [expr $current_cmd_no+1]
	set regex $regex_history($current_cmd_no)
    }
}
bind .regexf.regex <3> \"set regex {}\"

button .regexf.expanduser -text {Expand User Macro's}\\
    -command {prolog \"protect(fsa_tk:expand_user_macros('$regex',$interp))\"}

button .regexf.expandall -text {Expand All Macro's}\\
    -command {prolog \"protect(fsa_tk:expand_all_macros('$regex',$interp))\"}

pack .regexf.regexl -side left
pack .regexf.regex -side left -expand 1 -fill x
pack .regexf.expanduser .regexf.expandall -side left

label .stringf.stringl -text \"String:\"
entry .stringf.string -textvariable string -relief sunken -width 60\\
      -font $entryfont
bind .stringf.string <Return> {
    set string_last_cmd_no [expr $string_last_cmd_no+1]
    set string_current_cmd_no $string_last_cmd_no
    set string_history($string_last_cmd_no) $string
    prolog protect(fsa_tk:submit_from_tk(\\'$action\\',\\\"$string\\\"))
}
bind .stringf.string <Up> {
    if [expr $string_current_cmd_no > 1] {
	set string_current_cmd_no [expr $string_current_cmd_no-1]
	set string $string_history($string_current_cmd_no);
    }
}
bind .stringf.string <Down> {
    if [expr $string_current_cmd_no < $string_last_cmd_no] {
	set string_current_cmd_no [expr $string_current_cmd_no+1]
	set string $string_history($string_current_cmd_no)
    }
}
bind .stringf.string <3> \"set string {}\"

set action {Apply Exactly}
menubutton .stringf.actiontype -menu .stringf.actiontype.menu -textvariable action -relief raised
menu .stringf.actiontype.menu 

.stringf.actiontype.menu add radiobutton -label {Apply Exactly}\\
    -variable action
.stringf.actiontype.menu add radiobutton -label {Apply Approximately}\\
    -variable action

",[]),    tcl(Interp,"
button .stringf.submit -text {Submit} -command {
    set string_last_cmd_no [expr $string_last_cmd_no+1]
    set string_current_cmd_no $string_last_cmd_no
    set string_history($string_last_cmd_no) $string
    prolog protect(fsa_tk:submit_from_tk(\\'$action\\',\\\"$string\\\"))
}



pack .stringf.stringl -side left 
pack .stringf.string -side left -expand 1 -fill x
pack .stringf.submit .stringf.actiontype -side left
##pack .stringf.accept .stringf.transduce .stringf.w_transduce .stringf.approx -side left

canvas .t.c -bd 2 -yscrollcommand \".t.y set\" \\
	-xscrollcommand \".t.x set\" \\
	-background White\\
	-scrollregion \"-5000 -5000 5000 5000\" -width 0 -height 0

scrollbar .t.x -orient horiz -command \".t.c xview\"
scrollbar .t.y -command \".t.c yview\"
pack .t.x -side bottom -fill x
pack .t.y -side left -fill y
pack .t.c -fill both -expand 1

pack .menu -side top -fill x
pack .regexf -fill x
pack .stringf -fill x
pack .t -fill both -expand 1
pack .edits -fill x -expand 0


proc next_regex {lregex} {
    global regex_history regex current_cmd_no last_cmd_no
    set regex $lregex
    set last_cmd_no [expr $last_cmd_no+1]
    set current_cmd_no $last_cmd_no
    set regex_history($last_cmd_no) $lregex

}

bind .t.c <1> \"tk_popup .menu.file.m %X %Y\"
bind .t.c <2> \"tk_popup .menu.op.m %X %Y\"
bind .t.c <3> \"tk_popup .menu.viz.m %X %Y\"

proc move_node {state x y} {
    global interp
     prolog \"protect(fsa_tk:move_node0($state,$x,$y,$interp))\"
}
",[]),    tcl(Interp,"

proc edge-label {canvas name x y text tag {anchor center}} {
    label $name -text $text -relief raised \\
   -font \"-*-times-bold-r-*-*-*-120-*-*-*-*-*-*\"
   bind $name <Enter> \"myraise $name $canvas \\{$tag\\}\"
   bind $name <1> \"showtrans \\{$tag\\}\"
   $canvas create window $x $y -window $name -anchor $anchor
}

proc showtrans {tag} {
    prolog fsa_tk:showtrans([lindex $tag 0],[lindex $tag 1])
}

proc myraise {w canvas tag} {
    raise $w
    $canvas itemconfigure $tag -fill red -width 3
    after 1000 \"$canvas itemconfigure \\{$tag\\} -fill black -width 1\" 
}


proc send_url_to_netscape {url} {
    if [catch { exec netscape -remote openURL($url) } ] {
	exec netscape $url &
    }
}


",[]).


showtrans_out(I) :-
    fsa_globals:fsa_global_get(fa,Fa),
    format(user_error,"~ntransitions leaving ~w:~n",[I]),
    (	fsa_data:fsa_transition(Fa,I,Sym,J),
	format(user_error,"~w -> ~w: ~w~n",[I,J,Sym]),
	fail
    ;	fsa_data:fsa_jump(Fa,I,J),
	format(user_error,"~w -> ~w: <jump>~n",[I,J]),
	fail
    ;	true
    ).

showtrans_in(I) :-
    fsa_globals:fsa_global_get(fa,Fa),
    format(user_error,"~ntransitions entering ~w:~n",[I]),
    (	fsa_data:fsa_transition(Fa,J,Sym,I),
	format(user_error,"~w -> ~w: ~w~n",[J,I,Sym]),
	fail
    ;	fsa_data:fsa_jump(Fa,J,I),
	format(user_error,"~w -> ~w: <jump>~n",[J,I]),
	fail
    ;	true
    ).

showtrans(I,J) :-
    fsa_globals:fsa_global_get(fa,Fa),
    format(user_error,"~ntransitions from ~w to ~w:~n",[I,J]),
    (	fsa_data:fsa_transition(Fa,I,Sym,J),
	format(user_error,"~w~n",[Sym]),
	fail
    ;	fsa_data:fsa_jump(Fa,I,J),
	format(user_error,"<jump>~n",[]),
	fail
    ;	true
    ).

:- public module_current_state/8.  % fsa_visualization
module_current_state(Module,A,B,C,D,E,F,G) :-
    Module:current_state(A,B,C,D,E,F,G).
