%%% -*- Module: fsa_compiler_to_c; -*-
:- module(fsa_compiler_to_c,
	  [ fsa_compile_to_c/1,
	    fsa_compile_to_c/2,
	    fsa_compile_to_c_fa/2
	  ]).

#swi# :- style_check([-atom]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.

user:help_info(class,fsa_compiler_to_c,"C Code Generation","
FSA supports the production of C code on the basis of
a finite automaton. Before the automaton is translated into C code, 
it is determinized. In the case of transducers, Mohri's determinization
algorithm is applied. Note however that certain transductions cannot be
determinized: in that case the algorithm will not terminate. 

The C program will contain definitions of the following functions:

* recognizers: 

        int accepts(char *in)

* weighted recognizers

        int w_accepts(char *in,int *out)

* transducers

        int t_accepts(char *in,char *out)

* weighted transducers

        int wt_accepts(char *in,char *out, int *wout)
	
In each case, the functions return 1 if the string *in* is
accepted. Otherwise they return 0. For transducers the resulting
string or weight is available in *out* and *wout*.

If the global variable *c_with_main* is set to *on*, then the
resulting C program will also contain a main function. This function
is defined in such a way that it reads lines from standard input and
applies the corresponding accept functions for each line. For
recognizers, either *yes* or *no* is printed to standard error. For
transducers, the transduction is written to standard output; if the
input string is not in the domain of the transducer then *no* is
written to standard error. 

The representation of the finite-automaton in C is similar to the
technique explained on page 43 (table 4.2) of Jan Daciuk's
dissertation `Incremental Construction of Finite-State Automata and
Transducers and their use in the Natural Language Processing'.
Politechnika Gdanska, 1998.

The special input symbol ^A is used in the representation of the automaton
in C to indicate a symbol not otherwise mentioned in the automaton: it will
match any such symbol. Similarly, in transducers the symbol ^B is used
to indicate an unknown symbol with an associated identity. The corresponding
output symbol is also ^B. When a string is transduced this ^B is replaced
by the actual input symbol (by means of a queue).  An unknown
output symbol without an associated identity is represented using the
symbol given by the global flag fl_arbitrary_symbol.  In the case of
final states with multiple outputs a special meta-notation is used using
a special symbol given by the global variable fl_multiple_symbol_start
which starts a sequence of possible outputs where each output is seperated
using a symbol given by the global variable fl_multiple_symbol_sep.



").


user:help_info(xpred,fsa_compile_to_c,
"fsa_compile_to_c(+Fa) fsa_compile_to_c(+FileIn,+FileOut)","
In the first variant, a C program is written to standard output
for the automaton Fa. The C program will read lines from standard
input and report for each line whether it is a string accepted by
Fa. In the second variant, the C program is written to FileOut on
the basis of the automaton read in from FileIn.").

user:help_info(xpred,fsa_compile_to_c_fa,
"fsa_compile_to_c_fa(+Fa,+FileOut)",
"A C program is written to FileOut for the automaton Fa. The C program 
will read lines from standard input and report for each line whether 
it is a string accepted by Fa. ").


fsa_compile_to_c(Source,Target):-
    fsa_fl:fsa_fl_file(c,Source,Target).

fsa_compile_to_c(Fa):-
    fsa_fl:fsa_fl_fa(c,Fa,user).

fsa_compile_to_c_fa(Fa,Target):-
    fsa_fl:fsa_fl_fa(c,Fa,Target).
