:- module( fsa_fl, [ fsa_fl_fa/3,
		     fsa_fl_fa/4,
		     fsa_fl_file/3,
		     fsa_fl_file/4,
		     fsa_fl/4
		   ]).

%% the C++ part as suggested by Dale Gerdemann (although I ignored a
%% few suggestions for improvement ...)

#swi# :- style_check([-atom]).
#!swi# :- use_module(library(lists), [ select/3, member/2 ]).

fsa_fl_file(Lang,Type,Source,Target) :-
    fsa_io:fsa_read_file(Source,Fa),
    fsa_fl_fa(Lang,Type,Fa,Target).

fsa_fl_file(Lang,Source,Target) :-
    fsa_io:fsa_read_file(Source,Fa),
    fsa_fl_fa(Lang,Fa,Target).

fsa_fl_fa(Lang,Fa,Target) :-
    fsa_data:fsa_type(Fa,Type),
    fsa_fl_fa(Lang,Type,Fa,Target).

fsa_fl_fa(Lang,Type,Fa,Target0) :-
    fsa_fl_parse_file(Lang,Type,Target0,Target,Label),
    fsa_util:to_file(fsa_fl:fsa_fl(Lang,Type,Label,Fa),Target).

fsa_fl_parse_file(c,_,Path,Path,Path).
fsa_fl_parse_file(cpp,_,Path,Path,Path).
fsa_fl_parse_file(java,_,Path,File,Label) :-
    (	Path == user
    ->	File=user, Label=user
    ;	(   atom_concat(Path0,'.java',Path)
	->  File=Path
	;   atom_concat(Path,'.java',File),
	    Path0=Path
	),
	core_name(Path0,Label)
    ).
		 
%% core_name
core_name(Path,Name) :-
    atom_codes(Path,Codes),
    core_name_(Codes,T,T,NCodes),
    atom_codes(Name,NCodes).

core_name_([],[],Sol,Sol).
core_name_([H|T],Sol0,Sol,N):-
    (	H =:= 47
    ->	core_name_(T,S,S,N)
    ;   Sol0=[H|Sol1],
	core_name_(T,Sol1,Sol,N)
    ).

fsa_fl(Lang,Type,Label,Fa0) :-
    get_main(Lang,Main),
    get_conversion(Lang,Converse),
    converse_fa(Converse,Type,Fa0,Fa1,Specials),
    select_epsilon_start(Type,Fa1,Fa2,Ini),
    with_final_output(Type,Fa2,Fa,FinalOutput,FinalTable),
    calc_nr_next(Fa,NextTable),
    fsa_data:fsa_start_states(Fa,[Start]),
    fsa_data:fsa_transitions(Fa,T),
    length(T,Length0), Length is Length0+1,
    format_comment,
    format_headers(Lang),
    format_queue(Type,Lang),
    format_init(Lang,Type,Length,Label,Ini,FinalOutput,
		FinalTable,NextTable,Specials),
    format_sink(Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable),
    format_start(Lang,Type,Label,Start,Ini,FinalOutput,FinalTable,NextTable),
    format_transitions(T,Lang,Type,FinalOutput,FinalTable,NextTable,Fa),
    format_transitions_end(Lang),
    format_more(Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable),
    format_main(Main,Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable),
    format_tail(Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable).


get_main(java,Main) :-
    fsa_globals:fsa_global_get(java_with_main,Main).
get_main(c,Main) :-
    fsa_globals:fsa_global_get(c_with_main,Main).
get_main(cpp,Main) :-
    fsa_globals:fsa_global_get(cpp_with_main,Main).

get_conversion(java,Converse) :-
    fsa_globals:fsa_global_get(to_java_conversion,Converse).
get_conversion(c,Converse) :-
    fsa_globals:fsa_global_get(to_c_conversion,Converse).
get_conversion(cpp,Converse) :-
    fsa_globals:fsa_global_get(to_cpp_conversion,Converse).

converse_fa(off,_,Fa,Fa,[]).
converse_fa(on,Type,Fa0,Fa,Syms3):-
    symbols_talked_about(Fa0,Syms0),
    unknown_symbol(Special),
    n_expand_predicates_fa(Type,Fa0,Fa1,[Special|Syms0]),
    converse_fa(Type,Fa1,Fa),
    symbols_talked_about(Fa1,Syms2),
    sort(Syms2,Syms3).

atom_to_atoms(Atom,Atoms,Tail) :-
    fsa_globals:fsa_global_get(fl_ignore_complex_symbols,Ignore),
    atom_to_atoms(Ignore,Atom,Atoms,Tail).

atom_to_atoms(on,Atom,[Atom|Tail],Tail) :-
    atom_codes(Atom,[_]).
atom_to_atoms(on,Atom,Tail,Tail) :-
    atom_codes(Atom,[_,_|_]).
atom_to_atoms(off,Atom,Atoms,Tail) :-
    atom_codes(Atom,String),
    string_to_atoms(String,Atoms,Tail).

string_to_atoms([],Tail,Tail).
string_to_atoms([Int|Ints],[Atom|Atoms0],Atoms):-
    atom_codes(Atom,[Int]),
    string_to_atoms(Ints,Atoms0,Atoms).

converse_fa(recognizer,Fa0,Fa):-
    fsa_regex:fsa_regex_compile(determinize(fa(Fa0)),Fa).
converse_fa(w_recognizer,Fa0,Fa):-
    fsa_regex:fsa_regex_compile(w_determinize(fa(Fa0)),Fa).
converse_fa(transducer(_),Fa0,Fa):-
    fsa_regex:fsa_regex_compile(t_determinize(fa(Fa0)),Fa).
converse_fa(w_transducer(_),Fa0,Fa):-
    fsa_regex:fsa_regex_compile(wt_determinize(fa(Fa0)),Fa).

select_epsilon_start(recognizer,Fa,Fa,"\"\"").
select_epsilon_start(transducer(_),Fa0,Fa,IniOut):-
    fsa_data:fsa_copy_except(start_states,Fa0,Fa1,[Start0],[Start]),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    (	select(trans(Start0,[]/Out,Start),Trans0,Trans),
	\+ member(trans(Start0,_,_),Trans)
    ->	true
    ;   Fa0=Fa,
	Out=[]
    ),
    string_symbol(Out,IniOut).

select_epsilon_start(w_recognizer,Fa0,Fa,IniOut):-
    fsa_data:fsa_copy_except(start_states,Fa0,Fa1,[Start0],[Start]),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    (	select(trans(Start0,[]/Out,Start),Trans0,Trans),
	\+ member(trans(Start0,_,_),Trans)
    ->	true
    ;   Fa0=Fa, Out=0
    ),
    number_codes(Out,IniOut).

select_epsilon_start(w_transducer(_),Fa0,Fa,S/W):-
    fsa_data:fsa_copy_except(start_states,Fa0,Fa1,[Start0],[Start]),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    (	select(trans(Start0,[]/(Out/OutW),Start),Trans0,Trans),
	\+ member(trans(Start0,_,_),Trans)
    ->	true
    ;   Fa0=Fa, OutW=0, Out=[]
    ),
    number_codes(OutW,W),
    string_symbol(Out,S).


%% for each state, find out the position in the transition list.
calc_nr_next(Fa,Table):-
    fsa_data:fsa_states_number(Fa,Size),
    fsa_array:fsa_array_new(Table,Size),
    fsa_data:fsa_transitions(Fa,Trs),
    calc_nr_next0(Trs,Table).

calc_nr_next0([],_).
calc_nr_next0([trans(P,_,_)|Tail],Table):-
    calc_nr_next(Tail,P,2,2,Table).

calc_nr_next([],Prev,NrSoFar,_TransNr,Table):-
    fsa_array:fsa_array_access(Prev,NrSoFar,Table).
calc_nr_next([trans(P,_,_)|Tail],Prev,NrSoFar,TransNr,Table):-
    TransNr1 is TransNr+1,
    (	P==Prev
    ->	calc_nr_next(Tail,Prev,NrSoFar,TransNr1,Table)
    ;   fsa_array:fsa_array_access(Prev,NrSoFar,Table),
	calc_nr_next(Tail,P,TransNr1,TransNr1,Table)
    ).

with_final_output(recognizer,Fa,Fa,[],Table) :-
    fsa_data:index_finals(Fa,Table).

with_final_output(w_recognizer,Fa,Fa,[],Table) :-
    fsa_data:index_finals(Fa,Table).

with_final_output(transducer(_),Fa0,Fa,FinalOutputs,FTable):-
    fsa_data:fsa_states_number(Fa0,Size),
    fsa_array:fsa_array_new(FinalOutputs,Size),
    fsa_data:fsa_copy_except(transitions,Fa0,Fa1,Trans0,Trans),
    fsa_data:fsa_copy_except(final_states,Fa1,Fa,Finals0,Finals),
    with_final_output0(Trans0,Trans,Finals1,Finals0,FinalOutputs,t),
    sort(Finals1,Finals),
    fsa_data:index_finals(Fa,FTable).

with_final_output(w_transducer(_),Fa0,Fa,FinalOutputs,FTable):-
    fsa_data:fsa_states_number(Fa0,Size),
    fsa_array:fsa_array_new(FinalOutputs,Size),
    fsa_data:fsa_copy_except(transitions,Fa0,Fa1,Trans0,Trans),
    fsa_data:fsa_copy_except(final_states,Fa1,Fa,Finals0,Finals),
    with_final_output0(Trans0,Trans,Finals1,Finals0,FinalOutputs,w),
    sort(Finals1,Finals),
    fsa_data:index_finals(Fa,FTable).

with_final_output1([],[],F,F,[],_).
with_final_output1([trans(P,L/R,Q)|TransT],Trans,F0,F,Outs,TW):-
    (	L == []
    ->	Trans1=Trans,
	F0=[P|F1],
	Outs=[P-R|Outs1]
    ;	F0=F1,
	Trans=[trans(P,L/R,Q)|Trans1],
	Outs=Outs1
    ),
    with_final_output1(TransT,Trans1,F1,F,Outs1,TW).

with_final_output0(Tr0,Tr,F0,F,OutputsArray,TW) :-
    with_final_output1(Tr0,Tr,F0,F,Outputs0,TW),
    sort(Outputs0,Outputs),
    combine_outputs(Outputs,OutputsArray,TW).

combine_outputs([],_,_).
combine_outputs([P-Out|T],Ar,TW) :-
    combine_outputs(T,P,[Out],Ar,TW).

combine_outputs([],P,Outs,Ar,TW) :-
    fsa_array:fsa_array_access(P,Out,Ar),
    combine_output_seq(TW,Outs,Out).
combine_outputs([P1-Out1|T],P,Outs0,Ar,TW) :-
    (	P==P1
    ->	combine_outputs(T,P,[Out1|Outs0],Ar,TW)
    ;	fsa_array:fsa_array_access(P,Out,Ar),
	combine_output_seq(TW,Outs0,Out),
	combine_outputs(T,P1,[Out1],Ar,TW)
    ).

combine_output_seq(t,List,Out) :-
    combine_output_seq_t(List,Out).
combine_output_seq(w,List,Out/Score) :-
    get_scores(List,List1,Score),
    combine_output_seq_t(List1,Out).

get_scores([],[],_).
get_scores([H/Score|T],[H|NT],Score) :-
    get_scores(T,NT,Score).

combine_output_seq_t([H|T],Out):-
    combine_output_seq_t(T,H,Out).

combine_output_seq_t([],H,[Sym]) :-
    fsa_util:concat_all(H,Sym).
combine_output_seq_t([H|T],F,[Sym]):-
    flatten1([F,H|T],List),
    fsa_globals:fsa_global_get(fl_multiple_symbol_start,Start),
    fsa_globals:fsa_global_get(fl_multiple_symbol_separator,Sep),
    fsa_util:concat_all(List,Sym0,Sep),
    atom_concat(Start,Sym0,Sym).

flatten1([],[]).
flatten1([H|T],[Atom|L]):-
    fsa_util:concat_all(H,Atom),
    flatten1(T,L).

get_single_symbol(Module,Pred,L,Sym) :-
    findall(L0,fsa_pred_module:evaluate_predicate(Module,Pred,L0),Ls),
    check_get_single_symbol(Ls,Sym,Module,Pred),
    (	atom(Sym),
	atom_codes(Sym,[L])
    ->	true
    ;	number(Sym),
	number_codes(Sym,[L])
    ).

check_get_single_symbol([],_,Mod,P) :-
    user:throw(fsa_error("~w:~w is not true of a single symbol~n",[Mod,P])).

check_get_single_symbol([H|T],L,Module,Pred) :-
    check_get_single_symbol(T,H,L,Module,Pred).

check_get_single_symbol([],L,L,_,_).
check_get_single_symbol([_|_],_,_,Mod,P) :-
    user:throw(fsa_error("~w:~w is true of more than one symbol~n",[Mod,P])).
    
symbol_mark(recognizer,Fa,Pred,L,[]):-
    fsa_data:fsa_symbols_decl(Fa,r(Module)),
    get_single_symbol(Module,Pred,L,_).

symbol_mark(transducer(_),Fa,'$@'(Pred)/['$@'(Pred)],L,R):-
    fsa_data:fsa_symbols_decl(Fa,t(Module,Module)),
    !,
    get_single_symbol(Module,Pred,L,Sym),
    (	integer(Sym)
    ->  integer_symbol(Sym,R)
    ;   string_symbol([Sym],R)
    ).

symbol_mark(transducer(_),Fa,PredL/PredR,L,R):-
    fsa_data:fsa_symbols_decl(Fa,t(ModuleL,ModuleR)),
    fsa_pred_module:evaluate_predicate(ModuleL,PredL,SymL),
    get_single_symbol(ModuleL,PredL,L,SymL),
    evaluate_predicate_list(PredR,SymR,ModuleR),
    (	integer(SymR)
    ->  integer_symbol(SymR,R)
    ;   string_symbol(SymR,R)
    ).

symbol_mark(w_transducer(_),Fa,'$@'(Pred)/(['$@'(Pred)]/SymR),L,Sr/R):-
    fsa_data:fsa_symbols_decl(Fa,t(Module,Module,fsa_frozen)),
    !,
    get_single_symbol(Module,Pred,L,Sym),
    (	integer(Sym)
    ->  integer_symbol(Sym,Sr)
    ;   string_symbol([Sym],Sr)
    ),
    integer_symbol(SymR,R).


symbol_mark(w_transducer(_),Fa,PredL/(PredR/SymR),L,Sr/R):-
    fsa_data:fsa_symbols_decl(Fa,t(ModuleL,ModuleR,fsa_frozen)),
    fsa_pred_module:evaluate_predicate(ModuleL,PredL,SymL),
    get_single_symbol(ModuleL,PredL,L,SymL),
    evaluate_predicate_list(PredR,NrR,ModuleR),
    (	integer(NrR)
    ->  integer_symbol(NrR,Sr)
    ;   string_symbol(NrR,Sr)
    ),
    integer_symbol(SymR,R).

symbol_mark(w_recognizer,Fa,PredL/SymR,L,R):-
    fsa_data:fsa_symbols_decl(Fa,r(ModuleL,fsa_frozen)),
    get_single_symbol(ModuleL,PredL,L,_SymL),
    integer_symbol(SymR,R).

evaluate_predicate_list([],[],_).
evaluate_predicate_list([Pred|Preds],[Sym|Syms],Module) :-
    fsa_pred_module:evaluate_predicate(Module,Pred,Sym),
    evaluate_predicate_list(Preds,Syms,Module).

integer_symbol(Int,IntChars):-
    number_codes(Int,IntChars).

string_symbol(List,[34|Chars]) :-
    string_symbol0(List,Chars,[34]).

string_symbol0([],S,S).
string_symbol0([H|T],S0,S) :-
    escape_symbol(H,S0,S1),
    string_symbol0(T,S1,S).

escape_symbol(Sym,C0,C) :-
    (	number(Sym)
    ->	number_codes(Sym,[Code|Codes])
    ;   atom(Sym)
    ->	atom_codes(Sym,[Code|Codes])
    ;   format(user_error,"Warning: output symbol is not atomic: ~q~n",[Sym]),
	fsa_util:term_atom(Sym, SymAtom),
	atom_codes(SymAtom,[Code|Codes])
    ),
    escape_symbol1(Code,C0,C1),
    escape_symbol_(Codes,C1,C).

escape_symbol_([],C,C).
escape_symbol_([H|T],C0,C) :-
    escape_symbol1(H,C0,C1),
    escape_symbol_(T,C1,C).

escape_symbol1(Sym,C0,C) :-
    (	Sym =:= 92
    ->	C0=[92,92|C]
    ;   Sym =:= 34
    ->  C0 = [92,34|C]
    ;   Sym =:= 10   % newlines need to be escaped these days too!
    ->  C0 = [92,110|C]
    ;   C0=[Sym|C]
    ).

format_comment :-
    format(
"/* This file is generated automatically by FSA Utilities.
 * (c) 1993 - 2000 Gertjan van Noord RUG
 * More information from http://www.let.rug.nl/~~vannoord/Fsa/
 * 
 * The representation of a finite-automaton is
 * similar to the technique explained on page 43 (table
 * 4.2) of Jan Daciuk's dissertation `Incremental
 * Construction of Finite-State Automata and Transducers
 * and their use in the Natural Language Processing'.
 * Politechnika Gdanska, 1998, except that instead of the number
 * of transitions we have a boolean flag indicating for each
 * line whether that line is the last transition for the current
 * state.
 */

",[]).

format_headers(c) :-
    format("
#include <wchar.h>
#include <stdlib.h>

",[]).
format_headers(cpp) :-
    format("
#include <stdlib.h>
#include <iostream>
#include <string>
#include <queue>
using namespace std;

",[]). 

format_headers(java) :-
    format("import FSA.*;~n",[]).

format_init(cpp,Type,Length,_L,_I,_FinalOutput,_FinalTable,_NextTable,S) :-
    format_unknown_symbol(cpp,S),
    format_procedure_head_cpp(Type),
    (	Length < 2**22
    ->	format("
  static const struct transition_struct {
    unsigned char symbol :  8;      // current symbol
    bool last            :  1;      // last line of current state?
    bool final           :  1;      // final state?
    unsigned int next    : 22;      // line in table to go to next
",[])
    ;   format("
  static const struct transition_struct {
    char symbol              :  8;      // current symbol
    bool last                :  1;      // last line of current state?
    bool final               :  1;      // final state?
    unsigned long int next   : 54;      // line in table to go to next
",[])
    ),
    format_header_output_cpp(Type),
    format("  } trans[]= {~n",[]).

format_init(c,Type,Length,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable,S) :-
    format_unknown_symbol(c,S),
    (	Length < 2**22
    ->	format("
static const struct transition_struct {
  wchar_t symbol ;      /* current symbol                     */
  unsigned int last    :  1;      /* last line of current state?        */
  unsigned int final   :  1;      /* final state?                       */
  unsigned int next    : 22;      /* line in table to go to next        */
",[])
    ;   format("
static const struct transition_struct {
  wchar_t symbol ;      /* current symbol                */
  unsigned int last        :  1;      /* last line of current state?   */
  unsigned int      final  :  1;      /* final state?                  */
  unsigned long int next   : 54;      /* line in table to go to next   */
",[])
    ),
    format_header_output(Type),
    format("  } trans[]= {~n",[]).

format_init(java,Type,_L,Label,Ini,_FinalOutput,_FinalTable,_NextTable,S) :-
    format_class_head(Type,Label,Ini),
    format_unknown_symbol(java,S),
    format("    private static final Transition tA [] = {~n",[]).

format_header_output(recognizer).
format_header_output(transducer(_)) :-
    format(
"    wchar_t *out;                     /* output string                      */
    wchar_t *final_out;               /* output string if we terminate here */
",[]).
format_header_output(w_recognizer) :-
    format(
"    unsigned int out;              /* output weight                      */
    unsigned int final_out;        /* output weight if we terminate here */
",[]).
 
format_header_output(w_transducer(_)) :-
    format(
"    wchar_t *out;                 /* output string                      */
    wchar_t *final_out;           /* output string if we terminate here */
    unsigned int wout;         /* output weight                      */
    unsigned int final_wout;   /* output weight if we terminate here */
",[]).
 
format_header_output_cpp(recognizer).
format_header_output_cpp(transducer(_)) :-
    format(
"    char *out;                     // output string
    char *final_out;               // output string if we terminate here
",[]).
format_header_output_cpp(w_recognizer) :-
    format(
"    unsigned int wout;              // output weight
    unsigned int final_wout;        // output weight if we terminate here
",[]).
 
format_header_output_cpp(w_transducer(_)) :-
    format(
"    char *out;                 // output string
    char *final_out;           // output string if we terminate here
    unsigned int wout;         // output weight
    unsigned int final_wout;   // output weight if we terminate here
",[]).
 
format_procedure_head(recognizer) :-
    format("
int accepts(wchar_t *in) {~n",[]).
format_procedure_head(transducer(_)) :-
    format("
/* int t_accepts(wchar_t *in, wchar_t *out, int max)
   in is transduced to out. Note that memory for out should be allocated
   by the caller. At most max characters are writting into out. 
   The return values:

   0: no transduction possible
   1: transduction possible
   2: length of transduction would be > max
*/

int t_accepts(wchar_t *in,wchar_t *out,int max) {~n",[]).
format_procedure_head(w_recognizer) :-
    format("
int w_accepts(wchar_t *in, int *out) {~n",[]).

format_procedure_head(w_transducer(_)) :-
    format("
/* int wt_accepts(wchar_t *in, wchar_t *out, int *wout, int max)
   in is transduced to out, with associated weight wout. 
   Note that memory for out should be allocated
   by the caller. At most max characters are writting into out. 
   The return values:

   0: no transduction possible
   1: transduction possible
   2: length of transduction would be > max
*/

int wt_accepts(wchar_t *in, wchar_t *out, int *wout,int max) {~n",[]).

format_procedure_head_cpp(recognizer) :-
    format("
bool accepts(string::const_iterator input_iter,  
                const string::const_iterator end) {~n",[]).

format_procedure_head_cpp(transducer(_)) :-
    format("
bool t_accepts(string::const_iterator input_iter,  
                const string::const_iterator end,  
                string& out) {~n",[]).

format_procedure_head_cpp(w_recognizer) :-
    format("
bool w_accepts(string::const_iterator input_iter,  
                const string::const_iterator end,  
                unsigned int& wout) {~n",[]).

format_procedure_head_cpp(w_transducer(_)) :-
    format("
bool wt_accepts(string::const_iterator input_iter,  
                const string::const_iterator end,  
                string& out, 
                unsigned int& wout) {~n",[]).


format_class_head(recognizer,Class,_Ini) :-
    format("

public class ~w extends DFAapplet {

    public ~w () {
        super(new DFA(tA,unknown_string));
    }
",[Class,Class]).

format_class_head(transducer(_),Class,Ini) :-
    format("

public class ~w extends DFAapplet {

    public ~w () {
        super(new DFA(tA,~s,unknown_string));
    }
",[Class,Class,Ini]).

format_class_head(w_recognizer,Class,Ini) :-
    format("

public class ~w extends DFAapplet {

    public ~w () {
        super(new DFA(tA,~s,unknown_string));
    }
",[Class,Class,Ini]).


format_class_head(w_transducer(_),Class,IniS/IniW) :-
    format("

public class ~w extends DFAapplet {

    public ~w () {
        super(new DFA(tA,~s,~s,unknown_string));
    }
",[Class,Class,IniS,IniW]).


format_sink(c,Type,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable):-
    format("                     /* sink                               */~n",
	   []),
    format_sink_c(Type).
format_sink(cpp,Type,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable):-
    format("                     // sink~n",
	   []),
    format_sink_cpp(Type).
format_sink(java,Type,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable):-
    format("        /* sink                               */~n",
	   []),
    format_sink_java(Type).

format_sink_c(recognizer):-
    format('    {0,1,0,0},~n',[]).

format_sink_c(transducer(_)):-
    format('    {0,1,0,0,L"",L""},~n',[]).

format_sink_c(w_recognizer):-
    format('    {0,1,0,0,0,0},~n',[]).

format_sink_c(w_transducer(_)):-
    format('    {0,1,0,0,L"",L"",0,0},~n',[]).


format_sink_cpp(recognizer):-
    format('    {\'\\0\',1,0,0},~n',[]).

format_sink_cpp(transducer(_)):-
    format('    {\'\\0\',1,0,0,"",""},~n',[]).

format_sink_cpp(w_recognizer):-
    format('    {\'\\0\',1,0,0,0,0},~n',[]).

format_sink_cpp(w_transducer(_)):-
    format('    {\'\\0\',1,0,0,"","",0,0},~n',[]).


format_sink_java(recognizer):-
    format('        new Transition(0,true,false,0),~n',
         []).


format_sink_java(transducer(_)):-
    format('        new Transition(0,true,false,0,"",""),~n',
         []).


format_sink_java(w_recognizer):-
    format('        new Transition(0,true,false,0,0,0),~n',
         []).

format_sink_java(w_transducer(_)):-
    format('        new Transition(0,true,false,0,"","",0,0),~n',
         []).


format_start(Lang,Type,_Label,Start,_Ini,FinalOutput,FinalTable,NextTable):-
    final_mark(Lang,Start,FinalTable,Final),
    next_mark(Start,NextTable,Next),
    final_out_mark(Type,Start,FinalOutput,FinalOut),
    format_start_string(Lang,Type,Final,Next,FinalOut).

format_start_string(java,Type,Final,Next,FinalOut):-
    format("        /* start state                                 */~n",
	   []),
    format_start_string_java(Type,Final,Next,FinalOut).
format_start_string(c,Type,Final,Next,FinalOut):-
    format("                     /* start state                        */~n",
	   []),
    format_start_string_c(Type,Final,Next,FinalOut).
format_start_string(cpp,Type,Final,Next,FinalOut):-
    format("                     // start state~n",
	   []),
    format_start_string_cpp(Type,Final,Next,FinalOut).

format_start_string_c(recognizer,Final,Next,_) :-
    format('    {0,0,~w,~w},~n',[Final,Next]).
format_start_string_c(transducer(_),Final,Next,FinalOut) :-
    format('    {0,0,~w,~w,L"",L~s},~n',
	   [Final,Next,FinalOut]).
format_start_string_c(w_recognizer,Final,Next,FinalOut) :-
    format('    {0,0,~w,~w,0,~s},~n',
	   [Final,Next,FinalOut]).
format_start_string_c(w_transducer(_),Final,Next,FinalOut/Wout) :-
    format('    {0,0,~w,~w,L"",L~s,0,~s},~n',
	   [Final,Next,FinalOut,Wout]).

format_start_string_cpp(recognizer,Final,Next,_) :-
    format('    {\'\\0\',0,~w,~w},~n',[Final,Next]).
format_start_string_cpp(transducer(_),Final,Next,FinalOut) :-
    format('    {\'\\0\',0,~w,~w,"",~s},~n',
	   [Final,Next,FinalOut]).
format_start_string_cpp(w_recognizer,Final,Next,FinalOut) :-
    format('    {\'\\0\',0,~w,~w,0,~s},~n',
	   [Final,Next,FinalOut]).
format_start_string_cpp(w_transducer(_),Final,Next,FinalOut/Wout) :-
    format('    {\'\\0\',0,~w,~w,"",~s,0,~s},~n',
	   [Final,Next,FinalOut,Wout]).

format_start_string_java(recognizer,Final,Next,_) :-
    format('        new Transition(0,false,~w,~w),~n',
	   [Final,Next]).
format_start_string_java(transducer(_),Final,Next,FinalOut):-
    format('        new Transition(0,false,~w,~w,"",~s),~n',
	   [Final,Next,FinalOut]).
format_start_string_java(w_recognizer,Final,Next,FinalOut):-
    format('        new Transition(0,false,~w,~w,0,~s),~n',
	   [Final,Next,FinalOut]).
format_start_string_java(w_transducer(_),Final,Next,FinalOut/Wout):-
    format('        new Transition(0,false,~w,~w,"",~s,0,~s),~n',
	   [Final,Next,FinalOut,Wout]).

final_out_mark(Type,State,Outs,Mark) :-
    (	Outs==[]
    ->  nil_out_mark(Type,Mark)
    ;	fsa_array:fsa_array_get(State,Mark0,Outs)
    ->	(   Mark0=Mark1/Mark2
	->  Mark=MarkL/MarkR,
	    mark_it(Mark1,MarkL),
	    mark_it(Mark2,MarkR)
	;   mark_it(Mark0,Mark)
	)
    ;	nil_out_mark(Type,Mark)
    ).

mark_it(Mark0,Mark) :-
    (	integer(Mark0)
    ->	integer_symbol(Mark0,Mark)
    ;	string_symbol(Mark0,Mark)
    ).

final_mark(c,State,Table,Mark) :-
    fsa_array:fsa_array_access(State,V,Table),
    (	nonvar(V)
    ->	Mark=1
    ;   Mark=0
    ).

final_mark(cpp,State,Table,Mark) :-
    fsa_array:fsa_array_access(State,V,Table),
    (	nonvar(V)
    ->	Mark=1
    ;   Mark=0
    ).

final_mark(java,State,Table,Mark) :-
    fsa_array:fsa_array_access(State,V,Table),
    (	nonvar(V)
    ->	Mark=true
    ;   Mark=false
    ).

next_mark(State,Table,Mark):-
    fsa_array:fsa_array_access(State,Mark,0,Table).

format_transitions([],_,_,_,_,_,_).
format_transitions([H|T],Lang,Type,Outs,Finals,Nexts,Fa) :-
    format_transitions1(T,H,Lang,Type,Outs,Finals,Nexts,Fa).

format_transitions1([],trans(_,Sym,Q),Lang,Type,Outs,Finals,Nexts,Fa) :-
    last_mark(Lang,1,Last),
    format_transition(Lang,Type,Last,Sym,Q,Outs,Finals,Nexts,Fa).

format_transitions1([trans(P1,Sym1,Q1)|T],trans(P,Sym,Q),Lang,Type,Outs,
		    Finals,Nexts,Fa):-
    (	P1 =:= P
    ->	last_mark(Lang,0,Last)
    ;   last_mark(Lang,1,Last)
    ),
    format_transition(Lang,Type,Last,Sym,Q,Outs,Finals,Nexts,Fa),
    format_transitions1(T,trans(P1,Sym1,Q1),Lang,Type,Outs,Finals,Nexts,Fa).

last_mark(c,I,I).
last_mark(cpp,I,I).
last_mark(java,I,Bool) :-
    last_mark_java(I,Bool).

last_mark_java(0,false).
last_mark_java(1,true).

format_transition(Lang,Type,Last,Sym,Q,Outs,Finals,Nexts,Fa):-
    \+ \+ format_transition0(Lang,Type,Last,Sym,Q,Outs,Finals,Nexts,Fa).

format_transition0(Lang,Type,Last,Sym,Q,Outs,Finals,Nexts,Fa) :-
    symbol_mark(Type,Fa,Sym,SymIn,SymOut),
    final_mark(Lang,Q,Finals,Final),
    final_out_mark(Type,Q,Outs,FinalOut),
    next_mark(Q,Nexts,Next),
    format_transition_string(Lang,Type,SymIn,SymOut,Last,Final,Next,FinalOut).

format_transition_string(c,Type,SymIn,SymOut,Last,Final,Next,FinalOut) :-
    format_transition_string_c(Type,SymIn,SymOut,Last,Final,Next,FinalOut).
format_transition_string(cpp,Type,SymIn,SymOut,Last,Final,Next,FinalOut) :-
    format_transition_string_cpp(Type,SymIn,SymOut,Last,Final,Next,FinalOut).
format_transition_string(java,Type,SymIn,SymOut,Last,Final,Next,FinalOut) :-
    format_transition_string_java(Type,SymIn,SymOut,Last,Final,Next,FinalOut).

format_transition_string_java(recognizer,SymIn,_SymOut,Last,Final,Next,_):-
    format("        new Transition(~w,~w,~w,~w),~n",
	   [SymIn,Last,Final,Next]).
format_transition_string_java(transducer(_),SymIn,SymOut,Last,Final,Next,Out):-
    format("        new Transition(~w,~w,~w,~w,~s,~s),~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_java(w_recognizer,SymIn,SymOut,Last,Final,Next,Out):-
    format("        new Transition(~w,~w,~w,~w,~s,~s),~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_java(w_transducer(_),SymIn,SymOut/W0,Last,Final,Next,
			      Out/W):-
    format("        new Transition(~w,~w,~w,~w,~s,~s,~s,~s),~n",
	   [SymIn,Last,Final,Next,SymOut,Out,W0,W]).

format_transition_string_c(recognizer,SymIn,_SymOut,Last,Final,Next,_):-
    format("    {~w,~w,~w,~w},~n",
	   [SymIn,Last,Final,Next]).
format_transition_string_c(transducer(_),SymIn,SymOut,Last,Final,Next,Out):-
    format("    {~w,~w,~w,~w,L~s,L~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_c(w_recognizer,SymIn,SymOut,Last,Final,Next,Out):-
    format("    {~w,~w,~w,~w,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_c(w_transducer(_),SymIn,SymOut/W0,Last,Final,Next,Out/W):-
    format("    {~w,~w,~w,~w,L~s,L~s,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out,W0,W]).

format_transition_string_cpp(recognizer,SymIn,_SymOut,Last,Final,Next,_):-
    format("    {~w,~w,~w,~w},~n",
	   [SymIn,Last,Final,Next]).
format_transition_string_cpp(transducer(_),SymIn,SymOut,Last,Final,Next,Out):-
    format("    {~w,~w,~w,~w,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_cpp(w_recognizer,SymIn,SymOut,Last,Final,Next,Out):-
    format("    {~w,~w,~w,~w,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_cpp(w_transducer(_),SymIn,SymOut/W0,Last,Final,Next,Out/W):-
    format("    {~w,~w,~w,~w,~s,~s,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out,W0,W]).

format_transitions_end(c) :-
    format("};~n",[]).
format_transitions_end(cpp) :-
    format("    };~n",[]).
format_transitions_end(java) :-
    format("    };~n",[]).

format_more(java,_,_,_,_,_,_).
format_more(c,transducer(T),_Label,Ini,_FinalOutput,_FinalTable,_NextTable) :-
    !,
    format_procedure_head(transducer(T)),
    format_ini_out(transducer(T),Ini),
    format_procedure(transducer(T)),
    format_new_procedure_head(transducer(T)),
    format_ini_out(transducer(T),Ini),
    format_new_procedure_body(transducer(T)).

format_more(c,Type,_Label,Ini,_FinalOutput,_FinalTable,_NextTable) :-
    format_procedure_head(Type),
    format_ini_out(Type,Ini),
    format_procedure(Type).


format_more(cpp,Type,_Label,Ini,_FinalOutput,_FinalTable,_NextTable) :-
    format_ini_out_cpp(Type,Ini),
    format_procedure_cpp(Type).

format_ini_out(recognizer,_).
format_ini_out(transducer(_),String):-
    format('  const wchar_t *ini_out=L~s;~n',[String]).
format_ini_out(w_recognizer,String):-
    format('  const unsigned int ini_out=~s;~n',[String]).
format_ini_out(w_transducer(_),String/W):-
    format('  const wchar_t *ini_out=L~s;~n  const unsigned int ini_wout=~s;~n',[String,W]).

format_ini_out_cpp(recognizer,_).

format_ini_out_cpp(transducer(_),String):-
    format('
  const string ini_out(~s);',[String]).

format_ini_out_cpp(w_recognizer,W):-
    format('
  const unsigned int ini_wout(~s);',[W]).

format_ini_out_cpp(w_transducer(_),String/W):-
    format('
  const string ini_out(~s);
  const unsigned int ini_wout(~s);',[String,W]).


format_procedure(recognizer) :-
    unknown_symbol(U),
    atom_codes(U,[Code]),
    format("  register wchar_t c;
  register unsigned int i;
  register unsigned long int line=1;
  i=0;
  c=in[i];
  while (c != '\\0') {
    line=trans[line].next;
    if (trans[line].symbol == ~w && unknown_symbol(c) )
      ;
    else {
      while (!trans[line].last && c > trans[line].symbol) {  
        line++;
      }
      if (c != trans[line].symbol)
	return 0;
    }
    i++;
    c=in[i];
  }
  return trans[line].final;        /* ok if at final state */
}

",[Code]).

format_procedure(transducer(_)) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    queue_symbol(Q),
    atom_codes(Q,[Qcode]),
    format("  register wchar_t c;
  register unsigned int i;
  register unsigned long int line=1;
  struct queue q;
  int outl;
  qinit(&q);
  i=0;
  c=in[i];
  out[0]='\\0';
  outl=0;
  outl+=wcslen(ini_out);
  if (outl > max) return 2;
  wcscat(out,ini_out);
  while (c != '\\0') {
    line=trans[line].next;
    if (trans[line].symbol == ~w && unknown_symbol(c) )
      ;
    else {
      if (trans[line].symbol == ~w && unknown_symbol(c) ) {
	qinsert(&q,c);
      }
      else {
        while (!trans[line].last && c > trans[line].symbol) {  
          line++;
        }",[Ucode,Qcode]),
    format("
      if (c != trans[line].symbol)
        return 0;
      }
    }
    if (outl+wcslen(trans[line].out) > max) return 2;
    wcscat(&out[outl],trans[line].out);
    replace_from_queue(&q,&out[outl]);
    outl+=wcslen(trans[line].out);
    i++;
    c=in[i];
  }
  if (outl+wcslen(trans[line].final_out) > max) return 2;
  wcscat(&out[outl],trans[line].final_out);
  replace_from_queue(&q,&out[outl]);
  return trans[line].final;        /* ok if at final state */
}

",[]).

format_procedure(w_recognizer) :-
    unknown_symbol(U),
    atom_codes(U,[Code]),
    format("  register wchar_t c;
  register unsigned int i;
  register unsigned long int line=1;
  i=0;
  c=in[i];
  *out=0;
  *out+=ini_out;
  while (c != '\\0') {
    line=trans[line].next;
    if (trans[line].symbol == ~w && unknown_symbol(c) )
      ;
    else {
      while (!trans[line].last && c > trans[line].symbol) {  
        line++;
    }
",[Code]),
    format(
"      if (c != trans[line].symbol)
         return 0;
    }
    *out+=trans[line].out;
    i++;
    c=in[i];
  }
    *out+=trans[line].final_out;
    return trans[line].final;
    
}

",[]).


format_procedure(w_transducer(_)) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    queue_symbol(Q),
    atom_codes(Q,[Qcode]),
    format("  register wchar_t c;
  register unsigned int i;
  register unsigned long int line=1;
  struct queue q;
  int outl;
  qinit(&q);
  i=0;
  c=in[i];
  *wout=0;
  *wout+=ini_wout;
  out[0]='\\0';
  outl=0;
  outl+=wcslen(ini_out);
  outl+=wcslen(ini_out);
  if (outl > max) return 2;
  wcscat(out,ini_out);
  while (c != '\\0') {
    line=trans[line].next;
    if (trans[line].symbol == ~w && unknown_symbol(c) )
      ;
    else {
      if (trans[line].symbol == ~w && unknown_symbol(c) ) {
	qinsert(&q,c);
      }
      else {
        while (!trans[line].last && c > trans[line].symbol) {  
          line++;
        }
",[Ucode,Qcode]),
    format(
"      if (c != trans[line].symbol)
         return 0;
       }
    }
    *wout+=trans[line].wout;
    if (outl+wcslen(trans[line].out) > max) return 2;
    wcscat(&out[outl],trans[line].out);
    replace_from_queue(&q,&out[outl]);
    outl+=wcslen(trans[line].out);
    i++;
    c=in[i];
  }
  *wout+=trans[line].final_wout;
  if (outl+wcslen(trans[line].final_out) > max) return 2;
  wcscat(&out[outl],trans[line].final_out);
  replace_from_queue(&q,&out[outl]);
  return trans[line].final;
    
}

",[]).

format_procedure_cpp(recognizer) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    format("
  size_t line = 1; 
 
  for (;input_iter != end; ++input_iter) { 
    line=trans[line].next; 
    if (trans[line].symbol == ~w && unknown_symbol(*input_iter) )
      ;
    else {
      while (!trans[line].last && (unsigned char)*input_iter > trans[line].symbol) {   
        line++; 
      } 
      if ((unsigned char)*input_iter != trans[line].symbol) line=0;          // sink  
    }
  } 
  return trans[line].final; 
} ",[Ucode]).

format_procedure_cpp(transducer(_)) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    queue_symbol(Q),
    atom_codes(Q,[Qcode]),
    format("
  out = ini_out; 
 
  size_t line = 1; 

  queue< char > q;
  
  for (;input_iter != end; ++input_iter) { 
    line=trans[line].next; 
    if (trans[line].symbol == ~w && unknown_symbol(*input_iter) )
      ;
    else {
      if (trans[line].symbol == ~w && unknown_symbol(*input_iter) ) {
	q.push(*input_iter);
      }
      else {
        while (!trans[line].last && (unsigned char)*input_iter > trans[line].symbol) {   
          line++; 
        } 
        if ((unsigned char)*input_iter != trans[line].symbol) line=0;          // sink
      }
    }
    out += replace_from_queue(q,trans[line].out);
  } 
  out += replace_from_queue(q,trans[line].final_out); 
  return trans[line].final; 
} ",[Ucode,Qcode]).

format_procedure_cpp(w_recognizer) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    format("
  wout = ini_wout; 
 
  size_t line = 1; 
  for (;input_iter != end; ++input_iter) { 
    line=trans[line].next; 
    if (trans[line].symbol == ~w && unknown_symbol(*input_iter) )
      ;
    else {
      while (!trans[line].last && (unsigned char)*input_iter > trans[line].symbol) {   
        line++; 
      } 
      if ((unsigned char)*input_iter != trans[line].symbol) line=0;          // sink  
    }
    wout += trans[line].wout; 
  } 
  wout += trans[line].final_wout; 
  return trans[line].final; 
} ",[Ucode]).

format_procedure_cpp(w_transducer(_)) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    queue_symbol(Q),
    atom_codes(Q,[Qcode]),
    format("
  wout = ini_wout; 
  out = ini_out; 
 
  size_t line = 1; 
 
  queue< char > q;
  
  for (;input_iter != end; ++input_iter) { 
    line=trans[line].next; 
    if (trans[line].symbol == ~w && unknown_symbol(*input_iter) )
      ;
    else {
      if (trans[line].symbol == ~w && unknown_symbol(*input_iter) ) {
	q.push(*input_iter);
      }
      else {
        while (!trans[line].last && (unsigned char)*input_iter > trans[line].symbol) {   
          line++; 
        } 
        if ((unsigned char)*input_iter != trans[line].symbol) line=0;          // sink  
      }
    }
    wout += trans[line].wout; 
    out += replace_from_queue(q,trans[line].out); 
  } 
  wout += trans[line].final_wout; 
  out += replace_from_queue(q,trans[line].final_out); 
  return trans[line].final; 
} ",[Ucode,Qcode]).

format_main(off,_Lang,_Type,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable).
format_main(on,Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable) :-
    format_main(Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable).

format_main(c,Type,_,Ini,_,_,_):-
    format_main_c(Type,Ini).

format_main(cpp,Type,_,_,_,_,_):-
    format_main_cpp(Type).

format_main(java,_,Label,_,_,_,_):-
    format("
    public static void main(String argv[]) {
	DFA.main(argv, new ~w());
    }",[Label]).

format_main_c(recognizer,_) :-
    format("

#include <stdio.h>
#include <locale.h>
#include <errno.h>
#include <string.h>
#define MAXLINE 131072        /* this constrains line length */
int main(int argc, char *argv[]) {
  wchar_t in[MAXLINE];
  wchar_t *ptr;

  setlocale(LC_ALL,\"\");

  errno = 0;
  ptr = fgetws(in,MAXLINE,stdin);
  while(ptr != NULL) {
    wchar_t *p = wcschr(in, '\\n');
    if (p) { *p = '\\0'; }  /* chomp */
					   
    if (accepts(in)) 
      fprintf(stderr,\"yes\\n\");
    else
      fprintf(stderr,\"no\\n\");

    errno = 0;
    ptr = fgetws(in,MAXLINE,stdin);
  }
  if (errno) {
    fprintf(stderr,\"%s error: %s\\n\",argv[0],strerror(errno));
  }
	                                /* end-of-file */
  return 0;
}

",[]).

format_main_c(transducer(_),_Ini) :-
    format("

#include <stdio.h>
#include <locale.h>
#include <errno.h>
#include <string.h>

",[]),
    format("

int main(int argc, char *argv[]) {
  wchar_t *in;
  wchar_t *out;
  int in_max;
  int out_max;
  int offset;
  int return_value;
  wchar_t *ptr;

  in_max = 1000;
  out_max = 1000;
  in = (wchar_t *) malloc ((in_max + 1) * sizeof(wchar_t));
  out = (wchar_t *) malloc ((out_max + 1) * sizeof(wchar_t));
  if (in == NULL || out == NULL) {
    fprintf(stderr,\"error: out of memory\\n\");
    return 2;
  }

  setlocale(LC_ALL,\"\");

  for(;;) {

    /* read a complete line */
    errno = 0;
    offset = 0;
    for(;;) {

      /* grow input buffer if necessary */
      if (offset > in_max - 2) {
	in_max *= 2;
	in = (wchar_t *) realloc (in, (in_max + 1)  * sizeof(wchar_t));
	if (in == NULL) {
	  fprintf(stderr,\"error: out of memory\\n\");
	  return 2;
	}
      }

      /* read (part of) sentence */
      ptr = fgetws(in+offset,in_max-offset,stdin);

      if (ptr == NULL)
	  break;    /* EOF or error encountered: DONE reading line */

      /* check for newline */
      wchar_t *p = wcschr(in, '\\n');
      if (p) {
	*p = '\\0';  /* chomp */
	break;      /* newline encountered: DONE reading line */
      }

      offset = wcslen(in);
    } /* end of: read a complete line */

    /* if there is data, process it */
    if (ptr != NULL || offset > 0) {
      return_value = new_t_accepts(in,&out,&out_max);
      switch(return_value) {
      case 0:
	fprintf(stderr,\"no\\n\");
	break;
      case 1:
	printf(\"%ls\\n\",out);
	break;
      case 2:
	fprintf(stderr,\"error: out of memory\\n\");
	break;
      }
    }

    /* finish if no more data */
    if (ptr == NULL)
	break;
  }
  if (errno) {
    fprintf(stderr,\"%s error: %s\\n\",argv[0],strerror(errno));
  }
	                                /* end-of-file */

  /* to satisfy valgrind */
  free(in);
  free(out);

  return 0;
}
	  

",[]).

format_main_c(w_recognizer,_) :-
    format("

#include <stdio.h>
#include <locale.h>
#include <errno.h>
#include <string.h>
#define MAXLINE 131072        /* this constrains line length */
int main(int argc, char *argv[]) {
  wchar_t in[MAXLINE];
  unsigned int out;
  wchar_t *ptr;

  setlocale(LC_ALL,\"\");

  errno = 0;
  ptr = fgetws(in,MAXLINE,stdin);
  while(ptr != NULL) {
    wchar_t *p = wcschr(in, '\\n');
    if (p) { *p = '\\0'; }  /* chomp */
					  
    if (w_accepts(in,&out)) 
      printf(\"%d\\n\",out);
    else
      fprintf(stderr,\"no\\n\");

    errno = 0;
    ptr = fgetws(in,MAXLINE,stdin);
  }
  if (errno) {
    fprintf(stderr,\"%s error: %s\\n\",argv[0],strerror(errno));
  }
                                          /* end-of-file */
  return 0;
}

",[]).


format_main_c(w_transducer(_),_) :-
    format("

#include <stdio.h>
#include <locale.h>
#include <errno.h>
#include <string.h>
#define MAXLINE 131072        /* this constrains line length */
int main(int argc, char *argv[]) {
  wchar_t in[MAXLINE];
  wchar_t out[MAXLINE];
  unsigned int wout;
  int return_value;
  wchar_t *ptr;

  setlocale(LC_ALL,\"\");

  errno = 0;
  ptr = fgetws(in,MAXLINE,stdin);
  while(ptr != NULL) {
    wchar_t *p = wcschr(in, '\\n');
    if (p) { *p = '\\0'; }  /* chomp */

    return_value = wt_accepts(in,out,&wout,MAXLINE);
    switch(return_value) {
    case 0:
      fprintf(stderr,\"no\\n\");
      break;
    case 1:
      printf(\"%d:%S\\n\",wout,out);
      break;
    case 2:
      fprintf(stderr,\"error: output line too long\\n\");
      break;
    }

    errno = 0;
    ptr = fgetws(in,MAXLINE,stdin);
  }
  if (errno) {
    fprintf(stderr,\"%s error: %s\\n\",argv[0],strerror(errno));
  }
                                         /* end-of-file */
  return 0;
}

",[]).

format_new_procedure_head(transducer(_)) :-
    format("
int resize_buf(int n, int *max, wchar_t **buf) {
  while (n > *max)
    *max *= 2;
  *buf = (wchar_t *) realloc (*buf, (*max + 1)  * sizeof(wchar_t));
  return buf != NULL;
}

int new_t_accepts(wchar_t *in,wchar_t **out,int *max) {
",[]).

format_new_procedure_body(transducer(_)) :-
    format("
  register wchar_t c;
  register unsigned int i;
  register unsigned long int line=1;
  struct queue q;
  int outl;
  wchar_t *myout = *out;
  qinit(&q);
  i=0;
  c=in[i];
  myout[0]='\\0';
  outl=wcslen(ini_out);
  if (outl >= *max && !resize_buf(outl,max,&myout)) { *out = myout; return 2; }
  wcscat(myout,ini_out);
  while (c != '\\0') {
    line=trans[line].next;
    if (trans[line].symbol == 1 && unknown_symbol(c) )
      ;
    else {
      if (trans[line].symbol == 2 && unknown_symbol(c) ) {
	qinsert(&q,c);
      }
      else {
        while (!trans[line].last && c > trans[line].symbol) {
          line++;
        }
        if (c != trans[line].symbol) {
	  *out = myout;
	  return 0;
        }
      }
    }
    if (outl+wcslen(trans[line].out) >= *max && !resize_buf(outl+wcslen(trans[line].out),max,&myout)) { *out = myout; return 2; }
    wcscat(&myout[outl],trans[line].out);
    replace_from_queue(&q,&myout[outl]);
    outl+=wcslen(trans[line].out);
    i++;
    c=in[i];
  }
  if (outl+wcslen(trans[line].final_out) >= *max && !resize_buf(outl+wcslen(trans[line].final_out),max,&myout)) { *out = myout; return 2; }
  wcscat(&myout[outl],trans[line].final_out);
  replace_from_queue(&q,&myout[outl]);
  *out = myout;
  return trans[line].final;        /* ok if at final state */
}

",[]).
    
					 
format_main_cpp(recognizer) :-
    format("
int main() { 
 
  string in; 
  while(getline(cin,in)) { 
    if (accepts(in.begin(),in.end()))  
      cerr << \"yes\" << endl;
    else 
      cerr << \"no\" << endl; 
  } 
} 
",[]).


format_main_cpp(transducer(_)) :-
    format("
int main() { 
 
  string in, out; 
  while(getline(cin,in)) { 
    if (t_accepts(in.begin(),in.end(),out))  
      cout << out << endl; 
    else 
      cerr << \"no\" << endl; 
  } 
} 
",[]).


format_main_cpp(w_recognizer) :-
    format("
int main() { 
 
  unsigned int wout; 
  string in; 
  while(getline(cin,in)) { 
    if (w_accepts(in.begin(),in.end(),wout))  
      cout << wout << endl; 
    else 
      cerr << \"no\" << endl; 
  } 
} 
",[]).


format_main_cpp(w_transducer(_)) :-
    format("
int main() { 
 
  unsigned int wout; 
  string in, out; 
  while(getline(cin,in)) { 
    if (wt_accepts(in.begin(),in.end(),out,wout))  
      cout << wout << ':' << out << endl; 
    else 
      cerr << \"no\" << endl; 
  } 
} 
",[]).


format_tail(c,_,_,_,_,_,_).
format_tail(cpp,_,_,_,_,_,_).
format_tail(java,_Type,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable) :-
    format("~n}~n~n",[]).

nil_out_mark(recognizer,[]).
nil_out_mark(transducer(_),[34,34]).    % ""
nil_out_mark(w_recognizer,[48]).     % 0
nil_out_mark(w_transducer(_),[34,34]/[48]).     

format_unknown_symbol(Lang,Syms) :-
    format_unknown_symbol_start(Lang),
    format_unknown_symbol_list(Syms,Lang),
    format_unknown_symbol_end(Lang).

format_unknown_symbol_start(c) :-
    format("
int unknown_symbol(wchar_t c) {
  return ( 1~n",[]).

format_unknown_symbol_start(cpp) :-
    format("
bool unknown_symbol(char c) {
  return ( 1~n",[]).

format_unknown_symbol_start(java) :-
    format('        private static String unknown_string="',[]).

format_unknown_symbol_end(cpp) :-
    format("	   );
         
}
",[]).
format_unknown_symbol_end(java) :-
    format('";~n',[]).
format_unknown_symbol_end(c) :-
    format("	   );
         
}
",[]).


format_unknown_symbol_list([],_).
format_unknown_symbol_list([H|T],Lang) :-
    (	atom(H)
    ->	atom_codes(H,[N])
    ;	number(H)
    ->	number_codes(H,[N])
    ;	fail
    ),
    format_unknown_symbol_(Lang,N,H),
    format_unknown_symbol_list(T,Lang).

format_unknown_symbol_(cpp,N,_) :-
    format("        && c != (char)~w~n",[N]).
format_unknown_symbol_(java,_,H) :-
    format("~w",[H]).
format_unknown_symbol_(c,N,_) :-
    format("        && c != (wchar_t)~w~n",[N]).

%% returns the *input* symbols of Fa
symbols_talked_about(Fa,Syms) :-
    fsa_data:fsa_transitions(Fa,Trans),
    fsa_data:fsa_type(Fa,Type),
    fsa_data:fsa_symbols_decl(Fa,Decl),
    symbol_part_of_trans(Trans,SymTrans0),
    sort(SymTrans0,SymTrans),
    transition_symbols_talked_about(SymTrans,Type,Decl,Syms0),
    sort(Syms0,Syms).

symbol_part_of_trans([],[]).
symbol_part_of_trans([trans(_,S,_)|Ts],[S|Ss]) :-
    symbol_part_of_trans(Ts,Ss).

transition_symbols_talked_about([],_,_,[]).
transition_symbols_talked_about([S|Tr],Type,Decl,Syms):-
    transition_symbol_talked_about(Type,Decl,S,Syms,Syms1),
    transition_symbols_talked_about(Tr,Type,Decl,Syms1).

transition_symbol_talked_about(recognizer,r(Mod),S,Syms0,Syms) :-
    pred_talked_about(Mod,S,Syms0,Syms).
transition_symbol_talked_about(w_recognizer,r(Mod,_),S/_,Syms0,Syms) :-
    (	S == []
    ->	Syms0=Syms
    ;   pred_talked_about(Mod,S,Syms0,Syms)
    ).
transition_symbol_talked_about(transducer(Sub),t(Ma,Mb),Sa/Sb,Syms0,Syms):-
    transition_symbol_talked_about(Sub,Ma,Mb,Sa,Sb,Syms0,Syms).
transition_symbol_talked_about(w_transducer(Sub),t(Ma,Mb,_),Sa/(Sb/_),Syms0,
			       Syms):-
    transition_symbol_talked_about(Sub,Ma,Mb,Sa,Sb,Syms0,Syms).

transition_symbol_talked_about(_,Ma,_Mb,Sa,_Sb,Syms0,Syms):-
    (	Sa==[]
    ->	Syms0=Syms
    ;	Sa='$@'(Sa1)
    ->	pred_talked_about(Ma,Sa1,Syms0,Syms)
    ;   pred_talked_about(Ma,Sa,Syms0,Syms)
    ).

pred_talked_about(M,H,S0,S) :-
    findall(Sym,fsa_pred_module:talks_about(M,H,Sym),S0,S).


format_queue(recognizer,_).
format_queue(w_recognizer,_).
format_queue(transducer(_),Lang) :-
    format_queue_(Lang).
format_queue(w_transducer(_),Lang) :-
    format_queue_(Lang).

format_queue_(java).
format_queue_(cpp) :-
    queue_symbol(Q),
    atom_codes(Q,[Code]),
    format("
string replace_from_queue(queue< char >& q,string in)
{
  int i=0;
  char c;
  c=in[i];

  while (c != '\\0') {
    if (c == ~w) {
      in[i]=q.front();
      q.pop();
    }
    i++;
    c=in[i];
  }
  return in;
}
	  ",[Code]).


format_queue_(c) :-
    format("

/*
 * Queue ADT
 *
 * Taken from http://purists.org/queue/
 * Institute of Applied Iconoclasm
 *
 * GvN: added explicit casts for malloc() to prevent warnings
 */

#define QDATUM wchar_t
	  
/*
 *  Queue item structure
 */
struct qentry
{
        struct qentry *next;
        QDATUM d;
};
",[]),
format("
/*
 *  Queue container structure
 *
 *  This structure encapsulates a linked list of qentry items.
 */
struct queue
{
        struct qentry *begin, **end;
};

",[]),
format("
/*
 *  qinit: Initialize queue.
 *
 *  Parameters:
 *
 *    q         Pointer to a queue, or NULL if the user wishes to leave 
 *              it to qinit to allocate the queue.
 *
 *  Return values:
 *
 *    non-NULL  Queue has been initialized.
 *
 *    NULL      Insufficient memory.
 */
struct queue *
qinit(struct queue *q)
{
        if (q || (q = (struct queue *) malloc(sizeof(struct queue))) != NULL) {
                q->begin = NULL;
                q->end = &q->begin;
        }
        return q;
}

",[]),
format("
/*
 *  qinsert: append an item to the queue.
 *
 *  Parameters:
 *
 *    q         Pointer to a queue. It is assumed the queue has been
 *              initialized by a call to qinit.
 *
 *    d         Item to be appended.
 *
 *  Return values:      
 *
 *    1         The item has been appended.
 *
 *    0         The item could not be appended. Either the queue 
 *              pointer provided was NULL, or the function was unable 
 *              to allocate the amount of memory needed for a new 
 *              queue item.     
 */
int
qinsert(struct queue *q, QDATUM d)
{
        if (!q || !(*q->end = (struct qentry *) malloc(sizeof(struct qentry)))) return 0;
        (*q->end)->d = d;
        (*q->end)->next = NULL;
        q->end = &((*q->end)->next);
        return 1;
}

",[]),
format("
/*
 *  qremove: remove an item from the queue.
 *
 *  Parameters:
 *
 *    q         Pointer to a queue.
 *
 *    d         Pointer to the QDATUM variable that will hold the datum
 *              corresponding to the queue item.
 *
 *  Return values:
 *
 *    non-NULL  An item has been removed. The variable that d points 
 *              to now contains the datum associated with the item 
 *              in question.
 *              
 *    NULL      No item could be removed. Either the queue pointer 
 *              provided was NULL, or the queue was empty. The memory 
 *              location that d points to has not been modified. 
 */
QDATUM *
qremove(struct queue *q, QDATUM *d)
{
        struct qentry *tmp;
                
        if (!q || !q->begin) return NULL;
        tmp = q->begin;
        if (!(q->begin = q->begin->next)) q->end = &q->begin;
        *d = tmp->d;
        free(tmp);
        return d;       
}

",[]),
format("
/*
 *  qpeek: access an item without removing it from the queue.
 *
 *  Parameters:
 *
 *    q         Pointer to a queue.
 *
 *    d         Pointer to the QDATUM variable that will hold the datum
 *              associated with the first item in the queue, i. e.,
 *              the item that would be removed had qremove been called
 *              instead of qpeek.
 *
 *  Return values:
 * 
 *    See qremove.
 */
QDATUM *
qpeek(struct queue *q, QDATUM *d)
{
        if (!q || !q->begin) return NULL;
        *d = q->begin->d;
        return d;
}


",[]),
    queue_symbol(Q),
    atom_codes(Q,[Code]),
format("
void replace_from_queue(struct queue *q,wchar_t *in)
{
  int i=0;
  wchar_t c;
  wchar_t d;
  c=in[i];
  while (c != '\\0') {
    if (c == ~w) {
      qremove(q,&d);
      in[i]=d;
    }
    i++;
    c=in[i];
  }

}


",[Code]).



new_symbols_decl(recognizer,Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,_,r(fsa_frozen)).
new_symbols_decl(transducer(_Sub),Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,_,t(fsa_frozen,fsa_frozen)).
new_symbols_decl(w_recognizer,Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,_,r(fsa_frozen,fsa_frozen)).
new_symbols_decl(w_transducer(_Sub),Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,_,t(fsa_frozen,fsa_frozen,fsa_frozen)).

%%% limit input symbols to singleton set predicates of one-letter atoms
%%% for open-ended alphabets, use symbol '\002' 
%%% for identities over open-ended alphabets, use symbol
%%% '\001\' to indicate enqueue/dequeue
n_expand_predicates_fa(recognizer,Fa0,Fa,DefaultSymbols):-
    n_expand_predicates_fa(Fa0,Fa1,DefaultSymbols),
    new_symbols_decl(recognizer,Fa1,Fa).
n_expand_predicates_fa(w_recognizer,Fa0,Fa,DefaultSymbols):-
    n_expand_predicates_fa(Fa0,Fa1,DefaultSymbols),
    new_symbols_decl(w_recognizer,Fa1,Fa).
n_expand_predicates_fa(transducer(Sub),Fa0,Fa,DefaultSymbols):-
    fsa_regex:to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    n_expand_predicates_fa(Fa1,Fa2,DefaultSymbols),
    new_symbols_decl(transducer(Sub),Fa2,Fa).
n_expand_predicates_fa(w_transducer(Sub),Fa0,Fa,DefaultSymbols):-
    fsa_regex:w_to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    n_expand_predicates_fa(Fa1,Fa2,DefaultSymbols),
    new_symbols_decl(w_transducer(Sub),Fa2,Fa).

n_expand_predicates_fa(Fa0,Fa,DefaultSymbols) :-
    fsa_data:fsa_symbols_decl(Fa0,Mods),
    fsa_data:fsa_copy_except(transitions,Fa0,Fa1,Trans0,Trans),
    fsa_data:fsa_copy_except(states,Fa1,Fa,Nr0,Nr),
    n_expand_transitions(Trans0,Trans1,Mods,DefaultSymbols,Nr0,Nr),
    sort(Trans1,Trans).

n_expand_transitions([],[],_,_Ds,Nr,Nr).
n_expand_transitions([H|T],Tr,Mods,Ds,Nr0,Nr):-
    n_expand_transition(Mods,H,Tr,Tr1,Ds,Nr0,Nr1),
    n_expand_transitions(T,Tr1,Mods,Ds,Nr1,Nr).

n_expand_transition(r(Mod),trans(P,Pred,Q),Tr0,Tr,Ds,Nr0,Nr):-
    findall(trans(P,Sym,Q),n_eval_pred(Mod,Pred,Sym,Ds),Tr1),
    expand_multi(Tr1,Tr0,Tr,Nr0,Nr,r(Mod)).

n_expand_transition(r(Mod,F),trans(P,Pred/W,Q),Tr0,Tr,Ds,Nr0,Nr):-
    findall(trans(P,Sym/W,Q),n_eval_pred(Mod,Pred,Sym,Ds),Tr1),
    expand_multi(Tr1,Tr0,Tr,Nr0,Nr,r(Mod,F)).

n_expand_transition(t(ModD,ModR),trans(P,PredD/PredR,Q),Tr0,Tr,Ds,Nr0,Nr):-
    findall(trans(P,SymD/SymR,Q),
	    n_eval_pred_pair(ModD,ModR,PredD,PredR,SymD,SymR,Ds),
	    Tr1),
    expand_multi(Tr1,Tr0,Tr,Nr0,Nr,t(ModD,ModR)).

n_expand_transition(t(ModD,ModR,F),trans(P,PredD/(PredR/W),Q),Tr0,Tr,Ds,Nr0,Nr):-
    findall(trans(P,SymD/(SymR/W),Q),
	    n_eval_pred_pair(ModD,ModR,PredD,PredR,SymD,SymR,Ds),
	    Tr1),
    expand_multi(Tr1,Tr0,Tr,Nr0,Nr,t(ModD,ModR,F)).

expand_multi([],Tr,Tr,Nr,Nr,_).
expand_multi([H|T],Tr0,Tr,Nr0,Nr,Mod) :-
    expand_multi_(Mod,H,Tr0,Tr1,Nr0,Nr1),
    expand_multi(T,Tr1,Tr,Nr1,Nr,Mod).

expand_multi_(r(_),trans(P,Sym,Q),Trs0,Trs,Nr0,Nr) :-
    (	Sym==[]
    ->	Nr0=Nr,
	Trs0=[trans(P,Sym,Q)|Trs]
    ;	atom_to_atoms(Sym,Syms,[]),
	exp_transitions_r(Syms,P,Q,Trs0,Trs,Nr0,Nr)
    ).

expand_multi_(r(_,_),trans(P,Sym/W,Q),Trs0,Trs,Nr0,Nr) :-
    (	Sym==[]
    ->	Nr0=Nr,
	Trs0=[trans(P,Sym/W,Q)|Trs]
    ;	atom_to_atoms(Sym,Syms,[]),
	exp_transitions_wr(Syms,W,P,Q,Trs0,Trs,Nr0,Nr)
    ).

expand_multi_(t(_,_),trans(P,Sym/Out,Q),Trs0,Trs,Nr0,Nr) :-
    (	Sym == []
    ->	Nr0=Nr,
	Trs0=[trans(P,Sym/Out,Q)|Trs]
    ;	atom_to_atoms(Sym,Syms,[]),
	exp_transitions_t(Syms,Out,P,Q,Trs0,Trs,Nr0,Nr)
    ).

expand_multi_(t(_,_,_),trans(P,Sym/W,Q),Trs0,Trs,Nr0,Nr) :-
    (	Sym == []
    ->	Nr0=Nr,
	Trs0=[trans(P,Sym/W,Q)|Trs]
    ;	atom_to_atoms(Sym,Syms,[]),
	exp_transitions_wt(Syms,W,P,Q,Trs0,Trs,Nr0,Nr)
    ).

exp_transitions_r([],_,_,Trs,Trs,Nr,Nr).
exp_transitions_r([H|T],P,Q,[trans(P,H,Q1)|Trs0],Trs,Nr0,Nr):-
    exp_transitions_r_(T,Q1,Q,Trs0,Trs,Nr0,Nr).

exp_transitions_r_([],Q,Q,Trs,Trs,Nr,Nr).
exp_transitions_r_([H|T],Q0,Q,[trans(Q0,H,Q1)|Trs0],Trs,Q0,Nr) :-
    Nr1 is Q0+1,
    exp_transitions_r_(T,Q1,Q,Trs0,Trs,Nr1,Nr).

exp_transitions_wr([],_,_,_,Trs,Trs,Nr,Nr).
exp_transitions_wr([H|T],W,P,Q,[trans(P,H/W,Q1)|Trs0],Trs,Nr0,Nr):-
    exp_transitions_wr_(T,Q1,Q,Trs0,Trs,Nr0,Nr).

exp_transitions_wr_([],Q,Q,Trs,Trs,Nr,Nr).
exp_transitions_wr_([H|T],Q0,Q,[trans(Q0,H/Zero,Q1)|Trs0],Trs,Q0,Nr) :-
    fsa_semiring:zero(fsa_weights,Zero),
    Nr1 is Q0+1,
    exp_transitions_wr_(T,Q1,Q,Trs0,Trs,Nr1,Nr).

exp_transitions_t([],_,_,_,Trs,Trs,Nr,Nr).
exp_transitions_t([H|T],Out,P,Q,[trans(P,H/Out,Q1)|Trs0],Trs,Nr0,Nr):-
    exp_transitions_t_(T,Q1,Q,Trs0,Trs,Nr0,Nr).

exp_transitions_t_([],Q,Q,Trs,Trs,Nr,Nr).
exp_transitions_t_([H|T],Q0,Q,[trans(Q0,H/Zero,Q1)|Trs0],Trs,Q0,Nr) :-
    fsa_semiring:zero(fsa_strings,Zero),
    Nr1 is Q0+1,
    exp_transitions_t_(T,Q1,Q,Trs0,Trs,Nr1,Nr).

exp_transitions_wt([],_,_,_,Trs,Trs,Nr,Nr).
exp_transitions_wt([H|T],W,P,Q,[trans(P,H/W,Q1)|Trs0],Trs,Nr0,Nr):-
    exp_transitions_wt_(T,Q1,Q,Trs0,Trs,Nr0,Nr).

exp_transitions_wt_([],Q,Q,Trs,Trs,Nr,Nr).
exp_transitions_wt_([H|T],Q0,Q,[trans(Q0,H/Zero,Q1)|Trs0],Trs,Q0,Nr) :-
    fsa_semiring:zero(fsa_weighted_strings,Zero),
    Nr1 is Q0+1,
    exp_transitions_wt_(T,Q1,Q,Trs0,Trs,Nr1,Nr).

n_eval_pred_pair(Mod,Mod,'$@'(Pred0),'$@'(Pred0),Sym,Sym,Ds):-
    !,
    n_eval_pred(Mod,Pred0,Sym0,Ds),
    (	unknown_symbol(Sym0)
    ->	queue_symbol(Sym)
    ;	Sym0=Sym
    ).

n_eval_pred_pair(ModD,ModR,PredD,PredR,SymD,SymR,Ds):-
    (	PredD==[]
    ->  SymD=[]
    ;   n_eval_pred(ModD,PredD,SymD,Ds)
    ),
    (	PredR==[]
    ->  SymR=[]
    ;   n_eval_pred(ModR,PredR,SymR)
    ).

n_eval_pred(Mod,Pred,Sym,Ds) :-
    lists:member(Sym0,Ds),
    fsa_pred_module:evaluate_predicate(Mod,Pred,Sym0),
    ensure_atom(Sym0,Sym).

ensure_atom(Sym0,Sym) :-
    (	atom(Sym0)
    ->	Sym0=Sym
    ;	number(Sym0)
    ->	number_codes(Sym0,Codes),
	atom_codes(Sym,Codes)
    ;	fsa_util:term_atom(Sym0,Sym)
    ).

n_eval_pred(Mod,Pred,Sym) :-
    fsa_pred_module:evaluate_predicate(Mod,Pred,Sym0),
    (	var(Sym0)
    ->	fsa_globals:fsa_global_get(fl_arbitrary_symbol,Sym0)
    ;	true
    ),
    ensure_atom(Sym0,Sym).

unknown_symbol(X) :-
    (	var(X)
    ->	atom_codes(X,[1])
    ;	atom(X),
	atom_codes(X,[1])
    ).

queue_symbol(X) :-
    (	var(X)
    ->	atom_codes(X,[2])
    ;	atom(X),
	atom_codes(X,[2])
    ).
