:- module(fsa_options,[ parse/3 ]).         %% fsa_parse_options

#swi# :- style_check([-atom]).

#sic# :- use_module(library(system),  [ file_exists/2 ]).
#sic# :- use_module(library(charsio), [ read_from_chars/2 ]).
#yap# :- use_module(library(charsio), [ read_from_chars/2 ]).
#swi# :- use_module(charsio, [ read_from_chars/2 ]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous fsa_command/4.

%% command line parsing.
parse(Cmd) -->
    parse_options,
    continue_parse(Cmd).

%% normal case: a specific operation is required
continue_parse(cmd(Cmd)) -->
    fsa_command(Cmd),
    parse_options.

%% no operation: run interactively
continue_parse(true) -->
    { fsa_globals:fsa_global_get(interactive,OnOff),
      (	  OnOff == off
      ->  fsa_globals:fsa_global_set(interactive,on)
      ;   true
      )
    }.

input_output(File,File2) -->
    input_file(File),
    output_file_or_user(File2).
input_output(user,user) --> [].

input_file_or_user(File) -->
    input_file(File).
input_file_or_user(user) --> [].

input_file(File) --> 
    [File0],
    { absolute_file_name(File0,File),
#sic#       file_exists(File,[read])
#yap#       exists(File)
#swi#       exists_file(File)
    }.

output_file_or_user(File) -->
    output_file(File).
output_file_or_user(user) --> [].

output_file(File) -->
    [File].

fsa_command(Command) -->
    option(Option0),
    { abbreviates(Option0,Option) },
    fsa_command(Option,Command).

fsa_command(Command,P0,P) :-
    option(Option,P0,P1),
    user:user_command(Option,Command,P1,P).

:- multifile user:unknown_predicate_handler/2.
:- public user:unknown_predicate_handler/2.
user:unknown_predicate_handler(user_command(_,_,_,_),fail).

abbreviates(conv,         conversion).
abbreviates(conversion,   conversion).
abbreviates(cmdint,       cmdint).
abbreviates(cmd_int,      cmdint).
abbreviates(approx,       approx).
abbreviates(fsa2fsm,      fsa2fsm).
abbreviates(fsatofsm,     fsa2fsm).
abbreviates(tofsm,        fsa2fsm).
abbreviates(fsm2fsa,      fsm2fsa).
abbreviates(fsmtofsa,     fsm2fsa).
abbreviates(compile,      compile).
abbreviates(c,            compile).
abbreviates(cc,           compile_to_c).
abbreviates(c_c,          compile_to_c).
abbreviates(compile_c,    compile_to_c).
abbreviates(compile_toc,  compile_to_c).
abbreviates(compile_to_c, compile_to_c).
abbreviates(java,         java).
abbreviates(cpp,          cpp).
abbreviates('c++',        cpp).
abbreviates(cleanup,      cleanup).
abbreviates(compose,      compose).
abbreviates(comp,         compose).
abbreviates(complement,   complement).
abbreviates(compl,        complement).
abbreviates(count,        count).
abbreviates(density,      density).
abbreviates(davinci,      davinci).
abbreviates(dav,          davinci).
abbreviates(daVinci,      davinci).
abbreviates(vcg,          vcg).
abbreviates(xvcg,         vcg).
abbreviates(dot,          dot).
abbreviates(d,            determinize).
abbreviates(det,          determinize).
abbreviates(determinize,  determinize).
abbreviates(drgraph,      dreachgraph).
abbreviates(per_reachable_graph,
	                  dreachgraph).
abbreviates(dcrgraph,     dcreachgraph).
abbreviates(per_co_reachable_graph,
	                  dcreachgraph).
abbreviates(d7,           dgraph).
abbreviates(dgraph,       dgraph).
abbreviates(per_graph,    dgraph).
abbreviates(d8,           dsubset).
abbreviates(dsubset,      dsubset).
abbreviates(per_subset,   dsubset).
abbreviates(d10,          dstate).
abbreviates(dstate,       dstate).
abbreviates(per_state,    dstate).
abbreviates(efree,        efree).
abbreviates(e_free,       efree).
abbreviates(epsilonfree,  efree).
abbreviates(epsilon_free  ,efree).
abbreviates(jumpfree,     efree).
abbreviates(jump_free,    efree).
abbreviates(ignore,       ignore).
abbreviates(ign,          ignore).
abbreviates(difference,   difference).
abbreviates(diff,         difference).
abbreviates(a,            accepts).
abbreviates(apply,        accepts).
abbreviates(accepts,      accepts).
abbreviates(accept_all,   accept_all).
abbreviates(aa,           accept_all).
abbreviates(raa,          r_accept_all).
abbreviates(taa,          t_accept_all). % only for web-demo
abbreviates(command,      prolog).
abbreviates(prolog,       prolog).
abbreviates(leslie,       generate).
abbreviates(random,       generate).
abbreviates(g,            generate).
abbreviates(generate,     generate).
abbreviates(intersect,    intersect).
abbreviates(kleenestar,   kleene_star).
abbreviates(kleene_star,  kleene_star).
abbreviates(star,         kleene_star).
abbreviates(kleene,       kleene_star).
abbreviates(kleene_plus,  kleene_plus).
abbreviates(kleeneplus,   kleene_plus).
abbreviates(reverse,      reverse).
abbreviates(rev,          reverse).
abbreviates(reversal,     reverse).
abbreviates(inverse,      inverse).
abbreviates(invert,       inverse).
abbreviates(inversion,    inverse).
abbreviates(dom,          domain).
abbreviates(domain,       domain).
abbreviates(ran,          range).
abbreviates(range,        range).
abbreviates(id,           identity).
abbreviates(identity,     identity).
abbreviates(option,       option).
abbreviates(opt,          option).
abbreviates(optional,     option).
abbreviates(union,        union).
abbreviates(concat,       concat).
abbreviates(conc,         concat).
abbreviates(m,            minimize).
abbreviates(min,          minimize).
abbreviates(minimize,     minimize).
abbreviates(m1,           mhu).
abbreviates(min1,         mhu).
abbreviates(minimize1,    mhu).
abbreviates(mhu,          mhu).
abbreviates(m2,           mb).
abbreviates(min2,         mb).
abbreviates(minimize2,    mb).
abbreviates(mb,           mb).
abbreviates(m3,           mh).
abbreviates(min3,         mh).
abbreviates(minimize3,    mh).
abbreviates(mh,           mh).
abbreviates(t_minimize,   t_minimize).
abbreviates(t_m,          t_minimize).
abbreviates(tminimize,    t_minimize).
abbreviates(t_min,        t_minimize).
abbreviates(tmin,         t_minimize).
abbreviates(tm,           t_minimize).
abbreviates(w_m,          w_minimize).
abbreviates(w_minimize,   w_minimize).
abbreviates(wminimize,    w_minimize).
abbreviates(w_min,        w_minimize).
abbreviates(wmin,         w_minimize).
abbreviates(wm,           w_minimize).
abbreviates(w_m,          w_minimize).
abbreviates(wt_minimize,  wt_minimize).
abbreviates(wt_m,         wt_minimize).
abbreviates(wtminimize,   wt_minimize).
abbreviates(wt_min,       wt_minimize).
abbreviates(wtmin,        wt_minimize).
abbreviates(wtm,          wt_minimize).
abbreviates(produce,      produce).
abbreviates(sample,       sample).
abbreviates(p,            produce).
abbreviates(prod,         produce).
abbreviates(r,            regex).
abbreviates(regex,        regex).
abbreviates(tk,           tk).
abbreviates(x,            tk).
abbreviates('Tk',         tk).
abbreviates('X',          tk).
abbreviates(ps,           postscript).
abbreviates(postscript,   postscript).
abbreviates(dict2ph,         fsa_dict_to_perfect_hash).
abbreviates(dicttoph,        fsa_dict_to_perfect_hash).
abbreviates(fsa_dict_to_perfect_hash,
	                     fsa_dict_to_perfect_hash).
abbreviates(fsa_dict_to_fsa, fsa_dict_to_fsa).
abbreviates(dicttom,         fsa_dict_to_fsa).
abbreviates(dict2m,          fsa_dict_to_fsa).
abbreviates(pstricks,        pstricks).
abbreviates(pstricks_tex,    pstricks_tex).
abbreviates(pstricks_picture,pstricks_picture).
abbreviates(copy,            copy).
abbreviates(dict,            dict).
abbreviates(prefix_tree,     prefix_tree).
abbreviates(trie,            dict).
abbreviates(td,              t_determinize).
abbreviates(t_d,             t_determinize).
abbreviates(tdet,            t_determinize).
abbreviates(t_det,           t_determinize).
abbreviates(tdeterminize,    t_determinize).
abbreviates(t_determinize,   t_determinize).
abbreviates(wd,              w_determinize).
abbreviates(w_d,             w_determinize).
abbreviates(wdet,            w_determinize).
abbreviates(w_det,           w_determinize).
abbreviates(wdeterminize,    w_determinize).
abbreviates(w_determinize,   w_determinize).
abbreviates(wtd,             wt_determinize).
abbreviates(wt_d,            wt_determinize).
abbreviates(wtdet,           wt_determinize).
abbreviates(wt_det,          wt_determinize).
abbreviates(wtdeterminize,   wt_determinize).
abbreviates(wt_determinize,  wt_determinize).
abbreviates(aux,             aux_file).
abbreviates(auxfile,         aux_file).
abbreviates(aux_file,        aux_file).
abbreviates(pm_file,         pm_file).
abbreviates(pm,              pm_file).
abbreviates(auxrc,           aux_file_rc).
abbreviates(auxfilerc,       aux_file_rc).
abbreviates(aux_filerc,      aux_file_rc).
abbreviates(aux_rc,          aux_file_rc).
abbreviates(auxfile_rc,      aux_file_rc).
abbreviates(aux_file_rc,     aux_file_rc).
abbreviates(v,               verbose).
abbreviates(verbose,         verbose).
abbreviates(minimum_path,    minimum_path).
abbreviates(min_path,        minimum_path).
abbreviates(identical,       identical).

abbreviates(l,l).
abbreviates(cmd,cmd).
abbreviates(interactive,interactive).
abbreviates(vv,vv).
abbreviates(vvv,vvv).
abbreviates(vvvv,vvvv).
abbreviates(vvvvv,vvvvv).


user:help_info(class,option,"Command-line Arguments","
Usage: fsa [Flag=Val]* [ActionOption]

The fsa program can be started with command-line arguments
(options). A command-line consists of a number of global variable
assignments, following by (at most one) *action* option, followed by
more global variable assignments. If no action option is provided,
then the system runs in interactive mode. If the variable
*interpreter* is set to *on*, then fsa runs in interactive mode after
the action indicated by the action option has been performed.  If the
flag *interactive* has been set to *cmdint*, then fsa runs the FSA
command interpreter.  Otherwise, you get the ordinary SICStus Prolog
prompt.  

Typical actions that can be performed through the use of an action
option are:

* regular expression operations such as kleene closure,
complementation for given automata
* determinization and minimization of automata
* construction of automaton on the basis of regular expression
* visualization of given automaton
* apply automaton to a string or set of strings

").

user:help_info(option,aux,"-aux Aux",
"Aux is a file containing auxiliary regular expression operator definitions.
It is loaded into module fsa_regex_aux, and will be used for compiling 
regular expressions. 

Note that your file with definitions of regular expression operators
is consulted with the special Prolog-syntax operators for regular
expression notation loaded. Thus you can use * .. & etc. in your
definitions. Drawback is that you cannot use operator notation for
e.g. the is/2 predicate.  

A typical auxiliary definition will be:

	 macro(vowel,{a,e,i,o,u}).

A slightly more interesting one:

	 macro(free(Expr), ~ $ Expr).

You can also explicitly construct an automaton yourself, e.g.:

	 rx(my_operator(Expr),Fa) :-
	     fsa_regex:rx(Expr,Fa0),
	     my_operator_definition(Fa0,Fa).

so you can call fsa_regex:rx/2 for further compilations.

There can be multiple -aux options.
 ").

user:help_info(option,pm,"-pm File",
"File is supposed to contain the definition of a predicate module. The file
is loaded and moreover the global variable *pred_module* is set to the
name of the module defined in that file. There can be multiple -pm
options").

user:help_info(option,l,"-l File",
"The File is loaded, using use_module/1. There can be multiple
-l options.").

user:help_info(option,cmd,"-cmd Goal","evaluates Prolog Goal; Goal is parsed as Prolog term. Example:

fsa -cmd 'listing(library_directory),halt').

There can be multiple -cmd options
").


user:help_info(option,cmdint,"-cmdint","
Run interactively with the FSA6 command interpreter.").

fsa_command(cmdint,true) -->
    {  fsa_globals:fsa_global_set(interactive,cmdint) }.

user:help_info(option,accepts,"-a[ccepts] [In] String","
This option can be used to test a given string for acceptance
by an automaton read from In (or standard input). Fsa prints `yes'
or `no' to standard error. Example:

        % fsa -r 'a+' | fsa -a aaa

Prints `yes'.
").
fsa_command(accepts, fsa_interpreter:fsa_accepts_file(T,File)) -->
    input_file_or_user(File),
    string(T).

user:help_info(option,approx,"-approx [In] String","
This option can be used to get all best matches for a given string 
and an automaton read from In (or standard input). In must contain an
automaton.

        % fsa -r '[a,a]+' | fsa -approx aaa

Prints:

        aa
        aaaa
").

fsa_command(approx, fsa_interpreter:fsa_approx_file(File,T)) -->
    input_file_or_user(File),
    string(T).

user:help_info(option,fsa2fsm,"-fsa2fsm In Syms Aut | -fsa2fsm [In [Out]]","
If three file names are given, then the automaton read from the first
file is converted to an automaton in AT&T's *fsm* software
format. That automaton is written into Aut; Syms will contain
a mapping from the integers used in Aut to the actual symbols. If
less than three file names are given, then it is assumed that the
actual symbols can be ignored and no Syms is written. 
").

fsa_command(fsa2fsm, fsa_io:fsa2fsm_file(Fsa,Symbols,Fsm)) -->
    input_file(Fsa),
    output_file(Symbols),
    output_file(Fsm).

fsa_command(fsa2fsm, fsa_io:fsa2fsm_file(Fsa,Fsm)) -->
    input_output(Fsa,Fsm).

user:help_info(option,fsm2fsa,"-fsm2fsa [In [Out]]","
Converts an automaton in AT&T's *fsm* software format into fsa5
format. Note that a separate symbol definition file is currently not
supported. ").

fsa_command(fsm2fsa, fsa_io:fsm2fsa_file(Fsm,Fsa)) -->
    input_output(Fsm,Fsa).

user:help_info(option,compile,"-c[ompile] [In [Out]]","
For a given automaton read from file In, a Prolog program
is written to Out. For details, see the module fsa_compiler.").  

fsa_command(compile, fsa_compiler:fsa_compile_to_prolog(User,User2)) -->
    input_output(User,User2).

user:help_info(option,compile_to_c,"-c[ompile_to_]c [In [Out]]","
For a given automaton read from In a C program is written to
Out. For details, see module compiler_to_c.").

fsa_command(compile_to_c, fsa_compiler_to_c:fsa_compile_to_c(User,User2)) --> 
    input_output(User,User2).

user:help_info(option,java,"-java [In] Out",
"For a given automaton read from In, a JAVA program is written to
Out. For details, see module fsa_java.").

fsa_command(java,fsa_java:fsa_java(In,Out)) --> input_output(In,Out).
	       
user:help_info(option,cpp,"-cpp [In] Out",
"For a given automaton read from In, a C++ program is written to
Out. For details, see module fsa_cpp.").

user:help_info(option,'c++',"-c++ [In] Out",
"For a given automaton read from In, a C++ program is written to
Out. For details, see module fsa_cpp.").

fsa_command(cpp,fsa_cpp:fsa_cpp(In,Out)) --> input_output(In,Out).
	       
user:help_info(option,compose,"-compose A B [Out]","
The transducers read from A and B are composed, and the result
is written to Out. Equivalent to 

        fsa -r 'compose(file(A),file(B))' >Out
").
fsa_command(compose,
	    fsa_regex:fsa_regex_compile_file(compose(file(File0),file(File1)),User)) --> 
    input_file(File0),
    input_file(File1),
    output_file_or_user(User).

user:help_info(option,complement,"-complement [In [Out]]","
The complementation operator is applied to the automaton read from
In, and the result is written to Out.").

fsa_command(complement,
	    fsa_regex:fsa_regex_compile_file(complement(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,count,"-count [In [Out]]","
For the automaton read from In the number of transitions and symbols and
some other properties is written to Out. If *-short* is
specified, then the output is given as a single line consisting of the 
number of states, start states, final states, transitions, jumps, and
symbols respectively. Otherwise a more elaborate message is printed
meant for human consumption.").

fsa_command(count, fsa_data:fsa_count_file(User,User2)) -->
    option(short),
    {  fsa_globals:fsa_global_set(count,short) },
    input_output(User,User2).

fsa_command(count, fsa_data:fsa_count_file(User,User2)) -->
    input_output(User,User2).

user:help_info(option,density,"-density [In [Out]]","
For the automaton read from In various densities are reported to
Out. Deterministic density is the number of transition
divided by the number of states times the number of symbols; absolute
density is the number of transitions dividided by the number of 
states squared times the number of symbols. Jump density is the number 
of jumps dividied by the squared number of states. Deterministic
density can be used to characterize the difficulty of
determinization. For deterministic densities of around 2, exponential
blow-up of the output (and hence processing time) can be expected
(Leslie 1995). Jump density can be used to estimate the most efficient 
subset construction algorithm (van Noord 1998).").

fsa_command(density, fsa_data:fsa_density_file(User,User2)) -->
    input_output(User,User2).

user:help_info(option,davinci,"-davinci [In [Out]]","
For the automaton read from In a corresponding DaVinci term is written 
to Out. This can be used to visualize the automaton In using DaVinci:

        fsa -davinci a.nd > a.davinci
        daVinci a.davinci

").

fsa_command(davinci,fsa_visualization:fsa_davinci(User,User2)) -->
    input_output(User,User2).

user:help_info(option,vcg,"-vcg [In [Out]]","
For the automaton read from In a corresponding vcg term is written 
to Out. This can be used to visualize the automaton In using (x)vcg:

        fsa -vcg a.nd | xvcg -

").

fsa_command(vcg, fsa_visualization:fsa_vcg(User,User2)) -->
    input_output(User,User2).

user:help_info(option,dot,"-dot [In [Out]]","
For the automaton read from In a corresponding dot term is written 
to Out. This can be used to visualize the automaton In using dot or
dotty:

        fsa -dot a.nd | dotty -
        fsa -dot a.nd | dot -Tps | gv -
        fsa -dot a.nd | dot -Tgif | xv -

").

fsa_command(dot, fsa_visualization:fsa_dot(User,User2)) -->
    input_output(User,User2).

user:help_info(option,determinize,"
-d[eterminize] [In [Out]] | -dgraph [In [Out]] -drgraph [In [Out]]
-dsubset [In [Out]] | -dstate [In [Out]]","
The automaton read from In is determinized and written to Out. FSA6
supports four variants of the determinization algorithm. In the
first form, a heuristic is used (based on the jump density) to
select the variant of the determinization variant. The other
forms indicate the particular variant that is to be used. For details, 
refer to the *fsa_determinizer* module. 
").

fsa_command(determinize,fsa_regex:fsa_regex_compile_file(determinize(file(User)),User2)) -->
    input_output(User,User2).

fsa_command(dgraph,
	    fsa_regex:fsa_regex_compile_file(determinize(file(User),per_graph),User2)) -->
    input_output(User,User2).

fsa_command(dreachgraph,
	    fsa_regex:fsa_regex_compile_file(determinize(file(User),per_reachable_graph),User2)) -->
    input_output(User,User2).

fsa_command(dcreachgraph,
	    fsa_regex:fsa_regex_compile_file(determinize(file(User),per_co_reachable_graph),User2)) -->
    input_output(User,User2).

fsa_command(dsubset,
	    fsa_regex:fsa_regex_compile_file(determinize(file(User),per_subset),User2)) -->
    input_output(User,User2).

fsa_command(dstate,
	    fsa_regex:fsa_regex_compile_file(determinize(file(User),per_state),User2)) -->
    input_output(User,User2).

user:help_info(option,efree,"-efree [In [Out]]","
For the automaton read from In an equivalent automaton without any
epsilon transitions (jumps) is written to Out.
").

fsa_command(efree,fsa_regex:fsa_regex_compile_file(efree(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,ignore,"-ignore A B [Out]","
Equivalent to:

        fsa -r 'ignore(file(A),file(B))' > Out
").
fsa_command(ignore, fsa_regex:fsa_regex_compile_file(ignore(file(File0),file(File)),User2)) -->
    input_file(File0),
    input_file(File),
    output_file_or_user(User2).

user:help_info(option,difference,"-diff[erence] A B [Out]","
Equivalent to:

        fsa -r 'difference(file(A),file(B))' > Out
").

fsa_command(difference,
	    fsa_regex:fsa_regex_compile_file(difference(file(File0),file(File)),User2)) -->
    input_file(File0), 
    input_file(File),
    output_file_or_user(User2).

user:help_info(option,accept_all,"-aa In | -accept_all In | -raa Regex","
The program checks each string read from standard input for
acceptance by the automaton read from In. Depending on the type of the 
automaton, the system reports the transductions for each string, or
simply `yes' or `no' (for recognizers). In the third
form the recognizer is specified by regular expression Regex rather 
than by an automaton.").

fsa_command(accept_all,fsa_interpreter:fsa_accept_all(File,user,user)) -->
    input_file(File).

fsa_command(r_accept_all,
	    fsa_interpreter:fsa_regex_accept_all(Regex,user,user)) -->
    [Regex].

fsa_command(t_accept_all,
	    fsa_interpreter:fsa_regex_accept_all_t(Regex,user,user)) -->
    [Regex].

user:help_info(option,prolog,"-prolog Goal","
Evaluates Prolog goal. Example:

        fsa -prolog 'listing(user:file_search_path).'
").

fsa_command(prolog,user:Term) -->
    [CommandAtom],
    { fsa_util:atom_term(CommandAtom,Term) }.

user:help_info(option,generate,"-generate States Syms Dens [JDens]","
This option is used to generate random finite automata, using the
algorithm of Leslie 1995. States is the number of states, Syms is the
number of symbols, Dens is absolute density, and JDens is the jump
density. ").
fsa_command(generate, fsa_regex:fsa_regex_compile_file(random(St,Syms,Dens,JDens),User)) -->
    [St0], [Syms0], [Dens0], [JDens0],
    {
     fsa_util:number__atom(St,St0), integer(St),
     fsa_util:number__atom(Syms,Syms0), integer(Syms),
     atom_codes(Dens0,DensChars),
     number_codes(Dens,DensChars), Dens >=0, Dens =<1,     atom_codes(JDens0,JDensChars),
     number_codes(JDens,JDensChars), JDens >=0, JDens =<1
    },
    output_file_or_user(User).

fsa_command(generate,fsa_regex:fsa_regex_compile_file(random(St,Syms,Dens,0),User)) -->
    [St0], [Syms0], [Dens0],
    {
     fsa_util:number__atom(St,St0), integer(St),
     fsa_util:number__atom(Syms,Syms0), integer(Syms),
     atom_codes(Dens0,DensChars),
     number_codes(Dens,DensChars), Dens >=0, Dens =<1
    },
    output_file_or_user(User).

fsa_command(generate,fsa_regex:fsa_regex_compile_file(random,User)) -->
    output_file_or_user(User).

user:help_info(option,intersect,"-intersect A B [Out]","
Equivalent to:

        fsa -r 'intersect(file(A),file(B))' > Out
").

fsa_command(intersect,
	    fsa_regex:fsa_regex_compile_file(intersect(file(File0),file(File1)),User2)) --> 
    input_file(File0),
    input_file(File1),
    output_file_or_user(User2).

user:help_info(option,kleene_star,"-kleene_star [In [Out]]","
Equivalent to:

        fsa -r 'file(In)*' > Out
").
fsa_command(kleene_star,
	    fsa_regex:fsa_regex_compile_file(kleene_star(file(User)),User2)) --> 
    input_output(User,User2).

user:help_info(option,minimum_path,"-minimum_path [In]","
Writes the path with the minimum weigth of the automaton
read from In.").
fsa_command(minimum_path,
	    fsa_minimum_path:fsa_minimum_path_file(User)) -->
    input_file_or_user(User).

user:help_info(option,kleene_plus,"-kleene_plus [In [Out]]","
Equivalent to:

        fsa -r 'file(In)+' > Out
").
fsa_command(kleene_plus, fsa_regex:fsa_regex_compile_file(kleene_plus(file(User)),User2)) --> 
    input_output(User,User2).

user:help_info(option,reverse,"-reverse [In [Out]]","
Equivalent to:

        fsa -r 'reverse(file(In))' > Out
").
fsa_command(reverse,fsa_regex:fsa_regex_compile_file(reverse(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,inverse,"-inverse [In [Out]]","
Equivalent to:

        fsa -r 'inverse(file(In))' > Out
").
fsa_command(inverse, fsa_regex:fsa_regex_compile_file(inverse(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,domain,"-domain [In [Out]]","
Equivalent to:

        fsa -r 'domain(file(In))' > Out
").
fsa_command(domain,fsa_regex:fsa_regex_compile_file(domain(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,range,"-range [In [Out]]","
Equivalent to:

        fsa -r 'range(file(In))' > Out
").
fsa_command(range,fsa_regex:fsa_regex_compile_file(range(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,cleanup,"-cleanup [In [Out]]","
Equivalent to:

	fsa -r 'cleanup(file(In))' > Out
").

fsa_command(cleanup,fsa_regex:fsa_regex_compile_file(cleanup(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,identity,"-identity [In [Out]]","
Equivalent to:

        fsa -r 'identity(file(In))' > Out
").
fsa_command(identity,fsa_regex:fsa_regex_compile_file(identity(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,option,"-option [In [Out]]","
Equivalent to:

        fsa -r 'option(file(In))' > Out
").
fsa_command(option,fsa_regex:fsa_regex_compile_file(option(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,union,"-union A B [Out]]","
Equivalent to:

        fsa -r 'union(file(A),file(B))' > Out
").
fsa_command(union,fsa_regex:fsa_regex_compile_file(union(file(FileA),file(FileB)),User2)) -->
    input_file(FileA),
    input_file(FileB),
    output_file_or_user(User2).

user:help_info(option,concat,"-concat A B [Out]]","
Equivalent to:

        fsa -r 'concat(file(A),file(B))' > Out
").
fsa_command(concat,fsa_regex:fsa_regex_compile_file(concat(file(FileA),file(FileB)),User2)) -->
    input_file(FileA), 
    input_file(FileB),
    output_file_or_user(User2).

user:help_info(option,minimize,
"-m[inimize] [In [Out]] | -mb [In [Out]] | -mh [In [Out]]",
"Minimizes the automaton read from In. The first version uses the
default minimization algorithm (by Brzozwski). The other options
explicitly require the algorithms by, respectively, Brzozowski or 
Hopcroft. Refer to the fsa_regex module and the fsa_minimizer module
for details.").

fsa_command(minimize, fsa_regex:fsa_regex_compile_file(minimize(file(User)),User2)) -->
    input_output(User,User2).

fsa_command(mb, fsa_regex:fsa_regex_compile_file(mb(file(User)),User2)) -->
    input_output(User,User2).

fsa_command(mh, fsa_regex:fsa_regex_compile_file(mh(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,t_minimize,"-t_m[inimize] [In [Out]]","
Applies the minimization algorithm for transducers (by Mohri) to the
transducer read from In. Refer to the fsa_t_determinizer module for
details. ").
fsa_command(t_minimize, fsa_regex:fsa_regex_compile_file(t_minimize(file(User)),User2)) -->
    input_output(User,User2).

fsa_command(wt_minimize, fsa_regex:fsa_regex_compile_file(wt_minimize(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,produce,"-produce [In [Out]]","
For the recognizer read from In strings accepted by In are written to
Out. ").
fsa_command(produce, fsa_interpreter:fsa_produces(User,User2)) --> 
    input_output(User,User2).

user:help_info(option,sample,"-sample [In [Out]]","
For the automaton read from In strings or string pairs accepted by In are written to Out, using a sampling procedure based on the weights in In. If In is not a weighted automaton, then uniform weights are assumed. The size of the sample is determined by the flag nr_sol_max. Examples:

	fsa -r '[a::0.3*,b::5*]' |
	      fsa nr_sol_max=10000 -sample |sort |uniq -c | sort -nr


produces:

   5331 []
   2519 a
   1028 aa
    456 aaa
    187 aaaa
    172 b
    108 aaaaa
     73 ab
     32 aaaaaa
     28 aab
     24 aaaaaaa
     17 aaab
      8 aaaaaaaa
      5 bb
      3 abb
      3 aaaab
      3 aaaaab
      2 aaaaaaaaa
      1 aabb


	      
        fsa -r '[a*,b*]' |
	       fsa nr_sol_max=100 -sample |sort |uniq -c | sort -nr

produces:

     32 []
     16 b
     11 a
      9 bb
      5 abb
      5 ab
      3 bbbb
      3 bbb
      3 aa
      2 bbbbb
      2 abbbb
      2 aaa
      1 bbbbbbb
      1 abbbbbb
      1 abbb
      1 aab
      1 aaabb
      1 aaaabbb
      1 aaaaaab

	      
	      ").

fsa_command(sample, fsa_interpreter:fsa_sample(User,User2)) --> 
    input_output(User,User2).

user:help_info(option,regex,"-r[egex] [Regex] [Out]","
The automaton described by regular expression Regex are written to
Out. If Regex is not specified, it is read from standard input.").
fsa_command(regex,fsa_regex:fsa_regex_atom_compile_file(Expr,User2)) --> 
    [Expr],
    output_file_or_user(User2).

fsa_command(regex,fsa_regex:fsa_regex_read_compile_file(user)) -->
    [].

user:help_info(option,tk,"-tk [File] | -tk [-r Regex]","
Starts the graphical user interface on the automaton in File, or the
automaton defined by the regular expression Regex.
").
fsa_command(tk, fsa_tk:tk_fsa_file(File)) --> 
    input_file(File),
    {  fsa_globals:fsa_global_set(interactive,on) }.

fsa_command(tk, fsa_tk:tk_regex(Atom)) -->
    option(R), {  abbreviates(R,regex)  },
    [Atom],
    {  fsa_globals:fsa_global_set(interactive,on) }.

fsa_command(tk, fsa_tk:tk_fsa_read_regex) -->
    option(R), {  abbreviates(R,regex)  },
    {  fsa_globals:fsa_global_set(interactive,on) }.

fsa_command(tk, fsa_tk:tk_rx) -->
    {  fsa_globals:fsa_global_set(interactive,on) }.

user:help_info(option,postscript,"-postscript [In [Out]]","
Produces postscript version of the automaton read from In.
").
fsa_command(postscript, fsa_visualization:fsa_postscript(User,User2)) -->
    input_output(User,User2).

user:help_info(option,fsa_dict_to_perfect_hash,"dict2ph [In [Out]]","
A minimal string-to-weight transducer will be written to Out,
transducing each of the lines read from In into its rank in alphabetic 
ordering; in other words, the transducer computes a perfect hash for 
the keys read from In. For more info, see module *fsa_dict*.").

fsa_command(fsa_dict_to_perfect_hash,
	    fsa_dict:fsa_dict_to_perfect_hash_file(User,User2)) -->
    input_output(User,User2).

user:help_info(option,fsa_dict_to_fsa,"dict2m [In [Out]]","
A minimal recognizer will be written to Out, recognizing each of the
lines read from In.").

fsa_command(fsa_dict_to_fsa,fsa_dict:fsa_dict_to_fsa_file(User,User2)) -->
    input_output(User,User2).

fsa_command(pstricks,fsa_visualization:fsa_pstricks(User,User2)) -->
    input_output(User,User2).

user:help_info(option,pstricks,
"-pstricks_tex [In [Out]] | -pstricks_picture [In [Out]]","
Produces LaTeX code using PsTricks macro's for the automaton read from 
In. In the first variant a self-contained LaTeX document is produced;
in the second variant a LaTeX picture is produced to be included in
another document.").
fsa_command(pstricks_tex,fsa_visualization:fsa_pstricks_tex(User,User2)) -->
    input_output(User,User2).

fsa_command(pstricks_picture,fsa_visualization:fsa_pstricks_picture(User,User2)) -->
    input_output(User,User2).

user:help_info(option,copy,"-copy [In [Out]]","
Copies the automaton from In to Out. Useful to convert between
different formats using the *read* and *write* global variables:

        fsa read=fast write=normal -copy a.nd b.nd
").
fsa_command(copy,fsa_io:fsa_copy_file(User,User2)) -->
    input_output(User,User2).

fsa_command(dict,fsa_dict:fsa_dict_to_trie_file(User,User2)) -->
    input_output(User,User2).

user:help_info(option,prefix_tree,"-prefix_tree [In [Out]]",
"A so-called prefix tree will be written to Out, which is a weighted
recognizer of all the lines read from In, where the shape of the recognizer
is the corresponding trie, and the weights are derived from the counts
that each transition is used in In. Cf. Carrasco and Oncina, 1999.").

user:help_info(option,dict,"-dict [In [Out]]",
"A so-called trie will be written to Out, which is a deterministic
recognizer of all the lines read from In, where no states have more than
a single incoming transition.").

fsa_command(prefix_tree,fsa_dict:fsa_dict_to_weighted_trie_file(User,User2)) -->
    input_output(User,User2).

user:help_info(option,t_determinize,"-t_d[eterminize] [In [Out]]","
The determinization algorithm for transducers (by Mohri) is applied to 
string-to-string transducer In and yields Out. For details, refer to
module t_determinizer."). 

fsa_command(t_determinize,
	    fsa_regex:fsa_regex_compile_file(t_determinize(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,w_determinize,"-w_d[eterminize] [In [Out]]","
The determinization algorithm for transducers (by Mohri) is applied to 
string-to-weight transducer In and yields Out. For details, refer to
module t_determinizer."). 

fsa_command(w_determinize,
	    fsa_regex:fsa_regex_compile_file(w_determinize(file(User)),User2)) -->
    input_output(User,User2).

fsa_command(wt_determinize,
	    fsa_regex:fsa_regex_compile_file(wt_determinize(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,w_minimize,"-w_m[inimize] [In [Out]]","
The minimization algorithm for transducers (by Mohri) is applied to 
string-to-weight transducer In and yields Out. For details, refer to
module t_determinizer."). 

fsa_command(w_minimize,fsa_regex:fsa_regex_compile_file(w_minimize(file(User)),User2)) -->
    input_output(User,User2).

user:help_info(option,identical,"-identical In1 In2",
"Reports to standard output whether the two automata read from In and In2
are identical. The algorithm used to determine this first determines both
automata, and then renames the states of the automata in a canonical way.
The automata are identical in case their canonical representations are 
identical. Note that two non-identical automata might still be equivalent
in the sense that they define the same language or the same relation.").

fsa_command(identical,fsa_data:fsa_identical_files_report(A,B)) -->
    input_file(A),
    input_file(B).
	       
string(String) -->
    [Atom],
    {  atom_codes(Atom,String) }.

parse_options -->
    parse_option,
    parse_options.

parse_options -->
    [].

parse_option -->
    option(Option0),
    { abbreviates(Option0,Option) },
    extra_option(Option).

parse_option -->
    flag_option.

extra_option(aux_file) -->
    [File],
    { fsa_regex:fsa_load_aux_file(File) }.

extra_option(aux_file_rc) -->
    [File],
    { fsa_regex:fsa_reconsult_aux_file(File) }.

extra_option(l) -->
    [File],
#sic#     { user:compile(File) }.
#yap#     { user:compile(File) }.
#swi#     { user:consult(File) }.

extra_option(pm_file) -->
    [File],
    {
#!swi#        use_module(Module,File,[]),
#swi#        consult(File), 
#swi#        absolute_file_name(File,[file_type(prolog),
#swi#                                       access(read)],AbsFile), 
#swi#        current_module(Module,AbsFile),
       fsa_globals:fsa_global_set(pred_module,Module)
    }.

extra_option(cmd) -->
    [CommandAtom],
    { fsa_util:atom_term(CommandAtom,Term),
      call(user:Term)
    }.

extra_option(interactive) -->
    {  fsa_globals:fsa_global_set(interactive,on)  }.

extra_option(verbose) -->
    { fsa_globals:fsa_global_set(debug,1) }.

extra_option(vv) -->
    { fsa_globals:fsa_global_set(debug,2) }.

extra_option(vvv) -->
    { fsa_globals:fsa_global_set(debug,3) }.

extra_option(vvvv) -->
    { fsa_globals:fsa_global_set(debug,4) }.

extra_option(vvvvv) -->
    { fsa_globals:fsa_global_set(debug,5) }.

option(Flag) -->
    [Flag0],
    { atom_codes(Flag0,[45|Chars]),
      atom_codes(Flag,Chars)
    }.

flag_option -->
    option(flag),
    [Flag],
    [Val],
    { fsa_globals:fsa_global_set(Flag,Val),
      format(user_error,"~w=~w~n",[Flag,Val])
    }.

flag_option -->
    [FlagIsVal],
    { catch(fsa_util:atom_term(FlagIsVal,Flag=Val), 
            syntax_error(_,_,_,_,_),
            fail
	   ),
      fsa_globals:fsa_global_set(Flag,Val),
      format(user_error,"~w=~w~n",[Flag,Val])
    }.

