:- module(fsa_semiring, []).

#swi# :- style_check([-atom]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.


user:help_info(class,semiring,"Types of transducers",
"The determinization, minimization and minimum path algorithms for
transducers are implemented in a fully general way, i.e., for various
types of transducers (`semirings').  

For each supported type, a number of predicates must be defined in a 
corresponding module (these are called `semiring declarations'):

* zero(Val).
* fsa_plus(Val0,Val1,Sum).
* minus(Val0,Val1,Diff).
* minimum(Val0,Val1,MinVal).
* minimum_only(YesNo).

Currently, the following types of transducers are supported:

* fsa_strings (ordinary (string-to-string) transducers)
* fsa_weights (weighted recognizers, aka string-to-weight transducers)
* fsa_weighted_strings (weighted transducers, aka
  string-to-weighted-string transducer).

The examples directory SemiringModules might contain additional 
semiring declarations.  ").


user:help_info(semiring,zero,"zero(?Val).",
"The identity element for addition. For strings, this is the empty
string; for weights it is 0.").

user:help_info(semiring,fsa_plus,"fsa_plus(+Val0,+Val1,?Sum).",
"Addition. For weights this is number addition, for strings this is
concatenation.").

user:help_info(semiring,minus,"minus(+Val0,+Val1,?Diff).",
"Inverse of the addition operator. ").

user:help_info(semiring,minimum,"minimum(+Val0,+Val1,?Min).",
"Minimum value of two given values. For weights this is the minumum of
two numbers, for strings this is the longest common prefix.").

user:help_info(semiring,minimum_only,"minimum_only(+YesNo).",
"YesNo is one of the atoms

* yes
* no

indicating whether we are intested in all outputs associated with a
path or only in the minimal output. For weights this is `yes', for
strings this is `no'.").

zero(Val,A) :-
    (	Val==fsa_strings -> fsa_strings:zero(A)
    ;   Val==fsa_weights -> fsa_weights:zero(A)
    ;   Val==fsa_weighted_strings -> fsa_weighted_strings:zero(A)
    ;   Val:zero(A)
    ).


fsa_plus(Val,A,B,C) :-
    (	Val==fsa_strings -> fsa_strings:fsa_plus(A,B,C)
    ;   Val==fsa_weights -> fsa_weights:fsa_plus(A,B,C)
    ;   Val==fsa_weighted_strings -> fsa_weighted_strings:fsa_plus(A,B,C)
    ;   Val:fsa_plus(A,B,C)
    ).

minus(Val,A,B,C) :-
    (	Val==fsa_strings -> fsa_strings:minus(A,B,C)
    ;   Val==fsa_weights -> fsa_weights:minus(A,B,C)
    ;   Val==fsa_weighted_strings -> fsa_weighted_strings:minus(A,B,C)
    ;   Val:minus(A,B,C)
    ).

minimum(Val,A,B,C) :-
    (	Val==fsa_strings -> fsa_strings:minimum(A,B,C)
    ;   Val==fsa_weights -> fsa_weights:minimum(A,B,C)
    ;   Val==fsa_weighted_strings -> fsa_weighted_strings:minimum(A,B,C)
    ;   Val:minimum(A,B,C)
    ).

% minimal_path_minimum(Val,A,B,C) :-
%     (	Val==fsa_strings -> fsa_strings:minimal_path_minimum(A,B,C)
%     ;   Val==fsa_weights -> fsa_weights:minimal_path_minimum(A,B,C)
%     ;   Val==fsa_weighted_strings -> fsa_weighted_strings:minimal_path_minimum(A,B,C)
%     ;   Val:minimal_path_minimum(A,B,C)
%     ).

minimum_only(Val,A) :-
    (	Val==fsa_strings -> fsa_strings:minimum_only(A)
    ;   Val==fsa_weights -> fsa_weights:minimum_only(A)
    ;   Val==fsa_weighted_strings -> fsa_weighted_strings:minimum_only(A)
    ;   Val:minimum_only(A)
    ).


