%%% -*- Module: fsa_visualization; -*-
:- module(fsa_visualization,
	  [
#sic# 	    fsa_visualization/1,             % fsa_visualize
#sic# 	    fsa_visualization/2,
	    fsa_davinci/1,
	    fsa_davinci/2,
	    fsa_dot/1,
	    fsa_dot/2,
	    fsa_vcg/1,
	    fsa_vcg/2,
	    fsa_pstricks/2, fsa_pstricks/3, fsa_pstricks/4,
	    fsa_pstricks_picture/2, fsa_pstricks_tex/2,
	    fsa_postscript/2, fsa_postscript/3
	  ]).

#swi# :- style_check([-atom]).

#sic# :- use_module(library(system), [ exec/3,environ/2,mktemp/2 ]).
#sic# :- use_module(library(charsio),[ format_to_chars/3     ]).
#sic# :- use_module(library(lists),  [ member/2,append/3     ]).

#yap# :- use_module(library(charsio),[ format_to_chars/3     ]).
#yap# :- use_module(library(lists),  [ member/2,append/3     ]).

#swi# :- use_module(charsio,[ format_to_chars/3     ]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% compute_states %%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%% compute coordinates of the states
%% Coords is list of s(State,X,Y) triples.
compute_states(Coords,Fa) :-                 %2
    compute_states(Coords,Fa,[]).            %3

compute_states(Coords,Fa,Modules) :-         %3
    fsa_globals:fsa_global_get(v_algorithm,Alg),
    compute_states(Alg,Coords,Fa,Modules).   %4

#sic# compute_states(dot,Triples,Fa,_):-
#sic#     exec('dot -Grankdir=LR | dot2triples ',[pipe(Stream),pipe(Stream2),std],_Pid),
#sic#     current_output(Old),
#sic#     set_output(Stream),
#sic#     %% close Stream before trying to read Stream2!
#sic#     user:call_cleanup(fsa_visualization:fsa_dot(Fa),
#sic# 		      (	  close(Stream),
#sic# 			  set_output(Old)
#sic# 		      )),
#sic#     user:call_cleanup(read(Stream2,Triples),
#sic# 		      close(Stream2)
#sic# 		     ).

#sic# compute_states(tree,Coords,Fa,Modules) :-
#swi# compute_states(   _,Coords,Fa,Modules) :-
#yap# compute_states(   _,Coords,Fa,Modules) :-
    compute_states_tree(Modules,Coords,Fa).

compute_states_tree([],Coords,Fa) :- 
    findall(State,fsa_data:fsa_state(Fa,State),States0),
    order_states(States0,States,Fa),
    compute_tree(Fa,States,Trees,W),
    add_numbers(Trees,W,Coords,[]).

compute_states_tree([H|T],Coords,Fa) :-
    findall(State,fsa_data:fsa_state(Fa,State),States),
    recover_states(States,Coords,[H|T]),
    !.
compute_states_tree([_|_],Coords,Fa) :-
    compute_states_tree([],Coords,Fa).


recover_states([],[],_).
recover_states([State|States],[s(State,X,Y)|SStates],Modules) :-
    member(Module,Modules),
    fsa_tk:module_current_state(Module,State,X,Y,_,_,_,_),!,
    recover_states(States,SStates,Modules).

order_states(States0,States,Fa) :-
    shortest_path(States0,States,Fa).

shortest_path(States,Results,Fa) :-
    fsa_data:fsa_start_states(Fa,Starts),
    fsa_data:fsa_states_number(Fa,Size),
    fsa_array:fsa_array_new(Thing,Size),
    treat_start_states(Starts,Agenda,Thing),
    shortest_p(Agenda,Fa,Thing),
    results_p(States,Thing,Results0),
    keysort(Results0,Results).

treat_start_states([],[],_).
treat_start_states([Start|States],[0-Start|Ag],Thing) :-
    fsa_array:fsa_array_access(Start,0,Thing),
    treat_start_states(States,Ag,Thing).

results_p([],_,[]).
results_p([H|T0],Thing,[Key-H|T]) :-
    fsa_array:fsa_array_access(H,Key,-1,Thing),
    results_p(T0,Thing,T).

shortest_p([],_,_).
shortest_p([State|States0],Fa,Thing):-
    new_p(State,New,Fa),
    add_p(New,States0,States,Thing),
    shortest_p(States,Fa,Thing).

new_p(S0,Ss,Trans) :-
    findall(S,new_p0(S0,S,Trans),Ss0),
    sort(Ss0,Ss).

new_p0(Key0-State0,Key-State,Trans) :-
    transition_or_jump(Trans,State0,_,State),
    Key is Key0 + 1.

add_p([],S,S,_).
add_p([Key-H|T],S0,S,Thing) :-
    fsa_array:fsa_array_access(H,Key0,Thing),
    (	var(Key0)
    ->	Key0=Key, 
	add_p1(S0,Key-H,T,S,Thing)
    ;	add_p(T,S0,S,Thing)
    ).

add_p1([],H,T,[H|S],Thing) :-
    add_p(T,[],S,Thing).
add_p1([NKey-NQ|S0],Key-Q,T,S,Thing) :-
    (	Key < NKey
    ->	S=[Key-Q|S1],
	add_p(T,[NKey-NQ|S0],S1,Thing)
    ;	S=[NKey-NQ|S1],
	add_p1(S0,Key-Q,T,S1,Thing)
    ).

compute_tree(Fa,States,Trees,Width):-
    fsa_data:fsa_start_states(Fa,Starts0),
    findall(S,member(-1-S,States),Starts1),
    append(Starts0,Starts1,Starts),
    fsa_data:fsa_states_number(Fa,Size),
    fsa_array:fsa_array_new(Done,Size),
    ct_l(Starts,Trees,Fa,Done,0,Width,States).

ct_l([],[],_,_,W,W,_).
ct_l([H|T],Result,Fa,Done,W0,W,States) :-
    ct(H,Ht,Fa,Done,W1,States),
    (	Ht = r
    ->	Result=Tt
    ;	Result=[Ht|Tt]
    ),
    W2 is W0+W1,
    ct_l(T,Tt,Fa,Done,W2,W,States).

ct(State,Tree,Fa,Done,Max,States) :-
    fsa_array:fsa_array_access(State,Val,Done),
    (	var(Val) 
    ->	Val=State,
	Tree=tree(State,Max,Ds),
	ct_ds(State,Ds,Fa,Done,W,States),
	Max is max(W,1)
    ;	Tree=r, Max=0
    ).

ct_ds(State,Ds,Fa,Done,W,DistStates) :-
    findall(NState,transition_jump(Fa,State,NState,DistStates),States0),
    sort(States0,States),
    ct_l(States,Ds,Fa,Done,0,W,DistStates).

transition_jump(Fa,State0,State,States) :-
    fsa_globals:fsa_global_get(v_tree_depth,Val),
    (	Val == on
    ->	transition_jump_nc(Fa,State0,State)
    ;	transition_jump_c(Fa,State0,State,States)
    ).

transition_jump_c(Fa,State0,State,States) :-
    fsa_data:fsa_transition(Fa,State0,_,State),
    member(Dist0-State0,States),
    member(Dist-State,States),
    \+ Dist =< Dist0.
transition_jump_c(Fa,State0,State,States) :-
    fsa_data:fsa_jump(Fa,State0,State),
    member(Dist0-State0,States),
    member(Dist-State,States),
    \+ Dist =< Dist0.

transition_jump_nc(Fa,State0,State) :-
    fsa_data:fsa_transition(Fa,State0,_,State).
transition_jump_nc(Fa,State0,State) :-
    fsa_data:fsa_jump(Fa,State0,State).

add_numbers(Trees,Width,Coords0,Coords) :-
    fsa_globals:fsa_global_get(v_ycoord,Z),
    xdist(Delta),
    W is '*'(Width,Delta),
    Y is Z-(W/2),
    add_numbers(Trees,Width,Coords0,Coords,Z,Y,Delta).

add_numbers([],_,C,C,_,_,_).
add_numbers([tree(State,Width,Ds)|Tail],W,[s(State,X0,Y2)|C0],C,X0,Y0,Delta) :-
    Y1 is Y0 + '*'(Width,Delta),
    Y2 is Y0 + (Y1-Y0)/2,
    X1 is X0 + Delta,
    add_numbers(Ds,Width,C0,C1,X1,Y0,Delta),
    add_numbers(Tail,W,C1,C,X0,Y1,Delta).

%% default x distance between states
xdist(X) :-
    fsa_globals:fsa_global_get(v_xdist,X0),
    (	atom(X0)
    ->	%%fsa_util:atom_number(X,X0)
        atom_chars(X0,Xchars),
	number_chars(X,Xchars)
    ;	X= X0
    ).

transition_or_jump(Fa,P,Sym,Q) :-
    fsa_data:fsa_transition(Fa,P,Label,Q),
    fsa_data:fsa_symbols_decl(Fa,Decl),
    display_label(Decl,Label,Sym).
transition_or_jump(Fa,P,[],Q) :-
    fsa_data:fsa_jump(Fa,P,Q).

display_label(t(M1,M2,M3),A/(B/C),Label) :-
    !,
    display_label(t(M1,M2),A/B,SymA0),
    display_label(r(M3),C,Weight),
    (   Weight == 0
    ->  (   SymA0 = (L/R)
	->  Label = ':'(L,R)
	;   Label = SymA0
	)
    ;   (   SymA0 = (L/R)
	->  Label = ':'(L,':'(R,Weight))
	;   Label = '::'(SymA0,Weight)
	)
    ).
display_label(t(Mod,Mod),'$@'(L)/'$@'(L),Sym):-
    !,
    display_label0(L,Sym,Mod).
display_label(t(Mod,Mod),L/L,Sym):-
    fsa_pred_module:identity(Mod,L,L),
    !,
    display_label0(L,Sym,Mod).
display_label(t(Mod0,Mod1),L0/L1,Sym0/Sym1):-
    display_label0(L0,Sym0,Mod0),
    display_label0(L1,Sym1,Mod1).
display_label(r(Mod0,Mod1),L0/L1,Label) :-
    display_label0(L0,Sym0,Mod0),
    display_label0(L1,Sym1,Mod1),
    (   Sym1 == 0
    ->  Label = Sym0
    ;   Label = '::'(Sym0,Sym1)
    ).
    
display_label(r(Mod),L,Sym):-
    display_label0(L,Sym,Mod).

display_label0([],Sym,_):-
    !,
    Sym=[].
display_label0([H|T],Sym,Mod):-
    !,
    Sym=[SymH|SymT],
    display_label0(H,SymH,Mod),
    display_label0(T,SymT,Mod).

display_label0(S0,S,Module):-
    (	(   S0 = '$@'(S1)
	->  S3 = '$@'(S2)
	;   S0 = S1, S3=S2
	),
	fsa_pred_module:display_predicate(Module,S1,S2)
    ->	S3=S
    ;   S0=S
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% abbreviate_label  %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% the idea is to abbreviate [a,b,c,d,e,f,g,h,i,j,k,m,n,o,p] as [a-k,m-p]
%% moreover, to abbreviate [a/123.1289823749] as [a/123]
abbreviate_label(L0,L):-
    abbreviate_label0(L0,L1),
    output_ops(Ops),
    fsa_util:call_with_ops(Ops,fsa_visualization:term_atom_all(L1,L2)),
    fsa_util:concat_all(L2,L,' ').

output_ops([  op(395,xfy,(:)),
	      op(396,xfy,(::)),
	      op(398,fy,(`)),
	      op(397,xfx,(..))
	    ]).


term_atom_all([],[]).
term_atom_all([T0|Ts],[A0|As]) :-
    fsa_util:term_atom(T0,A0),
    term_atom_all(Ts,As).

abbreviate_label0([],[]).
abbreviate_label0([H|T],Syms) :-
    abbreviate_pairs([H|T],[H2|T2]),
    abbreviate_label0(T2,H2,H2,Syms,[]).

abbreviate_pairs([],[]).
abbreviate_pairs([H0|T0],[H|T]) :-
    (   H0 = A/B
    ->  H = ':'(A,B)
    ;   H=H0
    ),
    abbreviate_pairs(T0,T).

abbreviate_label0([],H,Begun,S0,S) :-
    a_to_b(Begun,H,S0,S).
abbreviate_label0([Next|Tail],H,Begun,S0,S) :-
    (	follows(Next,H)
    ->	abbreviate_label0(Tail,Next,Begun,S0,S)
    ;	a_to_b(Begun,H,S0,S1),
	abbreviate_label0(Tail,Next,Next,S1,S)
    ).

%% a_to_b(Begin,End,Result0,Result)
a_to_b(A,A,T0,T):-
    !,
    T0= [A|T].
a_to_b(A,B,T0,T) :-
    (	follows(B,A)
    ->	T0 = [A,B|T]
    ;	H = '..'(A,B), T0 = [H|T]
    ).

follows(A/A,B/B) :-
    !,
    follows(A,B).

follows(A,B) :-
    atom(A), atom(B),
    !,
    atom_codes(A,[Achar]),  az_or_09(Achar),
    atom_codes(B,[Bchar]),  az_or_09(Bchar),
    Bchar is Achar-1.

follows(A,B) :-
    integer(A), integer(B),
    A is B+1.

az_or_09(N) :-
    (   N >= 48, N =< 57
    ->  true
    ;   N >= 65, N =< 90
    ->  true
    ;   N >= 97, N =< 122
    ->  true
    ;   fail
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% escape_specials %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

escape_specials(In,Out,Patterns) :-
    (	atom(In)
    ->	atom_chars(In,InString),
	escape_specials0(InString,OutString,Patterns),
	atom_chars(Out,OutString)
    ;	In=Out
    ).

%% String0,String,In-OutList
escape_specials0([],[],_).
escape_specials0([H|T],OutList,Patterns):-
    (   apply_pattern([H|T],Rest,OutList,OutList0,Patterns)
    ->  true
    ;   OutList=[H|OutList0],
	Rest=T
    ),
    escape_specials0(Rest,OutList0,Patterns).

apply_pattern(List,Rest,OutList,OutList0,Patterns):-
    member(PatternIn-PatternOut,Patterns),
    append(PatternIn,Rest,List),
    append(PatternOut,OutList0,OutList).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% davinci %%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

user:help_info(xpred,fsa_davinci,"fsa_davinci(+File0,+File) fsa_davinci(+Fa)","
In the first variant, a representation accepted by the daVinci graph
visualization program is written to File on the basis of the automaton 
read from File0. In the second form, the representation for Fa is
written to standard output.").

fsa_davinci(Input,Output) :-
    fsa_io:fsa_read_file(Input,Fa),
    fsa_util:to_file(fsa_visualization:fsa_davinci(Fa),Output).

fsa_davinci(Fa0) :-
    limit_fsa_states_number(Fa0,Fa),
    fsa_data:fsa_start_states(Fa,Starts),
    fsa_data:fsa_states_number(Fa,Size),
    fsa_array:fsa_array_new(Done,Size),
    davinci_l(Starts,Trees,Fa,Done),
    write_term(Trees,[portrayed(true)]).
				% print(Trees).


davinci_l([],[],_,_).
davinci_l([H|T],Result,Fa,Done) :-
    davinci(H,Ht,Fa,Done),
    (	Ht = r(string(_))
    ->	Result=Tt
    ;	Result=[Ht|Tt]
    ),
    davinci_l(T,Tt,Fa,Done).

davinci(State,Tree,Fa,Done) :-
    fsa_array:fsa_array_access(State,Val,Done),
    (	var(Val)
    ->	Val=[], 
	Tree=l(string(State),n(string(State),
			       [a(string('OBJECT'),
				  string(State))|Atts],Edges)),
	state_attributes(State,Fa,Atts),
	(   davinci_transitions(Fa,State,Pairs)
	->  davinci_edges(Pairs,Edges,Fa,Done)
	;   Edges=[]
	)
    ;	Tree=r(string(State)) 
    ).

davinci_transitions(Fa,State,Pairs) :-
    findall(Sym/NState,extern_vis_transition(Fa,State,Sym,NState),Pairs0),
    sort(Pairs0,Pairs).

davinci_edges([],[],_,_Done).
davinci_edges([Sym/State|Tail],[R0|Rs],Fa,Done) :-
    davinci_edge(Sym,State,R0,Fa,Done),
    davinci_edges(Tail,Rs,Fa,Done).

davinci_edge(Sym,State,e(string(''),[],n(string(Sym),
					[a(string('OBJECT'),string(Sym)),
					 a(string('_GO'),string(text))|Atts],
                                         [e(string(''),[],Tree)])),
	Fa,Done) :-
    edge_attributes(Atts),
    davinci(State,Tree,Fa,Done).

:- multifile user:portray/1.

user:portray(string(X)) :-
    !,
    write('"'), write(X), write('"').

state_attributes(State,Fa,[a(string('FONTSTYLE'),
                                      string(italic))|Atts]) :-
    (	fsa_data:fsa_start_state(Fa,State)
    ->	(   fsa_data:fsa_final_state(Fa,State)
	->  start_final_attributes(Atts)
	;   start_attributes(Atts)
	)
    ;	(   fsa_data:fsa_final_state(Fa,State)
	->  final_attributes(Atts)
	;   simple_attributes(Atts)
	)
    ).

start_attributes([a(string('COLOR'),string(green)),
                  a(string('_GO'),string(rhombus))]).
final_attributes([a(string('COLOR'),string('#FF6347')),
		  a(string('_GO'),string(circle)),
                  a(string('BORDER'),string(double))]).
simple_attributes([a(string('COLOR'),string(lightblue)),
	          a(string('_GO'),string(circle))]).
start_final_attributes([a(string('COLOR'),string('#DCACDC')),
			a(string('BORDER'),string(double)),
	                a(string('_GO'),string(circle))]).
edge_attributes([]).

%%%%%%%%%%%%%%%%
%%%%% DOT %%%%%%
%%%%%%%%%%%%%%%%

user:help_info(xpred,fsa_dot,"fsa_dot(+File0,+File) fsa_dot(+Fa)"," 
In the first variant, a representation accepted by the dot / GraphViz
graph visualization program is written to File on the basis of the
automaton read from File0. In the second form, the representation for
Fa is written to standard output.").

fsa_dot(Input,Output) :-
    fsa_io:fsa_read_file(Input,Fa),
    fsa_util:to_file(fsa_visualization:fsa_dot(Fa),Output).

fsa_dot(Fa0) :-
    limit_fsa_states_number(Fa0,Fa),
    format(
'digraph "FSA Utilities" {
nodesep=0.3;
center=true;
',[]),
    findall(State,visual_state(Fa,State),States),
    dot_states(States,Fa),
    dot_transitions(Fa),
    format('}~n',[]).

%rankdir=LR;
%orientation=landscape;

dot_transitions(Fa) :-
    (	extern_vis_transition(Fa,State0,Sym,State),
	format('~d -> ~d [label="~w"];~n',
	       [State0,State,Sym]),
	fail
    ;	true
    ).

dot_states([],_).
dot_states([S0|S],Fa):-
    dot_state(S0,Fa),
    dot_states(S,Fa).

dot_state(State,Fa) :-
    (	fsa_data:fsa_start_state(Fa,State)
    ->	format('start~d [shape=plaintext,label=""];~n',[State]),
	format('start~d -> ~d [label=""];',[State,State]),
	(   fsa_data:fsa_final_state(Fa,State)
%	->  format('~d [style=filled,shape=diamond,color=orchid];~n',[State])
	->  format('~d [style=filled,shape=circle,color=orchid,peripheries=2];~n',[State])
%	;   format('~d [style=filled,shape=ellipse,color=olivedrab1];~n',[State])
	;   format('~d [style=filled,shape=circle,color=olivedrab1,peripheries=1];~n',[State])
	)
    ;	(   fsa_data:fsa_final_state(Fa,State)
%	->  format('~d [style=filled,shape=box,color=tomato1];~n',[State])
	->  format('~d [style=filled,shape=circle,color=tomato1,peripheries=2];~n',[State])
	;   format('~d [style=filled,shape=circle,color=lightblue1,peripheries=1];~n',[State])
	)
    ).

%%%%%%%%%%%%%
%%%% VCG %%%%
%%%%%%%%%%%%%

user:help_info(xpred,fsa_vcg,"fsa_vcg(+File0,+File) fsa_vcg(+Fa)"," 
In the first variant, a representation accepted by the vcg
graph visualization program is written to File on the basis of the
automaton read from File0. In the second form, the representation for
Fa is written to standard output.").

fsa_vcg(Input,Output) :-
    fsa_io:fsa_read_file(Input,Fa),
    fsa_util:to_file(fsa_visualization:fsa_vcg(Fa),Output).

fsa_vcg(Fa0) :-
    limit_fsa_states_number(Fa0,Fa),
    format('graph: {~n',[]),
    format('title: "FSA (automatically generated)"~n',[]),
    format('display_edge_labels:yes~n',[]),
    format('splines:yes~n',[]),
    format('color:white~n',[]),
    format('xspace:40~nyspace:20~n',[]),
    format('orientation:left_to_right~n',[]),
    format('node.color:cyan~n',[]),
    format('node.shape:ellipse~n',[]),
    findall(State,visual_state(Fa,State),States),
    vcg_states(States,Fa),
    vcg_transitions(Fa),
    format('}~n',[]).

vcg_transitions(Fa) :-
    (	extern_vis_transition(Fa,State0,Sym,State),
	format('edge: { sourcename: "~d" targetname: "~d" label: "~w" }~n',
	       [State0,State,Sym]),
	fail
    ;	true
    ).

vcg_states([],_).
vcg_states([S0|S],Fa):-
    vcg_state(S0,Fa),
    vcg_states(S,Fa).

vcg_state(State,Fa) :-
    format('node: { title:"~w" ',[State]),
    (	fsa_data:fsa_start_state(Fa,State)
    ->	(   fsa_data:fsa_final_state(Fa,State)
	->  format('color:lilac',[])
	;   format('color:green',[])
	)
    ;	(  fsa_data:fsa_final_state(Fa,State)
	->  format('color:red',[])
	;   true
	)
    ),
    format(' }~n',[]).

visual_state(Fa,State) :-
    fsa_globals:fsa_global_get(display_unused_states,Val),
    visual_state(Val,Fa,State).

visual_state(on,Fa,State) :-
    fsa_data:fsa_state(Fa,State).

visual_state(off,Fa,State) :-
    fsa_data:fsa_state(Fa,State),
    used(State,Fa).

used(State,Fa) :-
    (	fsa_data:fsa_transition(Fa,State,_,_)
    ;	fsa_data:fsa_transition(Fa,_,_,State)
    ;	fsa_data:fsa_jump(Fa,State,_)
    ;	fsa_data:fsa_jump(Fa,_,State)
    ;	fsa_data:fsa_start_state(Fa,State)
    ),
    !.

extern_vis_transition(Fa,State,Sym,NState) :-
    %% a real setof?
    setof(Sym0,transition_or_jump(Fa,State,Sym0,NState),Syms0),
    abbreviate_label(Syms0,Sym0),
    escape_specials(Sym0,Sym,[[34]-[92,34],[92]-[92,92]]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% pstricks %%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

user:help_info(xpred,fsa_pstricks_picture,"fsa_pstricks_picture(+File0,+File)","
A piece of LaTeX code with pstricks macro's which produces a picture
of the automaton read from File0 is written to File. This LaTeX
code is supposed to be included in a full LaTeX document. The global
variable pstricks_style influences the result.").

user:help_info(xpred,fsa_pstricks_tex,"fsa_pstricks_tex(+File0,+File)","
A standalone LaTeX document with pstricks macro's which produces a picture
of the automaton read from File0 is written to File. The global
variable pstricks_style influences the result.").

%% pstricks(InputFile,OutputFile)
%% for the current fsa (defined in Module) a TeX file OutputFile
%% is created...
fsa_pstricks(Input,Output) :-
    fsa_io:fsa_read_file(Input,Fa),
#!sic#    fsa_pstricks(Fa,Output,[]).
#sic#    fsa_pstricks(Fa,Output,[fsa_tk]).

fsa_pstricks_picture(Input,Output) :-
    fsa_io:fsa_read_file(Input,Fa),
#!sic#    fsa_pstricks(picture,Fa,Output,[]).
#sic#     fsa_pstricks(picture,Fa,Output,[fsa_tk]).

fsa_pstricks_tex(Input,Output) :-
    fsa_io:fsa_read_file(Input,Fa),
#!sic#    fsa_pstricks(tex,Fa,Output,[]).
#sic#     fsa_pstricks(tex,Fa,Output,[fsa_tk]).

fsa_pstricks(Fa,File,Overrule) :-
    fsa_pstricks(picture,Fa,File,Overrule).

fsa_pstricks(PicOrTex,Fa0,File,Mods):-
    limit_fsa_states_number(Fa0,Fa),
    fsa_util:to_file(fsa_visualization:fsa_pstricks0(PicOrTex,Fa,Mods),File),
    format(user_error,"TeX PsTricks saved in ~a~n",[File]).

fsa_pstricks0(PictureOrTex,Fa,OverruleModules):-
    compute_states(Coords,Fa,OverruleModules),
    begin(PictureOrTex,Coords,Xmin,Ymin,Ydist),
    tex_states(Coords,Fa,Xmin,Ymin,Ydist),
    tex_transitions(Coords,Fa),
    end(PictureOrTex).

declare_colors(fancy) :-
    format("
\\definecolor{tomato1}{rgb}{1, .39, .28}        %255 99 71
\\definecolor{plum2}{rgb}{.93, .68, .93}         %238 174 238
",[]).
declare_colors(plain).

begin(PictureOrTex,Coords,Xmin,Ymin,Ydist0) :-
    find_xys(Coords,Xdist0,Ydist0,Xmin,Ymin),
    Xdist is 2 + Xdist0/40,
    Ydist is 1 + Ydist0/40,
    prologue(PictureOrTex),
    fsa_globals:fsa_global_get(pstricks_style,Style),
    declare_colors(Style),
    format("
\\begin{pspicture}(-1,-1)(~f,~f)
%%\\newpsobject{showgrid}{psgrid}{subgriddiv=1,griddots=10,gridlabels=6pt}
%%\\showgrid
\\psset{unit=0.025cm}~n",[Xdist,Ydist]).

end(tex) :-
    format("\\end{pspicture}~n~n\\end{document}~n",[]).

end(picture) :-
    format("\\end{pspicture}~n~n%%\\end{document}~n",[]).


prologue(tex) :-
    format("% generated by FSA tools (c) 1995 - 1998 RUG, G van Noord~n",[]),
    format("
\\documentclass{article}
\\thispagestyle{empty}\\pagestyle{empty}
\\usepackage{pstricks,pstcol,pst-node}
\\begin{document}~n~n",[]).

prologue(picture) :-
    format("% generated by FSA tools (c) 1995 - 1998 RUG, G van Noord~n",[]),
    format("
%%\\documentclass{article}
%%\\thispagestyle{empty}\\pagestyle{empty}
%%\\usepackage{pstricks,pstcol,pst-node}
%%\\begin{document}~n~n",[]).

find_xys(Coords,Xdist,Ydist,Xmin,Ymin) :-
    findall(X/Y,xy_state(Coords,X,Y),Xys0),
    sort(Xys0,Xys),
    xys(Xys,Xmin,Xmax,Ymin,Ymax),
    Ydist is abs(Ymax-Ymin),
    Xdist is abs(Xmax-Xmin).

xys([X/Y|Tail],Xmin,Xmax,Ymin,Ymax) :-
    xys(Tail,X,X,Y,Y,Xmin,Xmax,Ymin,Ymax).

xys([],Xmin,Xmax,Ymin,Ymax,Xmin,Xmax,Ymin,Ymax).
xys([X/Y|Tail],Xmin0,Xmax0,Ymin0,Ymax0,Xmin,Xmax,Ymin,Ymax) :-
    xs(X,Xmin0,Xmax0,Xmin1,Xmax1),
    xs(Y,Ymin0,Ymax0,Ymin1,Ymax1),
    xys(Tail,Xmin1,Xmax1,Ymin1,Ymax1,Xmin,Xmax,Ymin,Ymax).

xs(X,Xmin0,Xmax0,Xmin1,Xmax1) :-
    (	X < Xmin0
    ->	Xmin1 = X
    ;	Xmin1 = Xmin0
    ),
    (	X > Xmax0
    ->	Xmax1 = X
    ;	Xmax1 = Xmax0
    ).

xy_state(Coords,X,Y) :-
    member(s(_,X,Y),Coords).

tex_states([],_,_,_,_).
tex_states([s(Symbol,X,Y)|Tail],Fa,Xmin,Ymin,Ydist) :-
    fsa_globals:fsa_global_get(pstricks_style,Style),
    tex_state(Style,X,Y,Symbol,Fa,Xmin,Ymin,Ydist),
    tex_states(Tail,Fa,Xmin,Ymin,Ydist).

tex_state(Style,X0,Y0,Symbol0,Fa,Xmin,Ymin,Ydist) :-
    X is X0-Xmin,
    Y is Ydist-(Y0-Ymin),
    tex_symbol(Symbol0,Symbol),
    add_options(Style,Symbol0,Fa,C),
    format("\\rput(~f,~f){\\circlenode[~a]{~w}{\\tiny ~w}}~n",[X,Y,C,Symbol,Symbol]),
    (	fsa_data:fsa_start_state(Fa,Symbol0)
    ->	X1 is X-40, 
	format("\\pnode(~f,~f){start~w}~n",[X1,Y,Symbol]),
	format("\\ncarc[arcangle=0]{->}{start~w}{~w}~n",[Symbol,Symbol])
    ;	true
    ).

add_options(plain,State,Fa,E) :-
    (	fsa_data:fsa_final_state(Fa,State)
    ->	E='doubleline=true'
    ;	E=''
    ).

add_options(fancy,State,Fa,E) :-
    (	fsa_data:fsa_final_state(Fa,State)
    ->	(   fsa_data:fsa_start_state(Fa,State)
	->  Type=start_final
	;   Type=final
	)
    ;   (   fsa_data:fsa_start_state(Fa,State)
	->  Type=start
	;   Type=normal
	)
    ),
    add_options0(Type,E).

add_options0(normal,'fillstyle=solid,fillcolor=cyan').
add_options0(final, 'doubleline=true,fillstyle=solid,fillcolor=tomato1,doublecolor=tomato1').
add_options0(start, 'fillstyle=solid,fillcolor=green').
add_options0(start_final,'doubleline=true,fillstyle=solid,fillcolor=plum2,doublecolor=plum2').

tex_transitions(Coords,Fa) :-
    tex_transitions(Coords,Coords,Fa).

tex_transitions([],_,_).
tex_transitions([H|T],All,Fa) :-
    tex_transitions0(H,All,Fa),
    tex_transitions(T,All,Fa).

%% combine all transitions connecting the same nodes into
%% a single transition with multiple labels
tex_transitions0(Start,All,Fa) :-
    (	setof(Sym,Fa^combined_tex_trans(Start,Sym,End,All,Fa),Syms),
	%% a real setof..
	tex_transition(Start,Syms,End),
	fail
    ;	true
    ).

combined_tex_trans(s(Start,_X0,_Y0),Sym,s(End,X,Y),All,Fa) :-
    member(s(End,X,Y),All),
    transition_or_jump(Fa,Start,Sym0,End),
    typeset_symbol(Sym0,Sym).

%typeset_symbol([],Sym) :-
%    !,
%    Sym='$E'.

typeset_symbol(S0/S,Sym0/Sym) :- !,
    typeset_symbol(S0,Sym0),
    typeset_symbol(S,Sym).

typeset_symbol(X,X).

tex_transition(s(Start,_,_),Syms0,s(End,_,_)) :-
    abbreviate_label(Syms0,Sym0),
    format("~n%transition from ~w to ~w with ~w~n",[Start,End,Sym0]),
    fsa_globals:fsa_global_get(v_angle,Angle0),
    Angle is 100*Angle0,
    tex_symbol(Sym0,Sym),
    (	Start==End
    ->	format("\\nccircle[angle=-90]{<-}{~w}{0.5cm}~n",[Start]),
	format("\\trput{\\rnode{x}{~w}}~n",[Sym])
    ;	format("\\ncarc[arcangle=~f]{->}{~w}{~w}~n",[Angle,Start,End]),
	format("\\naput{\\rnode{x}{~w}}~n",[Sym])
    ).

tex_symbol(Sym0,Sym) :-
    escape_specials(Sym0,Sym,
	[
	 [60]-[36,60,36],
	 [62]-[36,62,36],
	 [36,69]-[36,92,101,112,115,105,108,111,110,36],
	 [36]-[92,36],
	 [38]-[92,38],
	 [92]-[36,92,98,97,99,107,115,108,97,115,104,36],
	 [94]-[36,92,117,112,97,114,114,111,119,36],
	 [95]-[92,95],
	 [35]-[92,35],
	 [123]-[92,123],
	 [125]-[92,125]
	]).


%%%%%%%%%%%%%%%%%%%%%%%
%%%%% postscript %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%

user:help_info(xpred,fsa_postscript,"fsa_postscript(+File0,+File)","
Postscript code which produces visualization of automaton read from
File0 is written to File. The Postscript macro's are due to Peter
Kleiweg. The global variable postscript_res can be set to indicate
whether output is meant to be displayed on the screen, or printed.").

%% ps_fsa(InputFile,OutputFile)
%% for the current fsa (defined in Module) a postscript file OutputFile
%% is created...
%% This uses Peter Kleiweg's stuff...
fsa_postscript(Input,Output) :-
    fsa_io:fsa_read_file(Input,Fa),
#!sic#     fsa_postscript(Fa,Output,[]).
#sic#      fsa_postscript(Fa,Output,[fsa_tk]).

fsa_postscript(Fa,File,Mods):-
    fsa_util:to_file(fsa_visualization:fsa_postsript0(Fa,Mods),File),
    format(user_error,"Postscript figure saved in ~a~n",[File]).

fsa_postsript0(Fa,OverruleModules) :-
    compute_states(Coords0,Fa,OverruleModules),
    adjust_coords(Coords0,Coords,A,B,C,D),
    ps_begin(A,B,C,D),
    fsa_data:fsa_states_number(Fa,Size),
    fsa_array:fsa_array_new(NodeRefs,Size),
    ps_nodes(Coords,0,NodeRefs),
    ps_draw_nodes(Coords,NodeRefs,Fa),
    ps_initial_nodes(Coords,NodeRefs,Fa),
    ps_final_nodes(Coords,NodeRefs,Fa),
    ps_label_in_nodes(Coords,NodeRefs),
    ps_transitions(Coords,NodeRefs,Fa),
    ps_end.

%% y -> -y, and transpose to positive
%% x -> x-170
adjust_coords(Coords0,Coords,0,0,Xdist,Ydist) :-
    find_xys(Coords0,Xdist0,Ydist0,Xmin,_Ymin,_Xmax,Ymax),
    Xdist is Xdist0+150,
    Ydist is Ydist0+150,
    Yadj is Ymax+50,
    Xadj is Xmin-50,
    adjust_coords(Coords0,Coords,Xadj,Yadj).

adjust_coords([],[],_,_).
adjust_coords([s(Sym,X0,Y0)|Tail0],[s(Sym,X,Y)|Tail],Xadj,Yadj) :-
    X is X0-Xadj,
    Y is -Y0+Yadj,
    adjust_coords(Tail0,Tail,Xadj,Yadj).

find_xys(Coords,Xdist,Ydist,Xmin,Ymin,Xmax,Ymax) :-
    findall(X/Y,xy_state(Coords,X,Y),Xys0),
    sort(Xys0,Xys),
    xys(Xys,Xmin,Xmax,Ymin,Ymax),
    Ydist is abs(Ymax-Ymin),
    Xdist is abs(Xmax-Xmin).

ps_nodes([],_,_As).
ps_nodes([s(Symbol,X,Y)|Tail],I,Assoc) :-
    ps_node(X,Y,Symbol,I,NodeRef),
    I2 is I+1,
    fsa_array:fsa_array_access(Symbol,NodeRef,Assoc),
    ps_nodes(Tail,I2,Assoc).

ps_node(X,Y,_Symbol,I,NodeRef) :-
%    fsa_util:format_to_atom("q~w",[I],NodeRef),
    fsa_util:number__atom(I,J),
    atom_concat(q,J,NodeRef),
    format("/~a [ ~w ~w ] def~n",[NodeRef,X,Y]).

ps_initial_nodes([],_,_SSpec).
ps_initial_nodes([s(State,_,_)|Tail],Assoc,SSpec) :-
    fsa_array:fsa_array_access(State,NodeRef,Assoc),
    ps_initial_node(State,NodeRef,SSpec),
    ps_initial_nodes(Tail,Assoc,SSpec).

ps_initial_node(State,Ref,Spec) :-
    (	fsa_data:fsa_start_state(Spec,State)
    ->	format("~a 180 initialnode~n",[Ref])
    ;	true
    ).

ps_node_color(State,Fa,Color):-
    (	fsa_data:fsa_final_state(Fa,State)
    ->  (   fsa_data:fsa_start_state(Fa,State)
	->  Color='.93 .68 .93'   %% start and final
	;   Color='1 .39 .28'
	)
    ;   (   fsa_data:fsa_start_state(Fa,State)
	->  Color='0 1 0'
	;   Color='0 1 1'
	)
    ).

ps_draw_nodes([],_,_).
ps_draw_nodes([s(State,_,_)|Tail],Assoc,Fa) :-
    ps_node_color(State,Fa,Color),
    fsa_array:fsa_array_access(State,NodeRef,Assoc),
    format("~w ~w node~n",[NodeRef,Color]),
    ps_draw_nodes(Tail,Assoc,Fa).

ps_final_nodes([],_,_SSpec).
ps_final_nodes([s(State,_,_)|Tail],Assoc,SSpec) :-
    fsa_array:fsa_array_access(State,NodeRef,Assoc),
    ps_final_node(State,NodeRef,SSpec),
    ps_final_nodes(Tail,Assoc,SSpec).

ps_final_node(State,Ref,Spec) :-
    (	fsa_data:fsa_final_state(Spec,State)
    ->	format("~a finalnode~n",[Ref])
    ;	true
    ).

ps_label_in_nodes([],_).
ps_label_in_nodes([s(State,_,_)|Tail],Assoc) :-
    fsa_array:fsa_array_access(State,NodeRef,Assoc),
    ps_label_in_node(State,NodeRef),
    ps_label_in_nodes(Tail,Assoc).

ps_label_in_node(State,NodeRef) :-
    format("~a amoveto [ (~w) Rm ] Centre puttext~n",[NodeRef,State]).

ps_transitions(Coords,Assoc,TSpec) :-
    ps_transitions(Coords,Coords,Assoc,TSpec).

ps_transitions([],_,_,_).
ps_transitions([H|T],All,Refs,TSpec) :-
    ps_transitions0(H,All,Refs,TSpec),
    ps_transitions(T,All,Refs,TSpec).

%% combine all transitions combining the same nodes into
%% a single transition with multiple labels
ps_transitions0(Start,All,Refs,TSpec) :-
    (	setof(Sym,TSpec^combined_trans(Start,Sym,End,All,TSpec),Syms),
	%% a real setof..
	ps_transition(Start,Syms,Refs,End),
	fail
    ;	true
    ).

combined_trans(s(Start,_X0,_Y0),Sym,s(End,X,Y),All,TSpec) :-
    member(s(End,X,Y),All),
    transition_or_jump(TSpec,Start,Sym,End).

ps_transition(s(Start,X0,Y0),Syms0,Assoc,s(End,X,Y)) :-
    abbreviate_label(Syms0,Sym0),
    escape_specials(Sym0,Sym,[[92]-[92,92],[40]-[92,40],[41]-[92,41]]),
    fsa_array:fsa_array_access(Start,StartRef,Assoc),
    fsa_array:fsa_array_access(End,EndRef,Assoc),
    format("~n%transition from ~w to ~w with ~w~n",[Start,End,Sym]),
    get_angle(Angle),
    (	Start=End
    ->	format("~a ~a 0 .5 trans [ (~w) Rm ] East puttext~n",
	       [StartRef,EndRef,Sym])
    ;	direction(X0,Y0,X,Y,Dir),
	format("~a ~a ~w .5 trans [ (~w) Rm ] ~a puttext~n",
	       [StartRef,EndRef,Angle,Sym,Dir])
    ).

get_angle(Angle) :-
    fsa_globals:fsa_global_get(v_angle,Angle0),
    Angle1 is 100*Angle0,
    (	Angle1 > 120
    ->	Angle=120
    ;	Angle1 < 0
    ->	Angle=0
    ;	Angle=Angle1
    ).

%% x groter: rechts
%% y groter: hoger
direction(X,_,X,_,'East') :- !.
direction(_,Y,_,Y,'North'):- !.
direction(X0,Y0,X,Y,D) :-
    (	X0 < X, Y0 < Y
    ->	D = 'NorthWest'
    ;	X0 < X, Y0 > Y
    ->	D = 'SouthWest'
    ;	X0 > X, Y0 > Y
    ->	D = 'NorthWest'
    ;	X0 > X, Y0 < Y
    ->	D = 'SouthWest'
    ).

ps_begin(A0,B0,C0,D0) :-
    A is integer(A0),
    B is integer(B0),
    C is integer(C0),
    D is integer(D0),
    format("%!PS-Adobe-3.0 EPSF-3.0
%%BoundingBox: ~w ~w ~w ~w~n",[A,B,C,D]),
    fsa_globals:fsa_global_get(postscript_res,Res),
    format_postscript_header(Res),
    format("~npetersFSA begin~n",[]).

ps_end :-
    format("end~n~nshowpage~n",[]).

fsa_visualization(Type) :-
    fsa_globals:fsa_global_get(fa,Fa),
    fsa_visualization(Type,Fa).

user:help_info(xpred,fsa_visualization,
"fsa_visualization(+Format,+Fa)","
Starts an external visualization program visualizing Fa. Format
indicates what program is to be used and must be one of:

* vcg
* dot_ghostview (dot -Tps | gv )
* pstricks_ghostview (latex ; dvips ; gv)
* dotty
* davinci

").

#sic# fsa_visualization(vcg,Fa) :-
#sic#     exec('xvcg -',[pipe(Stream),std,std],_Pid),
#sic#     current_output(Old),
#sic#     set_output(Stream),
#sic#     user:call_cleanup(fsa_visualization:fsa_vcg(Fa),
#sic# 		      (	  set_output(Old),
#sic# 			  close(Stream)
#sic# 		      )).
#sic# 
#sic# %% problem: if ghostview is reading standard input, it doesn't honour
#sic# %% requested bounding box. Therefore we use temporary file.
#sic# fsa_visualization(dot_ghostview,Fa) :-
#sic#     fsa_temporary_file(Tmp0),
#sic#     atom_concat(Tmp0,'.ps',Tmp),
#sic#     format_to_chars('dot -Grankdir=LR -Tps >~a',[Tmp],CmdString),
#sic#     atom_codes(Cmd,CmdString),
#sic#     exec(Cmd,[pipe(Stream),std,std],_Pid),
#sic#     current_output(Old),
#sic#     set_output(Stream),
#sic#     user:call_cleanup(fsa_visualization:fsa_dot(Fa),
#sic# 		      (	  set_output(Old), close(Stream) )),
#sic#     format_to_chars('ghostview ~a',[Tmp],CmdString2),
#sic#     atom_codes(Cmd2,CmdString2),
#sic#     exec(Cmd2,[std,std,std],_Pid).
#sic# 
#sic# fsa_visualization(dot_xv,Fa) :-
#sic#     fsa_temporary_file(Tmp0),
#sic#     atom_concat(Tmp0,'.gif',Tmp),
#sic#     format_to_chars('dot -Grankdir=LR -Tgif >~w',[Tmp],CmdString),
#sic#     atom_codes(Cmd,CmdString),
#sic#     exec(Cmd,[pipe(Stream),std,std],_Pid),
#sic#     current_output(Old),
#sic#     set_output(Stream),
#sic#     user:call_cleanup(fsa_visualization:fsa_dot(Fa),
#sic# 		      (	  set_output(Old), close(Stream) )),
#sic#     format_to_chars('xv ~w',[Tmp],CmdString2),
#sic#     atom_codes(Cmd2,CmdString2),
#sic#     exec(Cmd2,[std,std,std],_Pid).
#sic# 
#sic# fsa_visualization(pstricks_ghostview,Fa) :-
#sic#     fsa_temporary_file(Base),
#sic#     atom_concat(Base,'.tex',Tmp),
#sic#     fsa_temporary_dir(Dir),
#sic#     fsa_pstricks(tex,Fa,Tmp,[fsa_tk]),
#sic#     format_to_chars(
#sic#        "( cd ~a ; latex ~a.tex ; dvips -f <~a.dvi >~a.ps ; ghostview ~a.ps )",
#sic#        [Dir,Base,Base,Base,Base],CmdString),
#sic#     atom_codes(Cmd,CmdString),
#sic#     exec(Cmd,[std,std,std],_).
#sic# 
#sic# fsa_visualization(dotty,Fa) :-
#sic#     exec('dotty -f fsa.lefty -',[pipe(Stream),std,std],_Pid),
#sic#     current_output(Old),
#sic#     set_output(Stream),
#sic#     user:call_cleanup(fsa_visualization:fsa_dot(Fa),
#sic# 		      (   set_output(Old),
#sic# 			  close(Stream)
#sic# 		      )).
#sic# 
#sic# fsa_visualization(davinci,Fa) :-
#sic#     exec('daVinci -pipe',[pipe(Stream),std,std],_Pid),
#sic#     current_output(Old),
#sic#     set_output(Stream),
#sic#     user:call_cleanup((   format('graph(new_placed(',[]),
#sic# 			  fsa_visualization:fsa_davinci(Fa),
#sic# 			  format('))',[])
#sic# 		      ),
#sic# 		      (	  set_output(Old),
#sic# 			  close(Stream)
#sic# 		      )).
#sic# 

#sic# fsa_temporary_file(File):-
#sic#     fsa_temporary_dir(TmpDir),
#sic#     fsa_util:format_to_atom("~a/fsaXXXXXX",[TmpDir],Pattern),
#sic#     mktemp(Pattern,File).
#sic# 
#sic# fsa_temporary_dir(TmpDir) :-
#sic#     (   environ('TMPDIR',TmpDir)
#sic#     ->  true
#sic#     ;   TmpDir = '/tmp'
#sic#     ).
#sic# 


limit_fsa_states_number(Fa0,Fa) :-
    fsa_data:fsa_states_number(Fa0,States),
    fsa_globals:fsa_global_get(no_display_beyond,Value),
    (	States =< Value
    ->	Fa0=Fa
    ;   format_to_chars("too large(~w); no_display_beyond=~w",[States,Value],Ch),
	atom_codes(Atom,Ch),
	fsa_regex:fsa_regex_compile(Atom,Fa)
    ).

format_postscript_header(normal) :-
    format("
%%%%%%%% START OF HEADER %%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                             %
%   Automaton transition diagram              %
%                                             %
%    (c) P. Kleiweg  1996, 1998               %
%                                             %
%   Manual available at:                      %
%   http://www.let.rug.nl/~~kleiweg/automata/  %
%                                             %  
%                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/petersFSA 256 dict  def
petersFSA begin

/NodeR  20 def			% Radius for node circle
/ENodeR NodeR 0.8 mul def	% Radius for inner circle of final node
/LoopR  NodeR 4   mul def	% Radius for loop arc
/LoopW	30 def			% Width for loop arc (degrees)
/NodeW   2 def			% Linewidth for node circle
/ENodeW  1 def			% Linewidth for inner circle of final node
/LineW   1 def			% Linewidth for arcs

/Color true def			% Should we use fill colors?

%%%% FONTS %%%%

/TopSkip    2 def		% Extra white space above text
/BottomSkip 2 def               % Extra white space below text
/SideSkip   2 def               % Extra white space aside of text

/SuperRaise  5 def		% Vertical position for superscript text
/SubLower   -3 def		% Vertical position for subscript text

% Stack in:  NewFontName EncodingVector OldFontName
% Stack out:
/RE {
  findfont
  dup maxlength dict begin
  {
    1 index /FID ne { def } { pop pop } ifelse
  } forall
  /Encoding exch def
  dup /FontName exch def
  currentdict end definefont pop
} bind def

/Times-Roman-ISOLatin1 ISOLatin1Encoding /Times-Roman RE

% Switch to regular font
% Stack in:
% Stack out:
/NormalFont {
  /Times-Roman-ISOLatin1 findfont
  11 scalefont
  setfont
} bind def

% Switch to font used for superscript and subscript
% Stack in:
% Stack out:
/SmallFont {
  /Times-Roman-ISOLatin1 findfont
  8 scalefont
  setfont
} bind def

% Switch to symbol font
% Stack in:
% Stack out:
/SymbolFont {
  /Symbol findfont
  11 scalefont
  setfont
} bind def

% Calculate height of normal font
gsave
  newpath
  NormalFont
  0 0 moveto
  (\\() false charpath
  pathbbox
grestore
/TOP exch def		% height above baseline
pop
/BOTTOM exch def	% depth below baseline (negative number)
pop
",[]),
    format("%%%% DISPLAYING TEXT %%%%

% Positions for displaying text relative to currentpoint
/Centre	0 def
/North	1 def
/East	2 def
/South	4 def
/West	8 def
/NorthEast North East or def
/NorthWest North West or def
/SouthEast South East or def
/SouthWest South West or def

% Text types
/Rm    1 def	% Use normal text (Roman)
/Super 2 def	% Use superscript
/Sub   3 def	% Use subscript
/Sym   4 def    % Use symbol font
/Ovl   5 def    % Use overlined normal text

% Calculate width of text in pixels
% Stack in:  [ (text1) type1 (text2) type2 ... ] 
% Stack out: width
/textwidth {
  /Txt exch def
  /L 0 def
  0 2 Txt length 2 sub {
    dup
    Txt exch get /T exch def
    Txt exch 1 add get /P exch def
    P Rm eq P Ovl eq or {
      /L L NormalFont T stringwidth pop add def
    } if
    P Super eq P Sub eq or {
      /L L SmallFont T stringwidth pop add def
    } if
    P Sym eq {
      /L L SymbolFont T stringwidth pop add def
    } if
  } for
  L
} bind def

% Display text at currentpoint
% Stack in:  [ (text1) type1 (text2) type2 ... ] position
% Stack out:
/puttext {
  /Pos exch def        % Centre, North, NorthEast, etc
  /Txt exch def

  /Width Txt textwidth def

  Pos West and 0 ne {
    Width neg SideSkip sub 0 rmoveto
  } {
    Pos East and 0 ne {
      SideSkip 0 rmoveto
    } {
      Width 2 div neg 0 rmoveto
    } ifelse
  } ifelse

  Pos North and 0 ne {
    0 BOTTOM neg BottomSkip add rmoveto
  } {
    Pos South and 0 ne {
      0 TOP neg TopSkip sub rmoveto
    } {
      0 TOP BOTTOM add 2 div neg rmoveto
    } ifelse
  } ifelse

  0 2 Txt length 2 sub {
    dup
    Txt exch get /T exch def
    Txt exch 1 add get /P exch def
    P Rm eq {
      NormalFont T show
    } if
    P Super eq {
      0 SuperRaise rmoveto
      SmallFont T show
      0 SuperRaise neg rmoveto
    } if
    P Sub eq {
      0 SubLower rmoveto
      SmallFont T show
      0 SubLower neg rmoveto
    } if
    P Sym eq {
      SymbolFont T show
    } if
    P Ovl eq {
      NormalFont
      gsave
        .5 setlinewidth
        0.6 T maxheight 1.5 add rmoveto
        T stringwidth pop 2 sub 0 rlineto stroke
      grestore
      T show
    } if
  } for
} bind def

% stack in:  (text)
% stack out: height
/maxheight {
  /MHT exch def
  /MHH 0 def
  0 1 MHT length 1 sub {
    MHT exch 1 getinterval
    gsave
      newpath
      0 0 moveto
      false charpath
      pathbbox
      dup MHH gt {
        /MHH exch def
      } {
        pop
      } ifelse
      pop pop pop
    grestore
  } for
  MHH
} bind def
",[]),
    format("%%%% DRAWING NODES %%%%

% Draw a node
% Stack in:  [ x y ] red green blue
% Stack out:
/node {
  4 -1 roll
  aload pop
  /Y exch def
  /X exch def

  NodeW setlinewidth
  X NodeR add Y moveto
  X Y NodeR 0 360 arc
  Color {
    gsave
      setrgbcolor
      fill
    grestore
  } {
    pop pop pop
  } ifelse
  stroke
} bind def

% Make node initial
% Stack in:   [ x y ] angle
% Stack uit:
/initialnode {
  gsave
    exch
    aload pop
    translate
    rotate
    NodeR NodeW 2 div add 0 translate
    /N NodeR 3 div def
    0 0 moveto
    N N neg lineto
    N 2 div 0 N 2 div 0 N N curveto
    closepath
    fill
  grestore
} bind def    

% Make node final
% Stack in:   [ x y ]
% Stack out:
/finalnode {
  aload pop
  /Y exch def
  /X exch def
  ENodeW setlinewidth
  X ENodeR add Y moveto
  X Y ENodeR 0 360 arc
  stroke
} bind def

%%%% DRAWING TRANSITIONS %%%%

% Draw arrowhead at x2 y2 from direction x1 y1
% Stack in:  x1 y1 x2 y2
% Stack out:
/arrowHead {
  /AY2 exch def
  /AX2 exch def
  /AY1 exch def
  /AX1 exch def
  gsave
    AX2 AY2 translate
    AY2 AY1 sub AX2 AX1 sub atan rotate
    0 0 moveto
    LineW neg 2 div 0 translate
    -8 LineW mul 2 LineW mul lineto
    -6 LineW mul 0
    -6 LineW mul 0
    -8 LineW mul -2 LineW mul curveto
    closepath
    fill
  grestore
} bind def

% Draw a transition
% Stack in:  [ x1 y1 ] [ x2 y2 ] angle part arrow1 arrow2
% Stack out:
/transition {
  /Arrow2 exch def	% Arrow to final node: true / false
  /Arrow1 exch def	% Arrow ot starting node: true / false
  /F exch def		% Position for label (value between 0 and 1)
  /A exch def		% Angle (value between -120 and 120 exclusive for non-loops)
  aload pop
  /Y2 exch def		% Final node
  /X2 exch def		%
  aload pop
  /Y1 exch def		% Starting node
  /X1 exch def		%

  /Clear LineW 4 mul def
  LineW setlinewidth
  X1 X2 eq Y1 Y2 eq and {
    Arrow1 {
      /XX0 A LoopW add cos NodeR NodeW 2 div add mul X1 add def
      /YY0 A LoopW add sin NodeR NodeW 2 div add mul Y1 add def
      /XX0i A LoopW add cos NodeR NodeW 2 div add Clear add mul X1 add def
      /YY0i A LoopW add sin NodeR NodeW 2 div add Clear add mul Y1 add def
    } {
      /XX0 A LoopW add cos NodeR mul X1 add def
      /YY0 A LoopW add sin NodeR mul Y1 add def
      /XX0i XX0 def
      /YY0i YY0 def
    } ifelse
    Arrow2 {
      /XX3 A LoopW sub cos NodeR NodeW 2 div add mul X1 add def
      /YY3 A LoopW sub sin NodeR NodeW 2 div add mul Y1 add def
      /XX3i A LoopW sub cos NodeR NodeW 2 div add Clear add mul X1 add def
      /YY3i A LoopW sub sin NodeR NodeW 2 div add Clear add mul Y1 add def
    } {
      /XX3 A LoopW sub cos NodeR mul X1 add def
      /YY3 A LoopW sub sin NodeR mul Y1 add def
      /XX3i XX3 def
      /YY3i YY3 def
    } ifelse
    /XX1 A LoopW add cos LoopR mul X1 add def
    /YY1 A LoopW add sin LoopR mul Y1 add def
    /XX2 A LoopW sub cos LoopR mul X1 add def
    /YY2 A LoopW sub sin LoopR mul Y1 add def
  } {
    /ALFA Y2 Y1 sub X2 X1 sub atan def
    Arrow1 {
      /XX0 A ALFA add cos NodeR NodeW 2 div add mul X1 add def
      /YY0 A ALFA add sin NodeR NodeW 2 div add mul Y1 add def
      /XX0i A ALFA add cos NodeR NodeW 2 div add Clear add mul X1 add def
      /YY0i A ALFA add sin NodeR NodeW 2 div add Clear add mul Y1 add def
    } {
      /XX0 A ALFA add cos NodeR mul X1 add def
      /YY0 A ALFA add sin NodeR mul Y1 add def
      /XX0i XX0 def
      /YY0i YY0 def
    } ifelse
    Arrow2 {
      /XX3  ALFA 180 add A sub cos NodeR NodeW 2 div add mul X2 add def
      /YY3  ALFA 180 add A sub sin NodeR NodeW 2 div add mul Y2 add def
      /XX3i ALFA 180 add A sub cos NodeR NodeW 2 div add Clear add mul X2 add def
      /YY3i ALFA 180 add A sub sin NodeR NodeW 2 div add Clear add mul Y2 add def
    } {
      /XX3  ALFA 180 add A sub cos NodeR mul X2 add def
      /YY3  ALFA 180 add A sub sin NodeR mul Y2 add def
      /XX3i XX3 def
      /YY3i YY3 def
    } ifelse
    /LEN
      XX0 XX3 sub dup mul YY0 YY3 sub dup mul add sqrt
      1 2 A cos mul add
      div
    def
    /XX1 XX0 LEN ALFA A add cos mul add def
    /YY1 YY0 LEN ALFA A add sin mul add def
    /XX2 XX3 LEN ALFA 180 add A sub cos mul add def
    /YY2 YY3 LEN ALFA 180 add A sub sin mul add def
  } ifelse

  XX0i YY0i moveto
  XX1 YY1 XX2 YY2 XX3i YY3i curveto
  stroke

  /Cx XX1 XX0i sub 3 mul def
  /Cy YY1 YY0i sub 3 mul def
  /Bx XX2 XX1 sub 3 mul Cx sub def
  /By YY2 YY1 sub 3 mul Cy sub def
  /Ax XX3i XX0i sub Cx sub Bx sub def
  /Ay YY3i YY0i sub Cy sub By sub def

  /XL F F F mul mul Ax mul F F mul Bx mul add F Cx mul add XX0i add def
  /YL F F F mul mul Ay mul F F mul By mul add F Cy mul add YY0i add def

  Arrow1 {
      XX1 YY1 XX0 YY0 arrowHead
  } if
  Arrow2 {
      XX2 YY2 XX3 YY3 arrowHead
  } if

  XL YL moveto
} bind def

% Draw a transition with arrow to final node
% Stack in:  [ x1 y1 ] [ x2 y2 ] angle part
% Stack out:
/trans {
  false true transition
} bind def

%%%% MISC %%%%

% Stack in:  [ x y ]
% Stack out:
% Result:    x y moveto
/amoveto {
  aload pop
  moveto
} bind def

end

%%%%%%%% END OF HEADER %%%%%%%%

",[]).

format_postscript_header(lowres) :-
    format("
%%%%%%%% START OF HEADER %%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                             %
%   Automaton transition diagram              %
%                                             %
%   (c) P. Kleiweg  1996, 1998                %
%                                             %
%   Manual available at:                      %
%   http://www.let.rug.nl/~~kleiweg/automata/  %
%                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/petersFSA 256 dict  def
petersFSA begin

/NodeR  20 def			% Radius for node circle
/ENodeR NodeR 0.8 mul def	% Radius for inner circle of final node
/LoopR  NodeR 4   mul def	% Radius for loop arc
/LoopW	30 def			% Width for loop arc (degrees)
/NodeW   2 def			% Linewidth for node circle
/ENodeW  1 def			% Linewidth for inner circle of final node
/LineW   1 def			% Linewidth for arcs

/Color true def			% Should we use fill colors?

%%%% FONTS %%%%

/TopSkip    4 def		% Extra white space above text
/BottomSkip 2 def               % Extra white space below text
/SideSkip   3 def               % Extra white space aside of text

/SuperRaise  5 def		% Vertical position for superscript text
/SubLower   -4 def		% Vertical position for subscript text

% Stack in:  NewFontName EncodingVector OldFontName
% Stack out:
/RE {
  findfont
  dup maxlength dict begin
  {
    1 index /FID ne { def } { pop pop } ifelse
  } forall
  /Encoding exch def
  dup /FontName exch def
  currentdict end definefont pop
} bind def

/Times-Roman-ISOLatin1 ISOLatin1Encoding /Times-Roman RE

% Switch to regular font
% Stack in:
% Stack out:
/NormalFont {
  /Times-Roman-ISOLatin1 findfont
  14 scalefont
  setfont
} bind def

% Switch to font used for superscript and subscript
% Stack in:
% Stack out:
/SmallFont {
  /Times-Roman-ISOLatin1 findfont
  10 scalefont
  setfont
} bind def

% Switch to symbol font
% Stack in:
% Stack out:
/SymbolFont {
  /Symbol findfont
  14 scalefont
  setfont
} bind def

% Calculate height of normal font
gsave
  newpath
  NormalFont
  0 0 moveto
  (\\() false charpath
  pathbbox
grestore
/TOP exch def		% height above baseline
pop
/BOTTOM exch def	% depth below baseline (negative number)
pop
",[]),
    format("%%%% DISPLAYING TEXT %%%%

% Positions for displaying text relative to currentpoint
/Centre	0 def
/North	1 def
/East	2 def
/South	4 def
/West	8 def
/NorthEast North East or def
/NorthWest North West or def
/SouthEast South East or def
/SouthWest South West or def

% Text types
/Rm    1 def	% Use normal text (Roman)
/Super 2 def	% Use superscript
/Sub   3 def	% Use subscript
/Sym   4 def    % Use symbol font
/Ovl   5 def    % Use overlined normal text

% Calculate width of text in pixels
% Stack in:  [ (text1) type1 (text2) type2 ... ] 
% Stack out: width
/textwidth {
  /Txt exch def
  /L 0 def
  0 2 Txt length 2 sub {
    dup
    Txt exch get /T exch def
    Txt exch 1 add get /P exch def
    P Rm eq P Ovl eq or {
      /L L NormalFont T stringwidth pop add def
    } if
    P Super eq P Sub eq or {
      /L L SmallFont T stringwidth pop add def
    } if
    P Sym eq {
      /L L SymbolFont T stringwidth pop add def
    } if
  } for
  L
} bind def

% Display text at currentpoint
% Stack in:  [ (text1) type1 (text2) type2 ... ] position
% Stack out:
/puttext {
  /Pos exch def        % Centre, North, NorthEast, etc
  /Txt exch def

  /Width Txt textwidth def

  Pos West and 0 ne {
    Width neg SideSkip sub 0 rmoveto
  } {
    Pos East and 0 ne {
      SideSkip 0 rmoveto
    } {
      Width 2 div neg 0 rmoveto
    } ifelse
  } ifelse

  Pos North and 0 ne {
    0 BOTTOM neg BottomSkip add rmoveto
  } {
    Pos South and 0 ne {
      0 TOP neg TopSkip sub rmoveto
    } {
      0 TOP BOTTOM add 2 div neg rmoveto
    } ifelse
  } ifelse

  0 2 Txt length 2 sub {
    dup
    Txt exch get /T exch def
    Txt exch 1 add get /P exch def
    P Rm eq {
      NormalFont T show
    } if
    P Super eq {
      0 SuperRaise rmoveto
      SmallFont T show
      0 SuperRaise neg rmoveto
    } if
    P Sub eq {
      0 SubLower rmoveto
      SmallFont T show
      0 SubLower neg rmoveto
    } if
    P Sym eq {
      SymbolFont T show
    } if
    P Ovl eq {
      NormalFont
      gsave
        .5 setlinewidth
        0.6 T maxheight 2 add rmoveto
        T stringwidth pop 2 sub 0 rlineto stroke
      grestore
      T show
    } if
  } for
} bind def

% stack in:  (text)
% stack out: height
/maxheight {
  /MHT exch def
  /MHH 0 def
  0 1 MHT length 1 sub {
    MHT exch 1 getinterval
    gsave
      newpath
      0 0 moveto
      false charpath
      pathbbox
      dup MHH gt {
        /MHH exch def
      } {
        pop
      } ifelse
      pop pop pop
    grestore
  } for
  MHH
} bind def
",[]),
    format("%%%% DRAWING NODES %%%%

% Draw a node
% Stack in:  [ x y ] red green blue
% Stack out:
/node {
  4 -1 roll
  aload pop
  /Y exch def
  /X exch def

  NodeW setlinewidth
  X NodeR add Y moveto
  X Y NodeR 0 360 arc
  Color {
    gsave
      setrgbcolor
      fill
    grestore
  } {
    pop pop pop
  } ifelse
  stroke
} bind def

% Make node initial
% Stack in:   [ x y ] angle
% Stack uit:
/initialnode {
  gsave
    exch
    aload pop
    translate
    rotate
    NodeR NodeW 2 div add 0 translate
    /N NodeR 3 div def
    0 0 moveto
    N N neg lineto
    N 2 div 0 N 2 div 0 N N curveto
    closepath
    fill
  grestore
} bind def    

% Make node final
% Stack in:   [ x y ]
% Stack out:
/finalnode {
  aload pop
  /Y exch def
  /X exch def
  ENodeW setlinewidth
  X ENodeR add Y moveto
  X Y ENodeR 0 360 arc
  stroke
} bind def

%%%% DRAWING TRANSITIONS %%%%

% Draw arrowhead at x2 y2 from direction x1 y1
% Stack in:  x1 y1 x2 y2
% Stack out:
/arrowHead {
  /AY2 exch def
  /AX2 exch def
  /AY1 exch def
  /AX1 exch def
  gsave
    AX2 AY2 translate
    AY2 AY1 sub AX2 AX1 sub atan rotate
    0 0 moveto
    LineW neg 2 div 0 translate
    -8 LineW mul 2 LineW mul lineto
    -6 LineW mul 0
    -6 LineW mul 0
    -8 LineW mul -2 LineW mul curveto
    closepath
    fill
  grestore
} bind def

% Draw a transition
% Stack in:  [ x1 y1 ] [ x2 y2 ] angle part arrow1 arrow2
% Stack out:
/transition {
  /Arrow2 exch def	% Arrow to final node: true / false
  /Arrow1 exch def	% Arrow ot starting node: true / false
  /F exch def		% Position for label (value between 0 and 1)
  /A exch def		% Angle (value between -120 and 120 exclusive for non-loops)
  aload pop
  /Y2 exch def		% Final node
  /X2 exch def		%
  aload pop
  /Y1 exch def		% Starting node
  /X1 exch def		%

  /Clear LineW 4 mul def
  LineW setlinewidth
  X1 X2 eq Y1 Y2 eq and {
    Arrow1 {
      /XX0 A LoopW add cos NodeR NodeW 2 div add mul X1 add def
      /YY0 A LoopW add sin NodeR NodeW 2 div add mul Y1 add def
      /XX0i A LoopW add cos NodeR NodeW 2 div add Clear add mul X1 add def
      /YY0i A LoopW add sin NodeR NodeW 2 div add Clear add mul Y1 add def
    } {
      /XX0 A LoopW add cos NodeR mul X1 add def
      /YY0 A LoopW add sin NodeR mul Y1 add def
      /XX0i XX0 def
      /YY0i YY0 def
    } ifelse
    Arrow2 {
      /XX3 A LoopW sub cos NodeR NodeW 2 div add mul X1 add def
      /YY3 A LoopW sub sin NodeR NodeW 2 div add mul Y1 add def
      /XX3i A LoopW sub cos NodeR NodeW 2 div add Clear add mul X1 add def
      /YY3i A LoopW sub sin NodeR NodeW 2 div add Clear add mul Y1 add def
    } {
      /XX3 A LoopW sub cos NodeR mul X1 add def
      /YY3 A LoopW sub sin NodeR mul Y1 add def
      /XX3i XX3 def
      /YY3i YY3 def
    } ifelse
    /XX1 A LoopW add cos LoopR mul X1 add def
    /YY1 A LoopW add sin LoopR mul Y1 add def
    /XX2 A LoopW sub cos LoopR mul X1 add def
    /YY2 A LoopW sub sin LoopR mul Y1 add def
  } {
    /ALFA Y2 Y1 sub X2 X1 sub atan def
    Arrow1 {
      /XX0 A ALFA add cos NodeR NodeW 2 div add mul X1 add def
      /YY0 A ALFA add sin NodeR NodeW 2 div add mul Y1 add def
      /XX0i A ALFA add cos NodeR NodeW 2 div add Clear add mul X1 add def
      /YY0i A ALFA add sin NodeR NodeW 2 div add Clear add mul Y1 add def
    } {
      /XX0 A ALFA add cos NodeR mul X1 add def
      /YY0 A ALFA add sin NodeR mul Y1 add def
      /XX0i XX0 def
      /YY0i YY0 def
    } ifelse
    Arrow2 {
      /XX3  ALFA 180 add A sub cos NodeR NodeW 2 div add mul X2 add def
      /YY3  ALFA 180 add A sub sin NodeR NodeW 2 div add mul Y2 add def
      /XX3i ALFA 180 add A sub cos NodeR NodeW 2 div add Clear add mul X2 add def
      /YY3i ALFA 180 add A sub sin NodeR NodeW 2 div add Clear add mul Y2 add def
    } {
      /XX3  ALFA 180 add A sub cos NodeR mul X2 add def
      /YY3  ALFA 180 add A sub sin NodeR mul Y2 add def
      /XX3i XX3 def
      /YY3i YY3 def
    } ifelse
    /LEN
      XX0 XX3 sub dup mul YY0 YY3 sub dup mul add sqrt
      1 2 A cos mul add
      div
    def
    /XX1 XX0 LEN ALFA A add cos mul add def
    /YY1 YY0 LEN ALFA A add sin mul add def
    /XX2 XX3 LEN ALFA 180 add A sub cos mul add def
    /YY2 YY3 LEN ALFA 180 add A sub sin mul add def
  } ifelse

  XX0i YY0i moveto
  XX1 YY1 XX2 YY2 XX3i YY3i curveto
  stroke

  /Cx XX1 XX0i sub 3 mul def
  /Cy YY1 YY0i sub 3 mul def
  /Bx XX2 XX1 sub 3 mul Cx sub def
  /By YY2 YY1 sub 3 mul Cy sub def
  /Ax XX3i XX0i sub Cx sub Bx sub def
  /Ay YY3i YY0i sub Cy sub By sub def

  /XL F F F mul mul Ax mul F F mul Bx mul add F Cx mul add XX0i add def
  /YL F F F mul mul Ay mul F F mul By mul add F Cy mul add YY0i add def

  Arrow1 {
      XX1 YY1 XX0 YY0 arrowHead
  } if
  Arrow2 {
      XX2 YY2 XX3 YY3 arrowHead
  } if

  XL YL moveto
} bind def

% Draw a transition with arrow to final node
% Stack in:  [ x1 y1 ] [ x2 y2 ] angle part
% Stack out:
/trans {
  false true transition
} bind def

%%%% MISC %%%%

% Stack in:  [ x y ]
% Stack out:
% Result:    x y moveto
/amoveto {
  aload pop
  moveto
} bind def

end

%%%%%%%% END OF HEADER %%%%%%%%
",[]).


